import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/*
A convenient class for displaying option dialogs.

The ask() method takes an array of labels (questions) and
default answers, and returns an array of answers.

The confirm() method takes a question, and returns true if
OK was pressed, or false if Cancel was pressed.
*/

public class GEAskDialog extends JDialog implements ActionListener
{
	protected Object[]				cDefaultAnswer;
	protected JComponent[]			cAnswer;
	protected JButton				cOKButton;
	protected JButton				cCancelButton;
	protected boolean				cConfirmed; // ok button pressed to dismiss dialog
	
	
	public static Object[] ask(JFrame owner, int width, String title, String[] label, Object[] defaultAnswer) {
		GEAskDialog					dialog;
		
		dialog = new GEAskDialog(width, title, label, defaultAnswer, owner);
		dialog.show();
		return dialog.getAnswer();
	}
	
	public static boolean confirm(JFrame owner, String title, String question) {
		return JOptionPane.showConfirmDialog(owner,
				question, title,
				JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION;
	}
	
	public GEAskDialog(int width, String title, String[] label, Object[] defaultAnswer,
			JFrame owner) {
		
		super(owner, true);
		
		JPanel						textPanel;
		JPanel						buttonPanel;
		int							textCounter;
		int							heightPerField;
		
		setTitle(title);
		
		if (defaultAnswer == null && label.length == 1) {
			JTextArea				textArea;
			
			textPanel = new JPanel(new BorderLayout());
			textPanel.setBorder(new EmptyBorder(4, 4, 4, 4));
			textArea = new JTextArea(label[0]);
			textArea.setBackground(getContentPane().getBackground());
			textArea.setEditable(false);
			textArea.setLineWrap(true);
			textArea.setWrapStyleWord(true);
			textPanel.add(textArea);
		} else {
			cDefaultAnswer = defaultAnswer;
			heightPerField = 18;
			
			// init text panel
			if (label == null)
				throw new IllegalArgumentException("No questions. (label == null)");
			if (defaultAnswer == null)
				throw new IllegalArgumentException("No default answers. (defaultAnswer == null)");
			if (defaultAnswer.length != label.length)
				throw new IllegalArgumentException("Not equal number of questions and answers. (label.length != defaultAnswer.length)");
			
			textPanel = new JPanel(new GridLayout(label.length, 2));
			textPanel.setBorder(new EmptyBorder(3, 3, 3, 0));
			cAnswer = new JComponent[label.length];
			for (textCounter = 0; textCounter < label.length; textCounter++) {
				if (defaultAnswer[textCounter] instanceof String) {
					textPanel.add(new JLabel(label[textCounter]));
					cAnswer[textCounter] = new JTextField((String) defaultAnswer[textCounter], 20);
					textPanel.add(cAnswer[textCounter]);
				}
				else if (defaultAnswer[textCounter] instanceof Integer) {
					textPanel.add(new JLabel(label[textCounter]));
					cAnswer[textCounter] = new JTextField(defaultAnswer[textCounter].toString(), 20);
					textPanel.add(cAnswer[textCounter]);
				}
				else if (defaultAnswer[textCounter] instanceof Boolean) {
					cAnswer[textCounter] = new JCheckBox(label[textCounter], ((Boolean) defaultAnswer[textCounter]).booleanValue());
					textPanel.add(new JLabel(""));
					textPanel.add(cAnswer[textCounter]);
					heightPerField = 22;
				} else {
					System.err.println("no valid default answer " + textCounter);
				}
			}
		}
		
		// init buttons
		buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		cOKButton = new JButton("OK");
		cOKButton.addActionListener(this);
		buttonPanel.add(cOKButton);
		cCancelButton = new JButton("Cancel");
		cCancelButton.addActionListener(this);
		buttonPanel.add(cCancelButton);
		
		cConfirmed = false;
		
		// wrap up layout
		if (label.length > 10) {
			JScrollPane				scrollPane;
			JViewport				port;
			Dimension				size;
			JPanel					viewPanel;
			
			textPanel.setPreferredSize(new Dimension(100, textPanel.getPreferredSize().height));
			viewPanel = new JPanel(new BorderLayout());
			viewPanel.add(textPanel, BorderLayout.NORTH);
			scrollPane = new JScrollPane(viewPanel, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
			
			getContentPane().add(scrollPane, BorderLayout.CENTER);
			setSize(width, 250);
		} else {
			getContentPane().add(textPanel, BorderLayout.NORTH);
			setSize(width, textPanel.getPreferredSize().height + 50);
		}
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
	}
	
	public void actionPerformed(ActionEvent theEvent) {
		Object						source;
		
		source = theEvent.getSource();
		
		if (source == cOKButton) {
			Object[]				answer;
			int						answerCounter;
			
			cConfirmed = true;
			
			// check that all fields have valid answers
			answer = getAnswer();
			
			if (answer != null)
				for (answerCounter = 0; answerCounter < answer.length; answerCounter++) {
					if (answer[answerCounter] == null) {
						// an invalid answer
						
						if (cAnswer[answerCounter] instanceof JTextField) {
							cAnswer[answerCounter].requestFocus();
							((JTextField)cAnswer[answerCounter]).selectAll();
						}
						
						return;
					}
				}
			
			dispose();
		} else if (source == cCancelButton) {
			cConfirmed = false;
			dispose();
		}
	}
	
	public Object[] getAnswer() {
		Object[]					returnValue;
		int							textCounter;
		
		if (!cConfirmed || cAnswer == null || cDefaultAnswer == null)
			return null;
		
		returnValue = new Object[cAnswer.length];
		for (textCounter = 0; textCounter < cAnswer.length; textCounter++) {
			if (cDefaultAnswer[textCounter] instanceof String)
				returnValue[textCounter] = ((JTextField) cAnswer[textCounter]).getText();
			else if (cDefaultAnswer[textCounter] instanceof Integer) {
				try {
					returnValue[textCounter] = new Integer(((JTextField) cAnswer[textCounter]).getText());
				} catch (NumberFormatException e) {
					returnValue[textCounter] = null;
				}
			} else if (cDefaultAnswer[textCounter] instanceof Boolean)
				returnValue[textCounter] = new Boolean(((JCheckBox) cAnswer[textCounter]).isSelected());
		}
		
		return returnValue;
	}
}

