import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;


/*
Sits in a GEImageEditFrame, and displays four color/alpha sliders
and a magnified view of a GPixels object, and handles most
of the image editing.
*/

public class GEImageEditCanvas extends JComponent
		implements MouseListener, MouseMotionListener
{
	static final int				kMargin = 8;
	static final int				kButtonHeight = 20;
	static final int				kSliderWidth = 16;
	static final int				kMinSliderHeight = 60;
	static final int				kMaxScale = 24;
	static final int				kMinScale = 2;
	static final int				kColorTextHeight = 15;
	
	static final int				kNoGrid = 0;
	static final int				kDotGrid = 1;
	static final int				kLineGrid = 2;
	static final int				kMinScaleGrid = 3;
	
	
	protected int					cPressedX;
	protected int					cPressedY;
	
	protected GEPixels				cImage;
	protected GEPixels				cUndoImage;
	
	protected int					cGridMode;
	protected int					cSliderHeight;
	
	protected boolean				cIgnoreRed;
	protected boolean				cIgnoreGreen;
	protected boolean				cIgnoreBlue;
	protected boolean				cIgnoreAlpha;
	
	protected int					cRedSlider;
	protected int					cGreenSlider;
	protected int					cBlueSlider;
	protected int					cAlphaSlider;
	
	protected int					cImageLeft;
	
	protected int					cScaleFactor;
	
	
	public GEImageEditCanvas(GPixels image) {
		cImage = (GEPixels) image;
		
		cImageLeft = 5 * kMargin + 4 * kSliderWidth;
		
		prepareUndo();
		
		addMouseListener(this);
		addMouseMotionListener(this);
		
		cGridMode = kNoGrid;
		
		cIgnoreRed = false;
		cIgnoreGreen = false;
		cIgnoreBlue = false;
		cIgnoreAlpha = false;
		
		cRedSlider = 127;
		cGreenSlider = 127;
		cBlueSlider = 127;
		cAlphaSlider = 255;
		
		prepareScaling();
	}
	
	public void prepareUndo() {
		if (cImage != null)
			cUndoImage = (GEPixels) cImage.clone();
	}
	
	public void undo() {
		GEPixels						temp;
		
		// swap cImage with cUndoImage, but all external
		// references needs to be kept intact
		temp = (GEPixels) cImage.clone();
		cImage.copyFrom(cUndoImage);
		cUndoImage.copyFrom(temp);
		
		prepareScaling();
		repaint();
	}
	
	public void transform(int mode) {
		prepareUndo();
		cImage.transform(mode);
		prepareScaling();
		repaint();
	}
	
	public void updateImageSize() {
		prepareScaling();
		repaint();
	}
	
	public void paint(Graphics theGraphics) {
		int							x;
		int							y;
		int							pixelIndex;
		int							left;
		int							top;
		int							size;
		
		theGraphics.setFont(new Font("SansSerif", Font.PLAIN, 10));
		
		// red slider + text
		theGraphics.setColor(Color.red.darker().darker());
		if (!cIgnoreRed)
			theGraphics.drawString("" + cRedSlider, kMargin, kMargin * 2 + cSliderHeight + kColorTextHeight);
		
		theGraphics.drawRect(kMargin, kMargin, kSliderWidth, cSliderHeight);
		theGraphics.setColor(Color.red);
		theGraphics.fillRect(kMargin + 1, kMargin + 1 + (255 - cRedSlider) * cSliderHeight / 255, kSliderWidth - 1, cSliderHeight - (255 - cRedSlider) * cSliderHeight / 255 - 1);
		
		// green slider + text
		theGraphics.setColor(Color.green.darker().darker());
		if (!cIgnoreGreen)
			theGraphics.drawString("" + cGreenSlider, kMargin * 2 + kSliderWidth, kMargin * 2 + cSliderHeight + kColorTextHeight);
		
		theGraphics.drawRect(kMargin * 2 + kSliderWidth, kMargin, kSliderWidth, cSliderHeight);
		theGraphics.setColor(Color.green);
		theGraphics.fillRect(kMargin * 2 + kSliderWidth + 1, kMargin + 1 + (255 - cGreenSlider) * cSliderHeight / 255, kSliderWidth - 1, cSliderHeight - (255 - cGreenSlider) * cSliderHeight / 255 - 1);
		
		// blue slider + text
		theGraphics.setColor(Color.blue.darker());
		if (!cIgnoreBlue)
			theGraphics.drawString("" + cBlueSlider, kMargin * 3 + kSliderWidth * 2, kMargin * 2 + cSliderHeight + kColorTextHeight);
		
		theGraphics.drawRect(kMargin * 3 + kSliderWidth * 2, kMargin, kSliderWidth, cSliderHeight);
		theGraphics.setColor(Color.blue.brighter());
		theGraphics.fillRect(kMargin * 3 + kSliderWidth * 2 + 1, kMargin + 1 + (255 - cBlueSlider) * cSliderHeight / 255, kSliderWidth - 1, cSliderHeight - (255 - cBlueSlider) * cSliderHeight / 255 - 1);
		
		// alpha slider + text
		theGraphics.setColor(Color.black);
		if (!cIgnoreAlpha)
			theGraphics.drawString("" + cAlphaSlider, kMargin * 4 + kSliderWidth * 3, kMargin * 2 + cSliderHeight + kColorTextHeight);
		
		theGraphics.drawLine(kMargin * 4 + kSliderWidth * 3 + 1, kMargin + (255 - cAlphaSlider) * cSliderHeight / 255, (kMargin + kSliderWidth) * 4, kMargin + (255 - cAlphaSlider) * cSliderHeight / 255);
		theGraphics.drawRect(kMargin * 4 + kSliderWidth * 3, kMargin, kSliderWidth, cSliderHeight);
		theGraphics.setColor(new Color(cRedSlider, cGreenSlider, cBlueSlider));
		theGraphics.fillRect(kMargin * 4 + kSliderWidth * 3 + 1, kMargin + 1 + (255 - cAlphaSlider) * cSliderHeight / 255, kSliderWidth - 1, cSliderHeight - (255 - cAlphaSlider) * cSliderHeight / 255 - 1);
		
		// the image
		theGraphics.setColor(Color.black);
		
		if (cScaleFactor >= kMinScaleGrid && cGridMode == kLineGrid) { // gridlines
			theGraphics.drawRect(cImageLeft, kMargin, cScaleFactor * cImage.getWidth() + 2, cScaleFactor * cImage.getHeight() + 2);
			size = cScaleFactor - 1;
			left = cImageLeft + 2;
			top = kMargin + 2;
		} else { // no gridlines
			theGraphics.drawRect(cImageLeft, kMargin, cScaleFactor * cImage.getWidth() + 1, cScaleFactor * cImage.getHeight() + 1);
			size = cScaleFactor;
			left = cImageLeft + 1;
			top = kMargin + 1;
		}
		pixelIndex = 0;
		
		for (y = 0; y < cImage.getHeight(); y++) {
			for (x = 0; x < cImage.getWidth(); x++) {
				theGraphics.setColor(new Color(cImage.getPixel(pixelIndex) | 0xFF000000));
				theGraphics.fillRect(left, top, size * ((cImage.getPixel(pixelIndex) >> 24) & 0xFF) / 255, size);
				
				if (cGridMode == kDotGrid) {
					theGraphics.setColor(new Color((cImage.getPixel(pixelIndex) | 0xFF000000) ^ 0x00808080));
					theGraphics.drawLine(left, top, left, top);
				}
				
				left += cScaleFactor;
				pixelIndex++;
			}
			
			top += cScaleFactor;
			left -= cScaleFactor * cImage.getWidth();
		}
		
		theGraphics.setColor(Color.black);
	}
	
	public void prepareScaling() {
		int							xScale;
		int							yScale;
		
		xScale = (getWidth() - kMargin - cImageLeft) / cImage.getWidth();
		yScale = (getHeight() - 2 * kMargin) / cImage.getHeight();
		
		cScaleFactor = (xScale > yScale) ? yScale : xScale;
		
		if (cScaleFactor < kMinScale)
			cScaleFactor = kMinScale;
		if (cScaleFactor > kMaxScale)
			cScaleFactor = kMaxScale;
	}
	
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		
		prepareScaling();
		cSliderHeight = getHeight() - kColorTextHeight - 3 * kMargin;
	}
	
	public void toggleGrid() {
		if (cGridMode == kNoGrid)
			cGridMode = kDotGrid;
		else if (cGridMode == kDotGrid)
			cGridMode = kLineGrid;
		else if (cGridMode == kLineGrid)
			cGridMode = kNoGrid;
		
		repaint();
	}
	
	public int getAlpha() {
		if (cIgnoreAlpha)
			return -1;
		return cAlphaSlider;
	}
	
	public int getRed() {
		if (cIgnoreRed)
			return -1;
		return cRedSlider;
	}
	
	public int getGreen() {
		if (cIgnoreGreen)
			return -1;
		return cGreenSlider;
	}
	
	public int getBlue() {
		if (cIgnoreBlue)
			return -1;
		return cBlueSlider;
	}
	
	public void handleMouse(MouseEvent theEvent, boolean repeating) {
		int							x;
		int							y;
		
		x = theEvent.getX();
		y = theEvent.getY();
		
		if (cPressedX < kSliderWidth * 4 + kMargin * 4.5 && cPressedY > cSliderHeight + kMargin * 2) {
			if (repeating)
				return;
			if (cPressedX < kSliderWidth + kMargin * 1.5)
				cIgnoreRed = !cIgnoreRed;
			else if (cPressedX < kSliderWidth * 2 + kMargin * 2.5)
				cIgnoreGreen = !cIgnoreGreen;
			else if (cPressedX < kSliderWidth * 3 + kMargin * 3.5)
				cIgnoreBlue = !cIgnoreBlue;
			else if (cPressedX < kSliderWidth * 4 + kMargin * 4.5)
				cIgnoreAlpha = !cIgnoreAlpha;
			repaint();
			return;
		}
		
		if (cPressedX < kSliderWidth + kMargin * 1.5) {
			// red
			int						newRed;
			
			newRed = 255 - 255 * (y - kMargin) / cSliderHeight;
			
			if (newRed < 0)
				newRed = 0;
			if (newRed > 255)
				newRed = 255;
			
			cRedSlider = newRed;
			
			repaint();
		}
		else if (cPressedX < kSliderWidth * 2 + kMargin * 2.5) {
			// green
			int						newGreen;
			
			newGreen = 255 - 255 * (y - kMargin) / cSliderHeight;
			
			if (newGreen < 0)
				newGreen = 0;
			if (newGreen > 255)
				newGreen = 255;
			
			cGreenSlider = newGreen;
			
			repaint();
		}
		else if (cPressedX < kSliderWidth * 3 + kMargin * 3.5) {
			// blue
			int						newBlue;
			
			newBlue = 255 - 255 * (y - kMargin) / cSliderHeight;
			
			if (newBlue < 0)
				newBlue = 0;
			if (newBlue > 255)
				newBlue = 255;
			
			cBlueSlider = newBlue;
			
			repaint();
		}
		else if (cPressedX < kSliderWidth * 4 + kMargin * 4.5) {
			// alpha
			int						newAlpha;
			
			newAlpha = 255 - 255 * (y - kMargin) / cSliderHeight;
			
			if (newAlpha < 0)
				newAlpha = 0;
			if (newAlpha > 255)
				newAlpha = 255;
			
			cAlphaSlider = newAlpha;
			
			repaint();
		} else {
			// image
			x = (x - cImageLeft) / cScaleFactor;
			y = (y - kMargin) / cScaleFactor;
			
			if (theEvent.isAltDown()) {
				// color picker
				if (x < 0 || x >= cImage.getWidth() || y < 0 || y >= cImage.getHeight())
					return;
				
				if (!cIgnoreRed)
					cRedSlider = (cImage.getPixel(y * cImage.getWidth() + x) >> 16) & 0xFF;
				if (!cIgnoreGreen)
					cGreenSlider = (cImage.getPixel(y * cImage.getWidth() + x) >> 8) & 0xFF;
				if (!cIgnoreBlue)
					cBlueSlider = (cImage.getPixel(y * cImage.getWidth() + x)) & 0xFF;
				if (!cIgnoreAlpha)
					cAlphaSlider = (cImage.getPixel(y * cImage.getWidth() + x) >> 24) & 0xFF;
			}
			else if (theEvent.isShiftDown()) {
				// fill
				int					previousColor;
				int					tempColor;
				int					fillColor;
				int					changes;
				int					fillX;
				int					fillY;
				int					alpha;
				int					red;
				int					green;
				int					blue;
				
				if (x < 0 || x >= cImage.getWidth() || y < 0 || y >= cImage.getHeight())
					return;
				
				if (!repeating)
					prepareUndo();
				
				previousColor = cImage.getPixel(y * cImage.getWidth() + x);
				
				// pick an unusual color as a temporary fill color
				tempColor = 0xFE01FD02;
				
				if (cIgnoreAlpha)
					alpha = previousColor & 0xFF000000;
				else
					alpha = cAlphaSlider << 24;
				
				
				fillColor =
						(cIgnoreAlpha ? (previousColor & 0xFF000000) : (cAlphaSlider << 24)) +
						(cIgnoreRed ? (previousColor & 0x00FF0000) : (cRedSlider << 16)) +
						(cIgnoreGreen ? (previousColor & 0x0000FF00) : (cGreenSlider << 8)) +
						(cIgnoreBlue ? (previousColor & 0x000000FF) : (cBlueSlider));
				
				cImage.setPixel(y * cImage.getWidth() + x, tempColor);
				
				do {
					changes = 0;
					
					for (fillX = 0; fillX < cImage.getWidth(); fillX++)
						for (fillY = 0; fillY < cImage.getHeight(); fillY++)
							if (cImage.getPixel(fillY * cImage.getWidth() + fillX) == previousColor) {
								// check all surrounding pixels for tempColor
								if ((fillX != 0) && cImage.getPixel(fillY * cImage.getWidth() + fillX - 1) == tempColor) {
									cImage.setPixel(fillY * cImage.getWidth() + fillX, tempColor);
									changes++;
								}
								else if ((fillX != (cImage.getWidth() - 1)) && cImage.getPixel(fillY * cImage.getWidth() + fillX + 1) == tempColor) {
									cImage.setPixel(fillY * cImage.getWidth() + fillX, tempColor);
									changes++;
								}
								else if ((fillY != 0) && cImage.getPixel((fillY - 1) * cImage.getWidth() + fillX) == tempColor) {
									cImage.setPixel(fillY * cImage.getWidth() + fillX, tempColor);
									changes++;
								}
								else if ((fillY != (cImage.getHeight() - 1)) && cImage.getPixel((fillY + 1) * cImage.getWidth() + fillX) == tempColor) {
									cImage.setPixel(fillY * cImage.getWidth() + fillX, tempColor);
									changes++;
								}
							}
				} while (changes > 0);
				
				for (fillX = 0; fillX < cImage.getWidth(); fillX++)
					for (fillY = 0; fillY < cImage.getHeight(); fillY++)
						if (cImage.getPixel(fillY * cImage.getWidth() + fillX) == tempColor)
							cImage.setPixel(fillY * cImage.getWidth() + fillX, fillColor);
			} else if (theEvent.isControlDown()) {
				// fill all
				if (!repeating) {
					prepareUndo();
					
					for (int yCounter = 0; yCounter < cImage.getWidth(); yCounter++)
						for (int xCounter = 0; xCounter < cImage.getWidth(); xCounter++)
							cImage.setPixel(xCounter, yCounter, getAlpha(), getRed(), getGreen(), getBlue());
				}
			} else {
				// pen
				if (!repeating) {
					prepareUndo();
					cImage.setPixel(x, y, getAlpha(), getRed(), getGreen(), getBlue());
					cImage.moveTo(x, y);
				} else
					cImage.lineTo(x, y, getAlpha(), getRed(), getGreen(), getBlue());
			}
			
			repaint();
		}
	}
	
	public void mousePressed(MouseEvent theEvent) {
		cPressedX = theEvent.getX();
		cPressedY = theEvent.getY();
		handleMouse(theEvent, false);
	}
	
	public void mouseDragged(MouseEvent theEvent) {
		handleMouse(theEvent, true);
	}
	
	public void mouseMoved(MouseEvent theEvent) {}
	public void mouseClicked(MouseEvent theEvent) {}
	public void mouseReleased(MouseEvent theEvent) {}
	public void mouseEntered(MouseEvent theEvent) {}
	public void mouseExited(MouseEvent theEvent) {}
}

