import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;


/*
Handles the image editing menus, and contains a GEImageEditCanvas
that takes care of most editing, and of displaying the result.
*/

public class GEImageFrame extends JFrame implements ActionListener
{
	protected GEImageEditCanvas		cImageCanvas;
	protected GPixels				cImage;
	protected boolean				cIsSprite;
	
	
	public GEImageFrame(GPixels image, boolean isSprite, String title) {
		JMenuBar					menuBar;
		JPanel						contents;
		
		cImage = image;
		cIsSprite = isSprite;
		
		setTitle(title);
		setLocation(30, 30);
		setSize(400, 250);
		
		// menu bar
		menuBar = new JMenuBar();
		setJMenuBar(menuBar);
		
		if (isSprite) {
			menuBar.add(GEditor.createMenu("Edit", new Object[] {"Undo", null,
					"Copy", "Paste", "Paste as Mask", null,
					"Toggle Grid", null,
					"Change Size", "Add Margins", "Rotate CW", "Rotate CCW", "Flip Horizontally", "Flip Vertically"},
					new int[] {KeyEvent.VK_Z, 0,
					KeyEvent.VK_C, KeyEvent.VK_V, 0, 0,
					KeyEvent.VK_G, 0,
					0, KeyEvent.VK_M, KeyEvent.VK_R, KeyEvent.VK_T, KeyEvent.VK_H, KeyEvent.VK_F}, this));
		} else {
			menuBar.add(GEditor.createMenu("Edit", new Object[] {"Undo", null,
					"Copy", "Paste", "Paste as Mask", null,
					"Toggle Grid", null,
					"Rotate CW", "Rotate CCW", "Flip Horizontally", "Flip Vertically"},
					new int[] {KeyEvent.VK_Z, 0,
					KeyEvent.VK_C, KeyEvent.VK_V, 0, 0,
					KeyEvent.VK_G, 0,
					KeyEvent.VK_R, KeyEvent.VK_T, KeyEvent.VK_H, KeyEvent.VK_F}, this));
		}
		
		cImageCanvas = new GEImageEditCanvas(cImage);
		contents = new JPanel(new BorderLayout(), false);
		contents.add(cImageCanvas);
		getContentPane().add(contents);
	}
	
	public void setVisible(boolean b) {
		if (b)
			super.setVisible(b);
		else {
			GEditor.sInstance.updateImageCanvas(false);
			GELevel.updateFrames();
			((GEPixels) cImage).frameDisposed();
			super.setVisible(b);
		}
	}
	
	public void changeSize() {
		Object[]					answer;
		
		answer = GEAskDialog.ask(this, 200, "Resize Image",
			new String[] {"Width", "Height"},
			new Integer[] {new Integer(cImage.getWidth()),
				new Integer(cImage.getHeight())});
		
		if (answer != null) {
			int						width;
			int						height;
			
			width = ((Integer) answer[0]).intValue();
			height = ((Integer) answer[1]).intValue();
			
			cImageCanvas.prepareUndo();
			((GEPixels) cImage).scaleTo(width, height);
			cImageCanvas.prepareScaling();
			cImageCanvas.repaint();
		}
	}
	
	public void prepareScaling() {
		cImageCanvas.prepareScaling();
		cImageCanvas.repaint();
	}
	
	public void addMargins() {
		Object[]					answer;
		
		answer = GEAskDialog.ask(this, 200, "Add Margins", 
			new String[] {"Left", "Top", "Right", "Bottom"},
			new Integer[] {new Integer(0), new Integer(0), new Integer(0),
				new Integer(0)});
		
		if (answer != null) {
			int						left;
			int						top;
			int						right;
			int						bottom;
			
			left = ((Integer) answer[0]).intValue();
			top = ((Integer) answer[1]).intValue();
			right = ((Integer) answer[2]).intValue();
			bottom = ((Integer) answer[3]).intValue();
			
			cImageCanvas.prepareUndo();
			((GEPixels) cImage).addMargins(left, top, right, bottom);
			cImageCanvas.prepareScaling();
			cImageCanvas.repaint();
		}
	}
	
	public void actionPerformed(ActionEvent theEvent) {
		String						command;
		
		command = theEvent.getActionCommand();
		
		if (command.equals("Undo")) {
			cImageCanvas.undo();
			repaint();
		} else if (command.equals("Copy")) {
			GEPixels.setClip(cImage);
		} else if (command.equals("Paste")) {
			if (!GEPixels.hasClip()) {
				System.err.println("Clip is empty");
				return;
			}
			
			cImageCanvas.prepareUndo();
			((GEPixels) cImage).pasteTransparent(GEPixels.getClip());
			cImageCanvas.repaint();
		} else if (command.equals("Paste as Mask")) {
			if (!GEPixels.hasClip()) {
				System.err.println("Clip is empty");
				return;
			}
			
			cImageCanvas.prepareUndo();
			((GEPixels) cImage).applyMask(GEPixels.getClip());
			cImageCanvas.repaint();
		} else if (command.equals("Toggle Grid"))
			cImageCanvas.toggleGrid();
		else if (command.equals("Change Size"))
			changeSize();
		else if (command.equals("Add Margins"))
			addMargins();
		else if (command.equals("Rotate CW")) {
			if (cIsSprite || GTileForm.getTileWidth() == GTileForm.getTileHeight())
				cImageCanvas.transform(1);
			else
				System.err.println("Can't rotate non-square tiles");
		} else if (command.equals("Rotate CCW")) {
			if (cIsSprite || GTileForm.getTileWidth() == GTileForm.getTileHeight())
				cImageCanvas.transform(3);
			else
				System.err.println("Can't rotate non-square tiles");
		} else if (command.equals("Flip Horizontally"))
			cImageCanvas.transform(4);
		else if (command.equals("Flip Vertically"))
			cImageCanvas.transform(6);
	}
}

