import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/*
Subclass of GLevel, that adds methods for saving, adding, moving,
removing and change the size of levels. Also handles the labels
for GLevel properties.
*/


public class GELevel extends GLevel
{
	public static final int			kMaxThumbnailWidth = 250;
	public static final int			kMaxThumbnailHeight = 75;
	
	protected static String[][]		sLabel = null;
	
	protected GELevelFrame			cFrame;
	protected boolean				cFrameDisposed;
	
	protected GPixels				cThumbnail;
	
	protected int					cPenX;
	protected int					cPenY;
	
	
	public GELevel() {
		super();
	}
	
	public void addMargins(int left, int top, int right, int bottom) {
		int							newWidth;
		int							newHeight;
		int[][]						newArray;
		int							x;
		int							y;
		int							spriteCounter;
		
		newWidth = getWidth() + left + right;
		newHeight = getHeight() + top + bottom;
		
		if (newWidth < 1 || newHeight < 1) {
			System.err.println("addMargins canceled; w, h: " + newWidth + ", " + newHeight);
			return;
		}
		
		if (newWidth > kMaxWidth || newHeight > kMaxHeight) {
			System.err.println("addMargins canceled; w, h: " + newWidth + ", " + newHeight);
			return;
		}
		
		newArray = new int[newWidth][newHeight];
		for (y = 0; y < newHeight; y++)
			for (x = 0; x < newWidth; x++) {
				if (x - left < 0 || x - left >= getWidth() || y - top < 0 || y - top >= getHeight())
					newArray[x][y] = 0;
				else
					newArray[x][y] = cTile[x - left][y - top]; 
			}
		
		cTile = newArray;
		
		// move sprites
		for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++) {
			int nx = getSprite(spriteCounter).getLocationX() + (1024 * left * GTileForm.getTileWidth());
			int ny = getSprite(spriteCounter).getLocationY() + (1024 * top * GTileForm.getTileHeight());
			getSprite(spriteCounter).setLocation(nx, ny);
		}
		
		removeInvalidSprites();
	}
	
	public void createThumbnail() {
		int							width;
		int							height;
		
		cThumbnail = GFactory.createPixels();
		width = getWidth();
		height = getHeight();
		
		if (width > kMaxThumbnailWidth)
			width = kMaxThumbnailWidth;
		if (height > kMaxThumbnailHeight)
			height = kMaxThumbnailHeight;
		
		cThumbnail.setSize(width, height);
		
		for (int x = 0; x < width; x++)
			for (int y = 0; y < height; y++)
				cThumbnail.setPixel(x, y, ((GEPixels) GTileForm.get(getTile(x, y)).getImage(0)).getAverage());
	}
	
	public GPixels getThumbnail() {
		if (cThumbnail == null)
			createThumbnail();
		
		return cThumbnail;
	}
	
	public void openFrame(String title) {
		if (cFrame == null) {
			cFrame = new GELevelFrame(this, title);
			cFrame.show();
			cFrameDisposed = false;
		} else {
			cFrame.setTitle(title);
			cFrame.show();
			cFrameDisposed = false;
		}
	}
	
	public void renameFrame(String title) {
		if (cFrame != null)
			cFrame.setTitle(title);
	}
	
	public void closeFrame() {
		if (cFrame != null && !cFrameDisposed) {
			cFrame.dispose();
			cFrame = null;
			cFrameDisposed = true;
		}
	}
	
	public GELevelFrame getFrame() {
		return cFrame;
	}
	
	public void frameDisposed() {
		cFrameDisposed = true;
	}
	
	public static void updateFrames() {
		for (int levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			if (GELevel.getE(levelCounter).getFrame() != null) {
				GELevel.getE(levelCounter).getFrame().setTitle("Level " + levelCounter);
				GELevel.getE(levelCounter).getFrame().refresh();
			}
	}
	
	public void writeToFile(GEBitWriter output) throws IOException {
		int[]						tileMapOneDimension;
		int							x;
		int							y;
		int							tileIndex;
		
		if (output == null)
			throw new IOException("GELevel.writeToFile: invalid file");
		
		output.pushInt(8, 1); // version
		
		((GEProperties) cProperties).writeToFile(output);
		
		output.pushInt(12, getWidth());
		output.pushInt(12, getHeight());
		
		tileMapOneDimension = new int[getWidth() * getHeight()];
		tileIndex = 0;
		for (y = 0; y < getHeight(); y++)
			for (x = 0; x < getWidth(); x++)
				tileMapOneDimension[tileIndex++] = cTile[x][y];
		
		output.compress(tileMapOneDimension, 16);
		
		output.pushInt(16, cStartScrollX);
		output.pushInt(16, cStartScrollY);
		
		output.pushInt(12, getSpriteCount());
		for (int spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++) {
			output.pushInt(8, cSprite[spriteCounter].getFormIndex());
			((GESprite) cSprite[spriteCounter]).writeToFile(output);
		}
	}
	
	public static void writeAllToFile(GEBitWriter output) throws IOException {
		int							levelCounter;
		int							last;
		
		if (output == null)
			throw new IOException("GELevel.readAllFromFile: invalid file");
		
		GOutput.log(" level version...               ");
		last = output.getBitsWritten();
		output.pushInt(8, 1); // version
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  level property count...       ");
		last = output.getBitsWritten();
		output.pushInt(8, GLevel.getStaticNumberCount());
		output.pushInt(8, GLevel.getStaticFlagCount());
		output.pushInt(8, GLevel.getStaticTextCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  level count...                ");
		last = output.getBitsWritten();
		output.pushInt(8, getCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		for (levelCounter = 0; levelCounter < getCount(); levelCounter++) {
			GOutput.log("  level " + levelCounter + "...                    ");
			last = output.getBitsWritten();
			((GELevel) sLevel[levelCounter]).writeToFile(output);
			GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		}
	}
	
	public static void writeAllLabelsToFile(GEBitWriter output) throws IOException {
		int							labelCounter;
		
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			output.pushString(8, sLabel[0][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			output.pushString(8, sLabel[1][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			output.pushString(8, sLabel[2][labelCounter]);
	}
	
	public void moveTo(int x, int y) {
		cPenX = x;
		cPenY = y;
	}
	
	public int getPenX() {
		return cPenX;
	}
	
	public int getPenY() {
		return cPenY;
	}
	
	public void lineTo(int x, int y, int tile) {
		int							x1;
		int							y1;
		int							x2;
		int							y2;
		int							curX;
		int							curY;
		
		if (x == cPenX && y == cPenY)
			return;
		
		if (Math.abs(x - cPenX) >= Math.abs(y - cPenY)) {
			// mostly horizontal
			
			if (x < cPenX) {
				x1 = x;
				y1 = y;
				x2 = cPenX;
				y2 = cPenY;
			} else {
				x1 = cPenX;
				y1 = cPenY;
				x2 = x;
				y2 = y;
			}
			
			for (curX = 0; curX <= x2 - x1; curX++) {
				curY = Math.round((float) (y2 - y1) * curX / (x2 - x1));
				setTile(x1 + curX, y1 + curY, tile);
			}
		} else {
			// mostly vertical
			
			if (y < cPenY) {
				y1 = y;
				x1 = x;
				y2 = cPenY;
				x2 = cPenX;
			} else {
				y1 = cPenY;
				x1 = cPenX;
				y2 = y;
				x2 = x;
			}
			
			for (curY = 0; curY <= y2 - y1; curY++) {
				curX = Math.round((float) (x2 - x1) * curY / (y2 - y1));
				setTile(x1 + curX, y1 + curY, tile);
			}
		}
		
		// prepare for next line
		moveTo(x, y);
	}
	
	public static void move(int fromIndex, int toIndex) {
		if (fromIndex < toIndex) {
			duplicate(toIndex);
			switchLevels(fromIndex, toIndex + 1);
			remove(fromIndex);
		} else if (toIndex < fromIndex) {
			duplicate(toIndex);
			switchLevels(toIndex + 1, toIndex);
			switchLevels(fromIndex + 1, toIndex);
			remove(fromIndex + 1);
		}
	}
	
	public static GELevel getE(int index) {
		return (GELevel) GLevel.get(index);
	}
	
	// adds the new level last
	public static void add() {
		GLevel[]					old;
		int							levelCounter;
		
		old = sLevel;
		sLevel = new GLevel[old.length + 1];
		for (levelCounter = 0; levelCounter < old.length; levelCounter++)
			sLevel[levelCounter] = old[levelCounter];
		sLevel[old.length] = GFactory.createLevel();
	}
	
	// adds the new level after the copied level
	public static void duplicate(int index) {
		GLevel[]					old;
		int							levelCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		old = sLevel;
		sLevel = new GLevel[old.length + 1];
		for (levelCounter = 0; levelCounter <= index; levelCounter++)
			sLevel[levelCounter] = old[levelCounter];
		for (levelCounter = index; levelCounter < old.length; levelCounter++)
			sLevel[levelCounter + 1] = old[levelCounter];
		
		sLevel[index + 1] = GFactory.createLevel(sLevel[index]);
	}
	
	public static void remove(int index) {
		GLevel[]					old;
		int							levelCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		if (getCount() == 1) {
			GOutput.logln("GELevel:remove(): Operation canceled. This object needs one or more levels.");
			return;
		}
		
		old = sLevel;
		sLevel = new GLevel[old.length - 1];
		if (index > 0)
			for (levelCounter = 0; levelCounter < index; levelCounter++)
				sLevel[levelCounter] = old[levelCounter];
		if (index < sLevel.length)
			for (levelCounter = index; levelCounter < sLevel.length; levelCounter++)
				sLevel[levelCounter] = old[levelCounter + 1];
	}
	
	public static void switchLevels(int index1, int index2) {
		GLevel						temp;
		
		if (index1 == index2)
			return;
		
		if (index1 < 0 || index1 >= getCount()) {
			GOutput.throwOB("index1", index1, 0, getCount() - 1);
			return;
		} else if (index2 < 0 || index2 >= getCount()) {
			GOutput.throwOB("index2", index2, 0, getCount() - 1);
			return;
		}
		
		temp = sLevel[index1];
		sLevel[index1] = sLevel[index2];
		sLevel[index2] = temp;
	}
	
	// inserts a slot after index
	public void insertTileFormReference(int index) {
		int							x;
		int							y;
		
		for (x = 0; x < getWidth(); x++)
			for (y = 0; y < getHeight(); y++)
				if (cTile[x][y] > index)
					cTile[x][y]++;
	}
	
	public void removeTileFormReference(int index) {
		int							x;
		int							y;
		
		for (x = 0; x < getWidth(); x++)
			for (y = 0; y < getHeight(); y++) {
				if (cTile[x][y] == index)
					cTile[x][y] = 0;
				else if (cTile[x][y] > index)
					cTile[x][y]--;
			}
	}
	
	public void switchTileFormReferences(int index1, int index2) {
		int							x;
		int							y;
		
		for (x = 0; x < getWidth(); x++)
			for (y = 0; y < getHeight(); y++) {
				if (cTile[x][y] == index1)
					cTile[x][y] = index2;
				else if (cTile[x][y] == index2)
					cTile[x][y] = index1;
			}
	}
	
	// inserts a slot after index
	public void insertSpriteFormReference(int index) {
		int							spriteCounter;
		
		for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++)
			if (cSprite[spriteCounter].getFormIndex() > index)
				cSprite[spriteCounter].setFormIndex(cSprite[spriteCounter].getFormIndex() + 1);
	}
	
	public void removeSpriteFormReference(int index) {
		// this will remove any sprites with the removed index
		int							spriteCounter;
		
		spriteCounter = 0;
		while (spriteCounter < getSpriteCount()) {
			if (cSprite[spriteCounter].getFormIndex() == index) {
				removeSprite(spriteCounter);
			} else if (cSprite[spriteCounter].getFormIndex() > index) {
				cSprite[spriteCounter].setFormIndex(cSprite[spriteCounter].getFormIndex() - 1);
				spriteCounter++;
			} else {
				spriteCounter++;
			}
		}
	}
	
	public void switchSpriteFormReferences(int index1, int index2) {
		int							spriteCounter;
		
		for (spriteCounter = 0; spriteCounter < getSpriteCount(); spriteCounter++) {
			if (cSprite[spriteCounter].getFormIndex() == index1)
				cSprite[spriteCounter].setFormIndex(index2);
			else if (cSprite[spriteCounter].getFormIndex() == index2)
				cSprite[spriteCounter].setFormIndex(index1);
		}
	}
	
	public static void readAllLabelsFromFile(GBitReader input) throws IOException {
		int							labelCounter;
		
		sLabel = new String[3][];
		
		sLabel[0] = new String[getStaticNumberCount()];
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			sLabel[0][labelCounter] = input.pullString();
		
		sLabel[1] = new String[getStaticFlagCount()];
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			sLabel[1][labelCounter] = input.pullString();
		
		sLabel[2] = new String[getStaticTextCount()];
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			sLabel[2][labelCounter] = input.pullString();
	}
	
	public static void clearLabels() {
		sLabel = new String[3][0];
		setLabelCount(0, getStaticNumberCount());
		setLabelCount(1, getStaticFlagCount());
		setLabelCount(2, getStaticTextCount());
	}
	
	public static String getLabel(int type, int index) {
		if (type < 0 || type > 2 || index < 0 || index >= sLabel[type].length)
			return "-";
		
		return sLabel[type][index];
	}
	
	public static String[] getLabels(int type) {
		if (type < 0 || type > 2)
			return new String[] {"incorrect type"};
		
		if (sLabel[type] == null)
			return new String[0];
		return sLabel[type];
	}
	
	public static void setLabel(int type, int index, String label) {
		if (type == GProperties.kNumber && index >= 0 && index < getStaticNumberCount())
			sLabel[type][index] = label;
		
		if (type == GProperties.kFlag && index >= 0 && index < getStaticFlagCount())
			sLabel[type][index] = label;
		
		if (type == GProperties.kText && index >= 0 && index < getStaticTextCount())
			sLabel[type][index] = label;
	}
	
	protected static void setLabelCount(int number, int flag, int text) {
		setLabelCount(0, number);
		setLabelCount(1, flag);
		setLabelCount(2, text);
	}
	
	protected static void setLabelCount(int type, int count) {
		if (type < 0 || type > 2 || count < 0 || sLabel == null)
			return;
		
		if (count <= 0)
			sLabel[type] = new String[0];
		else {
			String[]				oldLabel;
			int						labelCounter;
			
			oldLabel = sLabel[type];
			sLabel[type] = new String[count];
			for (labelCounter = 0; labelCounter < count; labelCounter++) {
				if (oldLabel != null && labelCounter < oldLabel.length)
					sLabel[type][labelCounter] = oldLabel[labelCounter];
				else
					sLabel[type][labelCounter] = "";
			}
		}
	}
}

