import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;


/*
Wraps a GLevelCanvas in a JComponent, and handles scrolling with
scroll bars. It also handles mouse actions for the level canvas.
Note that GELevelCanvas is not a subclass of GLevelCanvas.
*/


public class GELevelCanvas extends JComponent
		implements MouseListener, MouseMotionListener, AdjustmentListener
{
	protected GLevel				cLevel;
	protected GLevelCanvas			cCanvas;
	protected GEImageCanvas			cImageCanvas;
	protected JScrollBar			cHorizontalScrollBar;
	protected JScrollBar			cVerticalScrollBar;
	protected JFrame				cOwner;
	
	
	public GELevelCanvas(GLevel level, GEImageCanvas imageCanvas, 
			JScrollBar horizontalScrollBar, JScrollBar verticalScrollBar, JFrame owner) {
		cLevel = level;
		cImageCanvas = imageCanvas;
		cHorizontalScrollBar = horizontalScrollBar;
		cVerticalScrollBar = verticalScrollBar;
		cOwner = owner;
		
		cCanvas = new GLevelCanvas(this, cLevel, -1, -1);
		
		if (cVerticalScrollBar != null) {
			cVerticalScrollBar.addAdjustmentListener(this);
			cVerticalScrollBar.setUnitIncrement(GTileForm.getTileHeight());
		}
		
		if (cHorizontalScrollBar != null) {
			cHorizontalScrollBar.addAdjustmentListener(this);
			cHorizontalScrollBar.setUnitIncrement(GTileForm.getTileWidth());
		}
		
		updateScrollBars();
		
		addMouseListener(this);
		addMouseMotionListener(this);
	}
	
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		
		cCanvas.setSize(width, height);
		updateScrollBars();
		repaint();
	}
	
	public void refresh() {
		cCanvas.setSize(getBounds().width, getBounds().height);
		cCanvas.setNeedUpdate(true);
		updateScrollBars();
		repaint();
	}
	
	public GLevelCanvas getLevelCanvas() {
		return cCanvas;
	}
	
	public void updateScrollBars() {
		int							totalWidth;
		int							totalHeight;
		
		if (cHorizontalScrollBar != null) {
			int						newValue;
			
			// horizontal scroll bar
			totalWidth = cLevel.getWidth() * GTileForm.getTileWidth();
			newValue = cCanvas.getScrollX();
			
			if (newValue + GTileForm.getTileWidth() * (getBounds().width / GTileForm.getTileWidth()) > totalWidth)
				newValue = totalWidth - GTileForm.getTileWidth() * (getBounds().width / GTileForm.getTileWidth());
			if (newValue < 0)
				newValue = 0;
			
			cHorizontalScrollBar.setValues(newValue, GTileForm.getTileWidth() * (getBounds().width / GTileForm.getTileWidth()), 0, totalWidth);
			cHorizontalScrollBar.setBlockIncrement(getBounds().width / 2);
		}
		
		if (cVerticalScrollBar != null) {
			int						newValue;
			
			// vertical scroll bar
			totalHeight = cLevel.getHeight() * GTileForm.getTileHeight();
			newValue = cCanvas.getScrollY();
			
			if (newValue + GTileForm.getTileHeight() * (getBounds().height / GTileForm.getTileHeight()) > totalHeight)
				newValue = totalHeight - GTileForm.getTileHeight() * (getBounds().height / GTileForm.getTileHeight());
			if (newValue < 0)
				newValue = 0;
			
			cVerticalScrollBar.setValues(newValue, GTileForm.getTileHeight() * (getBounds().height / GTileForm.getTileHeight()), 0, totalHeight);
			cVerticalScrollBar.setBlockIncrement(getBounds().height / 2);
		}
	}
	
	public synchronized void paint(Graphics theGraphics) {
		if (cCanvas != null)
			cCanvas.paint(theGraphics);
	}
	
	public void adjustmentValueChanged(AdjustmentEvent theEvent) {
		if (theEvent.getSource() == cHorizontalScrollBar)
			cCanvas.setScrollX(cHorizontalScrollBar.getValue());
		else if (theEvent.getSource() == cVerticalScrollBar)
			cCanvas.setScrollY(cVerticalScrollBar.getValue());
		
		repaint();
	}
	
	public void editSpriteProperties(int spriteIndex) {
		int							formIndex;
		int							numberCount;
		int							numberCounter;
		int							flagCount;
		int							flagCounter;
		int							textCount;
		int							textCounter;
		int							statCount;
		int							statCounter;
		
		GSprite						sprite;
		
		String[]					label;
		Object[]					answer;
		
		sprite = cLevel.getSprite(spriteIndex);
		
		numberCount = sprite.getForm().getInstanceNumberCount();
		flagCount = sprite.getForm().getInstanceFlagCount();
		textCount = sprite.getForm().getInstanceTextCount();
		statCount = 7 + numberCount + flagCount + textCount;
		
		label = new String[statCount];
		answer = new Object[statCount];
		
		statCounter = 0;
		
		// sprite-specific properties
		label[statCounter] = "Location x";
		answer[statCounter++] = new Integer(sprite.getLocationX() >> 10);
		
		label[statCounter] = "Location y";
		answer[statCounter++] = new Integer(sprite.getLocationY() >> 10);
		
		label[statCounter] = "Can pass solid tiles";
		answer[statCounter++] = new Boolean(sprite.getCanPassSolidTiles());
		
		label[statCounter] = "Collision effect";
		answer[statCounter++] = new Integer(sprite.getCollisionEffect());
		
		label[statCounter] = "Accept collision events";
		answer[statCounter++] = new Boolean(sprite.getDetectCollision());
		
		label[statCounter] = "Collision width";
		answer[statCounter++] = new Integer(sprite.getCollisionWidth());
		
		label[statCounter] = "Collision height";
		answer[statCounter++] = new Integer(sprite.getCollisionHeight());
		
		// numbers, flags and texts
		for (numberCounter = 0; numberCounter < numberCount; numberCounter++) {
			label[statCounter] = ((GESpriteForm) sprite.getForm()).getInstanceLabel(GProperties.kNumber, numberCounter);
			answer[statCounter] = new Integer(sprite.getNumber(numberCounter));
			
			statCounter++;
		}
		
		for (flagCounter = 0; flagCounter < flagCount; flagCounter++) {
			label[statCounter] = ((GESpriteForm) sprite.getForm()).getInstanceLabel(GProperties.kFlag, flagCounter);
			answer[statCounter] = new Boolean(sprite.getFlag(flagCounter));
			
			statCounter++;
		}
		
		for (textCounter = 0; textCounter < textCount; textCounter++) {
			label[statCounter] = ((GESpriteForm) sprite.getForm()).getInstanceLabel(GProperties.kText, textCounter);
			answer[statCounter] = sprite.getText(textCounter);
			
			statCounter++;
		}
		
		answer = GEAskDialog.ask(cOwner, 350, "Edit Sprite Properties", label, answer);
		
		if (answer != null) {
			statCounter = 0;
			
			// sprite-specific properties
			sprite.setLocationX(((Integer) answer[statCounter++]).intValue() << 10);
			sprite.setLocationY(((Integer) answer[statCounter++]).intValue() << 10);
			sprite.setCanPassSolidTiles(((Boolean) answer[statCounter++]).booleanValue());
			sprite.setCollisionEffect(((Integer) answer[statCounter++]).intValue());
			sprite.setDetectCollision(((Boolean) answer[statCounter++]).booleanValue());
			sprite.setCollisionWidth(((Integer) answer[statCounter++]).intValue());
			sprite.setCollisionHeight(((Integer) answer[statCounter++]).intValue());
			
			// numbers, flags and texts
			for (numberCounter = 0; numberCounter < numberCount; numberCounter++)
				sprite.getProperties().setNumber(numberCounter, ((Integer) answer[statCounter++]).intValue());
			
			for (flagCounter = 0; flagCounter < flagCount; flagCounter++)
				sprite.getProperties().setFlag(flagCounter, ((Boolean) answer[statCounter++]).booleanValue());
			
			for (textCounter = 0; textCounter < textCount; textCounter++)
				sprite.getProperties().setText(textCounter, (String) answer[statCounter++]);
		}
	}
	
	public void handleMouse(MouseEvent theEvent, boolean repeating) {
		int							tileX;
		int							tileY;
		
		tileX = (theEvent.getX() + cHorizontalScrollBar.getValue()) / GTileForm.getTileWidth();
		tileY = (theEvent.getY() + cVerticalScrollBar.getValue()) / GTileForm.getTileHeight();
		
		if (tileX < 0 || tileX >= cLevel.getWidth() ||
				tileY < 0 || tileY >= cLevel.getHeight())
			return;
		
		// see if option-pressing/shift-pressing a sprite
		if (theEvent.isAltDown() || theEvent.isShiftDown()) {
			int						x;
			int						y;
			
			x = (theEvent.getX() + cHorizontalScrollBar.getValue()) << 10;
			y = (theEvent.getY() + cVerticalScrollBar.getValue()) << 10;
			
			// edit stats of the topmost sprite covering x, y
			for (int index = cLevel.getSpriteCount() - 1; index >= 0; index--) {
				GSprite			sprite;
				
				sprite = cLevel.getSprite(index);
				if (sprite.getLeft() < x &&
						sprite.getRight() > x &&
						sprite.getTop() < y &&
						sprite.getBottom() > y) {
					
					if (theEvent.isAltDown())
						editSpriteProperties(index);
					else
						cLevel.removeSprite(index);
					
					repaint();
					return;
				}
			}
		}
		
		if (theEvent.isAltDown()) {
			// tile picker
			cImageCanvas.setSelected(cLevel.getTile(tileX, tileY));
			return;
		}
		
		if (cImageCanvas.getSelected() < GTileForm.getCount()) {
			// tile
			if (theEvent.isShiftDown()) {
				// fill
				int					previousTile;
				int					tempTile;
				int					fillTile;
				int					changes;
				
				previousTile = cLevel.getTile(tileX, tileY);
				tempTile = 0x0FF00FF0;
				fillTile = cImageCanvas.getSelected();
				
				cLevel.setTile(tileX, tileY, tempTile);
				
				do {
					changes = 0;
					
					for (int fillX = 0; fillX < cLevel.getWidth(); fillX++)
						for (int fillY = 0; fillY < cLevel.getHeight(); fillY++)
							if (cLevel.getTile(fillX, fillY) == previousTile) {
								// check all surrounding tiles for tempTile
								if ((fillX != 0) && cLevel.getTile(fillX - 1, fillY) == tempTile)
									cLevel.setTile(fillX, fillY, tempTile);
								else if ((fillX != (cLevel.getWidth() - 1)) && cLevel.getTile(fillX + 1, fillY) == tempTile)
									cLevel.setTile(fillX, fillY, tempTile);
								else if ((fillY != 0) && cLevel.getTile(fillX, fillY - 1) == tempTile)
									cLevel.setTile(fillX, fillY, tempTile);
								else if ((fillY != (cLevel.getHeight() - 1)) && cLevel.getTile(fillX, fillY + 1) == tempTile)
									cLevel.setTile(fillX, fillY, tempTile);
								
								if (cLevel.getTile(fillX, fillY) == tempTile)
									changes++;
							}
				} while (changes > 0);
				
				for (int fillX = 0; fillX < cLevel.getWidth(); fillX++)
					for (int fillY = 0; fillY < cLevel.getHeight(); fillY++)
						if (cLevel.getTile(fillX, fillY) == tempTile)
							cLevel.setTile(fillX, fillY, fillTile);
				
				cCanvas.setNeedUpdate(true);
				repaint();
			} else {
				// pen
				if (!repeating) {
					cLevel.setTile(tileX, tileY, cImageCanvas.getSelected());
					((GELevel) cLevel).moveTo(tileX, tileY);
					cCanvas.setNeedUpdate(tileX, tileY);
				} else {
					int				penX = ((GELevel) cLevel).getPenX();
					int				penY = ((GELevel) cLevel).getPenY();
					
					((GELevel) cLevel).lineTo(tileX, tileY, cImageCanvas.getSelected());
					for (int x = Math.min(tileX, penX); x <= Math.max(tileX, penX); x++)
						for (int y = Math.min(tileY, penY); y <= Math.max(tileY, penY); y++)
							cCanvas.setNeedUpdate(x, y);
				}
				repaint();
			}
		} else {
			int						x;
			int						y;
			
			x = (theEvent.getX() + cHorizontalScrollBar.getValue()) << 10;
			y = (theEvent.getY() + cVerticalScrollBar.getValue()) << 10;
			
			cLevel.addSprite(x, y, cImageCanvas.getSelected() - GTileForm.getCount());
			repaint();
		}
	}
	
	public void mousePressed(MouseEvent theEvent) {
		handleMouse(theEvent, false);
	}
	
	public void mouseDragged(MouseEvent theEvent) {
		if (cImageCanvas.getSelected() < GTileForm.getCount() && !theEvent.isAltDown() && !theEvent.isShiftDown())
			handleMouse(theEvent, true);
	}
	
	public void mouseMoved(MouseEvent theEvent) {}
	public void mouseClicked(MouseEvent theEvent) {}
	public void mouseReleased(MouseEvent theEvent) {}
	public void mouseEntered(MouseEvent theEvent) {}
	public void mouseExited(MouseEvent theEvent) {}
}

