import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;


/*
Subclass of GPixel, that adds methods for saving, transforming,
drawing to, scaling and importing images. GEPixels also implements
a clipboard, and owns the image editing frames.
*/


public class GEPixels extends GPixels implements ImageConsumer
{
	protected static GPixels		sClip;
	
	protected int					cPenX;
	protected int					cPenY;
	
	protected GEImageFrame			cFrame;
	protected boolean				cFrameDisposed;
	
	protected boolean				cImportCompleted;
	
	
	public GEPixels() {
		super();
	}
	
	public GEPixels(Image image) {
		super();
		
		cImportCompleted = false;
		image.getSource().startProduction(this);
		
		cAlphaMod = 255;
	}
	
	public void copyFrom(GEPixels other) {
		cWidth = other.cWidth;
		cHeight = other.cHeight;
		cPixelArray = new int[other.cPixelArray.length];
		System.arraycopy(other.cPixelArray, 0, cPixelArray, 0, cPixelArray.length);
		cAlphaMod = other.cAlphaMod;
	}
	
	public void clear() {
		for (int pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++)
			cPixelArray[pixelIndex] = 0xFF9F9F9F;
	}
	
	protected void copyPixels(int[] canvas, int canvasWidth, int sourceStart, int destStart,
			int plotWidth, int plotHeight, int deltaX, int deltaY) {
		// make sure we do not need to keep track of whether this has a mask
		copyPixelsAlpha(canvas, canvasWidth, sourceStart, destStart, plotWidth, plotHeight, deltaX, deltaY);
	}
	
	public void moveTo(int x, int y) {
		cPenX = x;
		cPenY = y;
	}
	
	public void lineTo(int x, int y, int alpha, int red, int green, int blue) {
		int							x1;
		int							y1;
		int							x2;
		int							y2;
		int							curX;
		int							curY;
		
		if (x == cPenX && y == cPenY)
			return;
		
		if (Math.abs(x - cPenX) >= Math.abs(y - cPenY)) {
			// mostly horizontal
			
			if (x < cPenX) {
				x1 = x;
				y1 = y;
				x2 = cPenX;
				y2 = cPenY;
			} else {
				x1 = cPenX;
				y1 = cPenY;
				x2 = x;
				y2 = y;
			}
			
			for (curX = 0; curX <= x2 - x1; curX++) {
				curY = Math.round((float) (y2 - y1) * curX / (x2 - x1));
				setPixel(x1 + curX, y1 + curY, alpha, red, green, blue);
			}
		} else {
			// mostly vertical
			
			if (y < cPenY) {
				y1 = y;
				x1 = x;
				y2 = cPenY;
				x2 = cPenX;
			} else {
				y1 = cPenY;
				x1 = cPenX;
				y2 = y;
				x2 = x;
			}
			
			for (curY = 0; curY <= y2 - y1; curY++) {
				curX = Math.round((float) (x2 - x1) * curY / (y2 - y1));
				setPixel(x1 + curX, y1 + curY, alpha, red, green, blue);
			}
		}
		
		// prepare for next line
		moveTo(x, y);
	}
	
	public void setPixel(int x, int y, int value) {
		if (x >= 0 && y >= 0 && x < cWidth && y < cHeight)
			cPixelArray[x + y * cWidth] = value;
	}
	
	public void setPixel(int x, int y, int inAlpha, int inRed, int inGreen, int inBlue) {
		if (x >= 0 && y >= 0 && x < cWidth && y < cHeight) {
			int						alpha;
			int						red;
			int						green;
			int						blue;
			
			if (inAlpha == -1)
				alpha = cPixelArray[x + y * cWidth] & 0xFF000000;
			else
				alpha = inAlpha << 24;
			
			if (inRed == -1)
				red = cPixelArray[x + y * cWidth] & 0x00FF0000;
			else
				red = inRed << 16;
			
			if (inGreen == -1)
				green = cPixelArray[x + y * cWidth] & 0x0000FF00;
			else
				green = inGreen << 8;
			
			if (inBlue == -1)
				blue = cPixelArray[x + y * cWidth] & 0x000000FF;
			else
				blue = inBlue;
			
			cPixelArray[x + y * cWidth] = alpha + red + green + blue;
		}
	}
	
	// preserves the alpha information, used to rotate/mirror images
	protected void drawVisiblePreserve(int[] canvas, int canvasWidth,
			int localLeft, int localTop, int localRight, int localBottom,
			int leftInset, int topInset, int rightInset, int bottomInset,
			int deltaX, int deltaY, int mode) {
		// deltaX and deltaY refers to sourceIndex's modifications
		// destinationIndex will use 1 and canvasWidth respectively
		
		int							canvasHeight = canvas.length / canvasWidth;
		
		int							plotWidth;
		int							plotHeight;
		
		int							sourceStartX;
		int							sourceStartY;
		int							sourceStart;
		int							destStart;
		
		// clip
		if (localLeft < 0) {
			leftInset -= localLeft;
			localLeft = 0;
		}
		if (localTop < 0) {
			topInset -= localTop;
			localTop = 0;
		}
		if (localRight > canvasWidth) {
			rightInset += localRight - canvasWidth;
			localRight = canvasWidth;
		}
		if (localBottom > canvasHeight) {
			bottomInset += localBottom - canvasHeight;
			localBottom = canvasHeight;
		}
		
		plotWidth = localRight - localLeft;
		plotHeight = localBottom - localTop;
		
		// make sure size is still positive
		if (plotWidth <= 0 || plotHeight <= 0) {
			// clipped into non-existence
			return;
		}
		
		// calculate where source pixels index should start
		if (deltaX > 0)
			sourceStartX = leftInset;
		else
			sourceStartX = -(getWidth(mode) - leftInset - 1);
		
		if (deltaY > 0)
			sourceStartY = topInset;
		else
			sourceStartY = -(getHeight(mode) - topInset - 1);
		
		sourceStart = sourceStartX * deltaX + sourceStartY * deltaY;
		
		// destination pixel index always starts at top left
		destStart = localLeft + localTop * canvasWidth;
		
		// preserve the mask when applying transformations
		copyPixelsNoAlpha(canvas, canvasWidth, sourceStart, destStart,
				plotWidth, plotHeight, deltaX, deltaY);
	}
	
	// preserves the alpha information, used to rotate/mirror images
	public void drawPreserve(int[] canvas, int canvasWidth, int left, int top, int mode) {
		int							canvasHeight = canvas.length / canvasWidth;
		
		int							deltaX;
		int							deltaY;
		
		if (mode == 0) { deltaX = 1; deltaY = cWidth;
		} else if (mode == 1) { deltaX = -cWidth; deltaY = 1;
		} else if (mode == 2) { deltaX = -1; deltaY = -cWidth;
		} else if (mode == 3) { deltaX = cWidth; deltaY = -1;
		} else if (mode == 4) { deltaX = -1; deltaY = cWidth;
		} else if (mode == 5) { deltaX = cWidth; deltaY = 1;
		} else if (mode == 6) { deltaX = 1; deltaY = -cWidth;
		} else if (mode == 7) { deltaX = -cWidth; deltaY = -1;
		} else return;
		
		drawVisiblePreserve(canvas, canvasWidth, left, top, left + getWidth(mode), top + getHeight(mode),
				0, 0, 0, 0, deltaX, deltaY, mode);
	}
	
	/*
	takes the blue channel of image and uses it as an alpha channel
	for this image
	*/
	public void applyMask(GPixels image) {
		int							color;
		int							alpha;
		
		if (image == null) {
			System.err.println("No mask to apply");
			return;
		}
		
		if ((image.getWidth() != getWidth()) ||
				(image.getHeight() != getHeight())) {
			System.err.println("Invalid image dimensions (" + image.getWidth() + " * " + image.getHeight() + ")");
			return;
		}
		
		for (int pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++) {
			color = cPixelArray[pixelIndex] & 0x00FFFFFF;
			alpha = image.getPixel(pixelIndex) & 0x000000FF; // use blue channel
			
			// invert alpha channel
			alpha = 255 - alpha;
			
			cPixelArray[pixelIndex] = color + (alpha << 24);
		}
	}
	
	public void pasteTransparent(GPixels image) {
		if (image == null)
			System.err.println("Empty clipboard");
		else
			((GEPixels) image).draw(cPixelArray, cWidth, 0, 0, 0);
	}
	
	public void pasteSolid(GPixels image) {
		if (image == null)
			System.err.println("Empty clipboard");
		else if (image.getWidth() != cWidth || image.getHeight() != cHeight)
			System.err.println("Invalid image dimensions (" + image.getWidth() + " * " + image.getHeight() + ")");
		else
			copyFrom((GEPixels) image);
	}
	
	// implemented ImageConsumer methods:
	public void setDimensions(int width, int height) {
		cWidth = width;
		cHeight = height;
		cPixelArray = new int[cWidth * cHeight];
	}
	
	public void setPixels(int x, int y, int w, int h,
				ColorModel model, byte[] pixels, int off, int scansize) {
		try {
			for (int yCounter = y; yCounter < y + h; yCounter++)
				for (int xCounter = x; xCounter < x + w; xCounter++)
					cPixelArray[yCounter * cWidth + xCounter] = model.getRGB(pixels[yCounter * scansize + xCounter + off]);
		} catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("SetPixels AIOOB exception, arguments: " + x + ", " + y + ", " + w + ", " + h + ", " + off + ", " + scansize);
		}
	}
	
	public void setPixels(int x, int y, int w, int h,
			ColorModel model, int pixels[], int off, int scansize) {
		try {
			for (int yCounter = y; yCounter < y + h; yCounter++)
				for (int xCounter = x; xCounter < x + w; xCounter++)
					cPixelArray[yCounter * cWidth + xCounter] = model.getRGB(pixels[yCounter * scansize + xCounter + off]);
		} catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("SetPixels AIOOB exception, arguments: " + x + ", " + y + ", " + w + ", " + h + ", " + off + ", " + scansize);
		}
	}
	
	public void imageComplete(int status) {
		if (!cImportCompleted) {
			System.out.println("Image complete, status = " + status + "; (" + (new String[] {"error", "frame done", "image done", "aborted"})[status - 1] + ")");
			GEditor.sInstance.updateImageCanvas(false);
		}
		cImportCompleted = true;
	}
	
	// ignored ImageConsumer methods:
	public void setProperties(Hashtable props) {}
	public void setColorModel(ColorModel model) {}
	public void setHints(int hintflags) {}
	
	// takes some random pixels and returns their average
	// alpha is ignored
	// used to create level thumbnails.
	public int getAverage() {
		int							sampleCount = 16;
		int							pixelCounter;
		int							red = 0;
		int							green = 0;
		int							blue = 0;
		int							index = 0;
		
		for (pixelCounter = 0; pixelCounter < sampleCount; pixelCounter++) {
			index = (int) (Math.random() * cPixelArray.length);
			red += (cPixelArray[index] >> 16) & 0x000000FF;
			green += (cPixelArray[index] >> 8) & 0x000000FF;
			blue += (cPixelArray[index]) & 0x000000FF;
		}
		
		red /= sampleCount;
		green /= sampleCount;
		blue /= sampleCount;
		
		return 0xFF000000 | red << 16 | green << 8 | blue;
	}
	
	public void scaleTo(int width, int height) {
		int[]						newArray;
		int							x;
		int							y;
		
		if (width < 1 || width > kMaxSize) {
			GOutput.throwOB("width", width, 1, kMaxSize);
			return;
		}
		
		if (height < 1 || height > kMaxSize) {
			GOutput.throwOB("height", height, 1, kMaxSize);
			return;
		}
		
		newArray = new int[width * height];
		for (y = 0; y < height; y++)
			for (x = 0; x < width; x++) {
				int					oldX;
				int					oldY;
				
				oldX = x * cWidth / width;
				oldY = y * cHeight / height;
				
				newArray[x + y * width] = cPixelArray[oldX + oldY * cWidth];
			}
		
		cWidth = width;
		cHeight = height;
		cPixelArray = newArray;
	}
	
	public void addMargins(int left, int top, int right, int bottom) {
		int							newWidth;
		int							newHeight;
		int[]						newArray;
		int							x;
		int							y;
		
		newWidth = cWidth + left + right;
		newHeight = cHeight + top + bottom;
		
		if (newWidth < 1 || newHeight < 1) {
			System.err.println("addMargins canceled; w, h: " + newWidth + ", " + newHeight);
			return;
		}
		
		newArray = new int[newWidth * newHeight];
		for (y = 0; y < newHeight; y++)
			for (x = 0; x < newWidth; x++) {
				if (x - left < 0 || x - left >= cWidth || y - top < 0 || y - top >= cHeight)
					newArray[x + y * newWidth] = 0xFF000000;
				else
					newArray[x + y * newWidth] = cPixelArray[x - left + (y - top) * cWidth]; 
			}
		
		cWidth = newWidth;
		cHeight = newHeight;
		cPixelArray = newArray;
	}
	
	public void transform(int mode) {
		int[]						newArray;
		
		newArray = new int[cWidth * cHeight];
		
		drawPreserve(newArray, getWidth(mode), 0, 0, mode);
		
		cPixelArray = newArray;
		
		if (mode % 2 == 1 && cWidth != cHeight) {
			int						newWidth;
			
			newWidth = cHeight;
			cHeight = cWidth;
			cWidth = newWidth;
		}
	}
	
	public void openFrame(String title, boolean isSprite) {
		if (cFrame == null) {
			cFrame = new GEImageFrame(this, isSprite, title);
			cFrame.show();
			cFrameDisposed = false;
		} else {
			cFrame.setTitle(title);
			cFrame.show();
			cFrameDisposed = false;
		}
	}
	
	public void renameFrame(String title) {
		if (cFrame != null)
			cFrame.setTitle(title);
	}
	
	public void closeFrame() {
		if (cFrame != null && !cFrameDisposed) {
			cFrame.dispose();
			cFrame = null;
			cFrameDisposed = true;
		}
	}
	
	public GEImageFrame getFrame() {
		return cFrame;
	}
	
	public void frameDisposed() {
		cFrameDisposed = true;
	}
	
	public void writeToFile(GEBitWriter output) throws IOException {
		int							pixelIndex;
		boolean						hasMask;
		
		if (output == null)
			throw new IOException("GEPixels.writeToFile: invalid file");
		
		output.pushInt(8, 1);
		
		output.pushInt(10, cWidth);
		output.pushInt(10, cHeight);
		
		// evaluate if mask is used, else save 24-bit color info
		// does not use cHasMask because the status might have changed
		hasMask = false;
		
		for (pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++)
			if ((cPixelArray[pixelIndex] & 0xFF000000) != 0xFF000000)
				hasMask = true;
		
		if (hasMask) {
			// reduce the number of colors by setting all transparent
			// pixels to 0x00000000.
			for (pixelIndex = 0; pixelIndex < cPixelArray.length; pixelIndex++)
				if ((cPixelArray[pixelIndex] & 0xFF000000) == 0x00000000)
					cPixelArray[pixelIndex] = 0x00000000;
		}
		
		output.pushBoolean(hasMask);
		
		if (hasMask)
			output.compress(cPixelArray, 32);
		else
			output.compress(cPixelArray, 24);
	}
	
	public static boolean hasClip() {
		return (sClip != null);
	}
	
	public static GPixels getClip() {
		return sClip;
	}
	
	public static void setClip(GPixels clip) {
		if (sClip == null)
			sClip = (GPixels) clip.clone();
		else
			((GEPixels) sClip).copyFrom((GEPixels) clip);
	}
}

