import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/*
Subclass of GSpriteForm that adds methods for saving, moving,
adding and removing sprite forms. It also handles the labels for
GSpriteForms and GSprites.
*/


public class GESpriteForm extends GSpriteForm
{
	protected static String[][]		sLabel = null;
	protected String[][]			cInstanceLabel = null;
	
	
	public void readLabelsFromFile(GBitReader input) throws IOException {
		int							labelCounter;
		
		cInstanceLabel = new String[3][];
		
		cInstanceLabel[0] = new String[getInstanceNumberCount()];
		for (labelCounter = 0; labelCounter < getInstanceNumberCount(); labelCounter++)
			cInstanceLabel[0][labelCounter] = input.pullString();
		
		cInstanceLabel[1] = new String[getInstanceFlagCount()];
		for (labelCounter = 0; labelCounter < getInstanceFlagCount(); labelCounter++)
			cInstanceLabel[1][labelCounter] = input.pullString();
		
		cInstanceLabel[2] = new String[getInstanceTextCount()];
		for (labelCounter = 0; labelCounter < getInstanceTextCount(); labelCounter++)
			cInstanceLabel[2][labelCounter] = input.pullString();
	}
	
	public void clear() {
		super.clear();
		
		clearInstanceLabels();
	}
	
	public void clearInstanceLabels() {
		cInstanceLabel = new String[3][];
		setInstanceLabelCount(0, getInstanceNumberCount());
		setInstanceLabelCount(1, getInstanceFlagCount());
		setInstanceLabelCount(2, getInstanceTextCount());
	}
	
	public static void clearAllInstanceLabels() {
		for (int index = 0; index < getCount(); index++)
			((GESpriteForm) get(index)).clearInstanceLabels();
	}
	
	public void moveImage(int fromIndex, int toIndex) {
		if (fromIndex < toIndex) {
			duplicateImage(toIndex);
			switchImages(fromIndex, toIndex + 1);
			removeImage(fromIndex);
		} else if (toIndex < fromIndex) {
			duplicateImage(toIndex);
			switchImages(fromIndex + 1, toIndex);
			removeImage(fromIndex + 1);
		}
	}
	
	public void addImage() {
		GPixels[]					old;
		int							imageCounter;
		
		old = cImage;
		cImage = new GPixels[old.length + 1];
		for (imageCounter = 0; imageCounter < old.length; imageCounter++)
			cImage[imageCounter] = old[imageCounter];
		cImage[old.length] = GFactory.createPixels();
	}
	
	public void addImage(GPixels image) {
		if (image == null)
			return;
		
		addImage();
		cImage[cImage.length - 1] = image;
	}
	
	public void duplicateImage(int index) {
		GPixels[]					old;
		int							imageCounter;
		
		if (index < 0 || index >= getImageCount()) {
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
			return;
		}
		
		old = cImage;
		cImage = new GPixels[old.length + 1];
		for (imageCounter = 0; imageCounter <= index; imageCounter++)
			cImage[imageCounter] = old[imageCounter];
		for (imageCounter = index; imageCounter < old.length; imageCounter++)
			cImage[imageCounter + 1] = old[imageCounter];
		
		cImage[index + 1] = GFactory.createPixels(cImage[index]);
	}
	
	public void removeImage(int index) {
		GPixels[]					old;
		int							imageCounter;
		
		if (index < 0 || index >= getImageCount()) {
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
			return;
		}
		
		if (getImageCount() == 1) {
			GOutput.throwRuntime("GESpriteForm:removeImage(): Operation canceled. A SpriteForm must have one or more images.");
			return;
		}
		
		// we know that cImage is not null
		
		old = cImage;
		cImage = new GPixels[old.length - 1];
		if (index > 0)
			for (imageCounter = 0; imageCounter < index; imageCounter++)
				cImage[imageCounter] = old[imageCounter];
		if (index < cImage.length)
			for (imageCounter = index; imageCounter < cImage.length; imageCounter++)
				cImage[imageCounter] = old[imageCounter + 1];
	}
	
	public void switchImages(int index1, int index2) {
		GPixels						temp;
		
		if (index1 == index2)
			return;
		
		temp = cImage[index1];
		cImage[index1] = cImage[index2];
		cImage[index2] = temp;
	}
	
	public void writeToFile(GEBitWriter output) throws IOException {
		int							imageCounter;
		
		if (output == null)
			throw new IOException("GESpriteForm.writeToFile: invalid file");
		
		output.pushInt(8, 1); // version
		
		((GEProperties) cProperties).writeToFile(output);
		
		output.pushInt(8, getImageCount());
		for (imageCounter = 0; imageCounter < getImageCount(); imageCounter++)
			((GEPixels) cImage[imageCounter]).writeToFile(output);
		
		output.pushBoolean(cCanPassSolidTiles);
		output.pushInt(16, cCollisionEffect);
		output.pushBoolean(cDetectCollision);
		output.pushInt(8, cCollisionWidth);
		output.pushInt(8, cCollisionHeight);
		
		output.pushInt(8, cInstanceNumberCount);
		output.pushInt(8, cInstanceFlagCount);
		output.pushInt(8, cInstanceTextCount);
	}
	
	public void writeLabelsToFile(GEBitWriter output) throws IOException {
		int							labelCounter;
		
		// instance labels
		for (labelCounter = 0; labelCounter < getInstanceNumberCount(); labelCounter++)
			output.pushString(8, cInstanceLabel[0][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getInstanceFlagCount(); labelCounter++)
			output.pushString(8, cInstanceLabel[1][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getInstanceTextCount(); labelCounter++)
			output.pushString(8, cInstanceLabel[2][labelCounter]);
	}
	
	public static void writeAllToFile(GEBitWriter output) throws IOException {
		int							spriteCounter;
		int							last;
		
		if (output == null)
			throw new IOException("GESpriteForm.readAllFromFile: invalid file");
		
		GOutput.log(" sprite form version...         ");
		last = output.getBitsWritten();
		output.pushInt(8, 1); // version
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  sprite form property count... ");
		last = output.getBitsWritten();
		output.pushInt(8, GSpriteForm.getStaticNumberCount());
		output.pushInt(8, GSpriteForm.getStaticFlagCount());
		output.pushInt(8, GSpriteForm.getStaticTextCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  sprite form count...          ");
		last = output.getBitsWritten();
		output.pushInt(8, getCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		for (spriteCounter = 0; spriteCounter < getCount(); spriteCounter++) {
			GOutput.log("  sprite form " + spriteCounter + "...              ");
			last = output.getBitsWritten();
			((GESpriteForm) sSpriteForm[spriteCounter]).writeToFile(output);
			GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		}
	}
	
	public static void writeAllLabelsToFile(GEBitWriter output) throws IOException {
		int							labelCounter;
		
		// instance labels
		for (int index = 0; index < getCount(); index++)
			((GESpriteForm) get(index)).writeLabelsToFile(output);
		
		// sprite labels
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			output.pushString(8, sLabel[0][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			output.pushString(8, sLabel[1][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			output.pushString(8, sLabel[2][labelCounter]);
	}
	
	public static void move(int fromIndex, int toIndex) {
		if (fromIndex < toIndex) {
			duplicate(toIndex);
			switchSpriteForms(fromIndex, toIndex + 1);
			remove(fromIndex);
		} else if (toIndex < fromIndex) {
			duplicate(toIndex);
			switchSpriteForms(toIndex + 1, toIndex);
			switchSpriteForms(fromIndex + 1, toIndex);
			remove(fromIndex + 1);
		}
	}
	
	public static void updateFrames(int spriteIndex) {
		int							minSprite;
		int							maxSprite;
		
		if (spriteIndex == -1) {
			minSprite = 0;
			maxSprite = GSpriteForm.getCount() - 1;
		} else {
			minSprite = spriteIndex;
			maxSprite = spriteIndex;
		}
		
		for (int spriteCounter = minSprite; spriteCounter <= maxSprite; spriteCounter++)
			for (int imageCounter = 0; imageCounter < GSpriteForm.get(spriteCounter).getImageCount(); imageCounter++)
				if (((GEPixels) GSpriteForm.get(spriteCounter).getImage(imageCounter)).getFrame() != null)
					((GEPixels) GSpriteForm.get(spriteCounter).getImage(imageCounter)).getFrame().setTitle("Sprite " + spriteCounter + ": Image " + imageCounter);
	}
	
	public static GESpriteForm getE(int index) {
		return (GESpriteForm) GSpriteForm.get(index);
	}
	
	public static void add() {
		GSpriteForm[]				old;
		int							spriteCounter;
		
		old = sSpriteForm;
		sSpriteForm = new GSpriteForm[old.length + 1];
		for (spriteCounter = 0; spriteCounter < old.length; spriteCounter++)
			sSpriteForm[spriteCounter] = old[spriteCounter];
		sSpriteForm[old.length] = GFactory.createSpriteForm();
	}
	
	public static void duplicate(int index) {
		GSpriteForm[]				old;
		int							spriteCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		old = sSpriteForm;
		sSpriteForm = new GSpriteForm[old.length + 1];
		for (spriteCounter = 0; spriteCounter <= index; spriteCounter++)
			sSpriteForm[spriteCounter] = old[spriteCounter];
		for (spriteCounter = index; spriteCounter < old.length; spriteCounter++)
			sSpriteForm[spriteCounter + 1] = old[spriteCounter];
		
		sSpriteForm[index + 1] = GFactory.createSpriteForm(sSpriteForm[index]);
		
		insertReference(index);
	}
	
	public static void remove(int index) {
		GSpriteForm[]				old;
		int							spriteCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		if (getCount() == 1) {
			GOutput.logln("GESpriteForm.remove(): Operation canceled. This object needs one or more sprite forms.");
			return;
		}
		
		old = sSpriteForm;
		sSpriteForm = new GSpriteForm[old.length - 1];
		if (index > 0)
			for (spriteCounter = 0; spriteCounter < index; spriteCounter++)
				sSpriteForm[spriteCounter] = old[spriteCounter];
		if (index < sSpriteForm.length)
			for (spriteCounter = index; spriteCounter < sSpriteForm.length; spriteCounter++)
				sSpriteForm[spriteCounter] = old[spriteCounter + 1];
		
		removeReference(index);
	}
	
	public static void switchSpriteForms(int index1, int index2) {
		GSpriteForm					temp;
		
		if (index1 == index2)
			return;
		
		if (index1 < 0 || index1 >= getCount()) {
			GOutput.throwOB("index1", index1, 0, getCount() - 1);
			return;
		} else if (index2 < 0 || index2 >= getCount()) {
			GOutput.throwOB("index2", index2, 0, getCount() - 1);
			return;
		}
		
		temp = sSpriteForm[index1];
		sSpriteForm[index1] = sSpriteForm[index2];
		sSpriteForm[index2] = temp;
		
		switchReferences(index1, index2);
	}
	
	// inserts a slot after index
	public static void insertReference(int index) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).insertSpriteFormReference(index);
	}
	
	public static void removeReference(int index) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).removeSpriteFormReference(index);
	}
	
	public static void switchReferences(int index1, int index2) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).switchSpriteFormReferences(index1, index2);
	}
	
	public String getInstanceLabel(int type, int index) {
		if (type < 0 || type > 2 || index < 0 || index >= cInstanceLabel[type].length)
			return "-";
		
		return cInstanceLabel[type][index];
	}
	
	public String[] getInstanceLabels(int type) {
		if (type < 0 || type > 2)
			return new String[] {"incorrect type"};
		
		return cInstanceLabel[type];
	}
	
	public void setInstanceLabel(int type, int index, String label) {
		if (type < 0 || type > 2 || index < 0 || index >= cInstanceLabel[type].length)
			return;
		
		cInstanceLabel[type][index] = label;
	}
	
	protected void setInstanceLabelCount(int type, int count) {
		if (type < 0 || type > 2 || count < 0 || cInstanceLabel == null)
			return;
		
		if (count <= 0)
			cInstanceLabel[type] = null;
		else {
			String[]				oldLabel;
			int						labelCounter;
			
			oldLabel = cInstanceLabel[type];
			cInstanceLabel[type] = new String[count];
			for (labelCounter = 0; labelCounter < count; labelCounter++) {
				if (oldLabel != null && labelCounter < oldLabel.length)
					cInstanceLabel[type][labelCounter] = oldLabel[labelCounter];
				else
					cInstanceLabel[type][labelCounter] = "";
			}
		}
	}
	
	public void setInstanceNumberCount(int count) {
		super.setInstanceNumberCount(count);
		
		setInstanceLabelCount(GProperties.kNumber, cInstanceNumberCount);
	}
	
	public void setInstanceFlagCount(int count) {
		super.setInstanceFlagCount(count);
		
		setInstanceLabelCount(GProperties.kFlag, cInstanceFlagCount);
	}
	
	public void setInstanceTextCount(int count) {
		super.setInstanceTextCount(count);
		
		setInstanceLabelCount(GProperties.kText, cInstanceTextCount);
	}
	
	public static void readAllLabelsFromFile(GBitReader input) throws IOException {
		int							labelCounter;
		
		// instance labels
		for (int index = 0; index < getCount(); index++)
			((GESpriteForm) get(index)).readLabelsFromFile(input);
		
		// sprite labels
		sLabel = new String[3][];
		
		sLabel[0] = new String[getStaticNumberCount()];
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			sLabel[0][labelCounter] = input.pullString();
		
		sLabel[1] = new String[getStaticFlagCount()];
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			sLabel[1][labelCounter] = input.pullString();
		
		sLabel[2] = new String[getStaticTextCount()];
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			sLabel[2][labelCounter] = input.pullString();
	}
	
	public static void clearLabels() {
		// sprite labels
		sLabel = new String[3][];
		setLabelCount(0, getStaticNumberCount());
		setLabelCount(1, getStaticFlagCount());
		setLabelCount(2, getStaticTextCount());
	}
	
	public static String getLabel(int type, int index) {
		if (type < 0 || type > 2 || index < 0 || index >= sLabel[type].length)
			return "-";
		
		return sLabel[type][index];
	}
	
	public static String[] getLabels(int type) {
		if (type < 0 || type > 2)
			return new String[] {"incorrect type"};
		
		if (sLabel[type] == null)
			return new String[0];
		return sLabel[type];
	}
	
	public static void setLabel(int type, int index, String label) {
		if (type < 0 || type > 2 || index < 0 || index >= sLabel[type].length)
			return;
		
		sLabel[type][index] = label;
	}
	
	protected static void setLabelCount(int number, int flag, int text) {
		setLabelCount(0, number);
		setLabelCount(1, flag);
		setLabelCount(2, text);
	}
	
	protected static void setLabelCount(int type, int count) {
		if (type < 0 || type > 2 || count < 0 || sLabel == null)
			return;
		
		if (count <= 0)
			sLabel[type] = null;
		else {
			String[]				oldLabel;
			int						labelCounter;
			
			oldLabel = sLabel[type];
			sLabel[type] = new String[count];
			for (labelCounter = 0; labelCounter < count; labelCounter++) {
				if (oldLabel != null && labelCounter < oldLabel.length)
					sLabel[type][labelCounter] = oldLabel[labelCounter];
				else
					sLabel[type][labelCounter] = "";
			}
		}
	}
}

