import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;


/*
Subclass of GTileForm that adds methods for saving, moving,
adding and removing tile forms. It also handles the labels for
GTileForms.
*/


public class GETileForm extends GTileForm
{
	protected static String[][]		sLabel = null;
	
	
	public void moveImage(int fromIndex, int toIndex) {
		if (fromIndex < toIndex) {
			duplicateImage(toIndex);
			switchImages(fromIndex, toIndex + 1);
			removeImage(fromIndex);
		} else if (toIndex < fromIndex) {
			duplicateImage(toIndex);
			switchImages(fromIndex + 1, toIndex);
			removeImage(fromIndex + 1);
		}
	}
	
	public void addImage() {
		GPixels[]					old;
		int							imageCounter;
		
		old = cImage;
		cImage = new GPixels[old.length + 1];
		for (imageCounter = 0; imageCounter < old.length; imageCounter++)
			cImage[imageCounter] = old[imageCounter];
		cImage[old.length] = GFactory.createPixels();
	}
	
	public void addImage(GPixels image) {
		if (image == null)
			return;
		
		addImage();
		cImage[cImage.length - 1] = image;
	}
	
	public void duplicateImage(int index) {
		GPixels[]					old;
		int							imageCounter;
		
		if (index < 0 || index >= getImageCount()) {
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
			return;
		}
		
		old = cImage;
		cImage = new GPixels[old.length + 1];
		for (imageCounter = 0; imageCounter <= index; imageCounter++)
			cImage[imageCounter] = old[imageCounter];
		for (imageCounter = index; imageCounter < old.length; imageCounter++)
			cImage[imageCounter + 1] = old[imageCounter];
		
		cImage[index + 1] = GFactory.createPixels(cImage[index]);
	}
	
	public void removeImage(int index) {
		GPixels[]					old;
		int							imageCounter;
		
		if (index < 0 || index >= getImageCount()) {
			GOutput.throwOB("index", index, 0, getImageCount() - 1);
			return;
		}
		
		if (getImageCount() == 1) {
			GOutput.throwRuntime("GETileForm:removeImage(): Operation canceled. A TileForm must have one or more images.");
			return;
		}
		
		// we know that cImage is not null
		
		old = cImage;
		cImage = new GPixels[old.length - 1];
		if (index > 0)
			for (imageCounter = 0; imageCounter < index; imageCounter++)
				cImage[imageCounter] = old[imageCounter];
		if (index < cImage.length)
			for (imageCounter = index; imageCounter < cImage.length; imageCounter++)
				cImage[imageCounter] = old[imageCounter + 1];
	}
	
	public void switchImages(int index1, int index2) {
		GPixels						temp;
		
		if (index1 == index2)
			return;
		
		temp = cImage[index1];
		cImage[index1] = cImage[index2];
		cImage[index2] = temp;
	}
	
	public void writeToFile(GEBitWriter output) throws IOException {
		int							imageCounter;
		
		if (output == null)
			throw new IOException("GETileForm.writeToFile: invalid file");
		
		output.pushInt(8, 1); // version
		
		((GEProperties) cProperties).writeToFile(output);
		
		output.pushInt(8, getImageCount());
		for (imageCounter = 0; imageCounter < getImageCount(); imageCounter++)
			((GEPixels) cImage[imageCounter]).writeToFile(output);
		
		output.pushInt(8, cAnimationPeriod);
		output.pushBoolean(cSolid);
	}
	
	public static void writeAllToFile(GEBitWriter output) throws IOException {
		int							tileCounter;
		int							last;
		
		if (output == null)
			throw new IOException("GETileForm.readAllFromFile: invalid file");
		
		GOutput.log(" tile form version...           ");
		last = output.getBitsWritten();
		output.pushInt(8, 1); // version
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  tile form property count...   ");
		last = output.getBitsWritten();
		output.pushInt(8, GTileForm.getStaticNumberCount());
		output.pushInt(8, GTileForm.getStaticFlagCount());
		output.pushInt(8, GTileForm.getStaticTextCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.log("  tile form count...            ");
		last = output.getBitsWritten();
		output.pushInt(8, getCount());
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		for (tileCounter = 0; tileCounter < getCount(); tileCounter++) {
			GOutput.log("  tile form " + tileCounter + "...                ");
			last = output.getBitsWritten();
			((GETileForm) sTileForm[tileCounter]).writeToFile(output);
			GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		}
	}
	
	public static void writeAllLabelsToFile(GEBitWriter output) throws IOException {
		int							labelCounter;
		
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			output.pushString(8, sLabel[0][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			output.pushString(8, sLabel[1][labelCounter]);
		
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			output.pushString(8, sLabel[2][labelCounter]);
	}
	
	public static void updateFrames(int tileIndex) {
		int							minTile;
		int							maxTile;
		
		if (tileIndex == -1) {
			minTile = 0;
			maxTile = GTileForm.getCount() - 1;
		} else {
			minTile = tileIndex;
			maxTile = tileIndex;
		}
		
		for (int tileCounter = minTile; tileCounter <= maxTile; tileCounter++)
			for (int imageCounter = 0; imageCounter < GTileForm.get(tileCounter).getImageCount(); imageCounter++)
				if (((GEPixels) GTileForm.get(tileCounter).getImage(imageCounter)).getFrame() != null) {
					((GEPixels) GTileForm.get(tileCounter).getImage(imageCounter)).getFrame().setTitle("Tile " + tileCounter + ": Image " + imageCounter);
					((GEPixels) GTileForm.get(tileCounter).getImage(imageCounter)).getFrame().prepareScaling();
				}
	}
	
	public static GETileForm getE(int index) {
		return (GETileForm) GTileForm.get(index);
	}
	
	public static void move(int fromIndex, int toIndex) {
		if (fromIndex < toIndex) {
			duplicate(toIndex);
			switchTileForms(fromIndex, toIndex + 1);
			remove(fromIndex);
		} else if (toIndex < fromIndex) {
			duplicate(toIndex);
			switchTileForms(toIndex + 1, toIndex);
			switchTileForms(fromIndex + 1, toIndex);
			remove(fromIndex + 1);
		}
	}
	
	public static void add() {
		GTileForm[]					old;
		int							tileCounter;
		
		old = sTileForm;
		sTileForm = new GTileForm[old.length + 1];
		for (tileCounter = 0; tileCounter < old.length; tileCounter++)
			sTileForm[tileCounter] = old[tileCounter];
		sTileForm[old.length] = GFactory.createTileForm();
	}
	
	public static void duplicate(int index) {
		GTileForm[]					old;
		int							tileCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		old = sTileForm;
		sTileForm = new GTileForm[old.length + 1];
		for (tileCounter = 0; tileCounter <= index; tileCounter++)
			sTileForm[tileCounter] = old[tileCounter];
		for (tileCounter = index; tileCounter < old.length; tileCounter++)
			sTileForm[tileCounter + 1] = old[tileCounter];
		
		sTileForm[index + 1] = GFactory.createTileForm(sTileForm[index]);
		
		insertReference(index);
	}
	
	public static void remove(int index) {
		GTileForm[]					old;
		int							tileCounter;
		
		if (index < 0 || index >= getCount()) {
			GOutput.throwOB("index", index, 0, getCount() - 1);
			return;
		}
		
		if (getCount() == 1) {
			GOutput.logln("GETileForm.remove(): Operation canceled. This object needs one or more tile forms.");
			return;
		}
		
		old = sTileForm;
		sTileForm = new GTileForm[old.length - 1];
		if (index > 0)
			for (tileCounter = 0; tileCounter < index; tileCounter++)
				sTileForm[tileCounter] = old[tileCounter];
		if (index < sTileForm.length)
			for (tileCounter = index; tileCounter < sTileForm.length; tileCounter++)
				sTileForm[tileCounter] = old[tileCounter + 1];
		
		removeReference(index);
	}
	
	public static void switchTileForms(int index1, int index2) {
		GTileForm					temp;
		
		if (index1 == index2)
			return;
		
		if (index1 < 0 || index1 >= getCount()) {
			GOutput.throwOB("index1", index1, 0, getCount() - 1);
			return;
		} else if (index2 < 0 || index2 >= getCount()) {
			GOutput.throwOB("index2", index2, 0, getCount() - 1);
			return;
		}
		
		temp = sTileForm[index1];
		sTileForm[index1] = sTileForm[index2];
		sTileForm[index2] = temp;
		
		switchReferences(index1, index2);
	}
	
	// inserts a slot after index
	public static void insertReference(int index) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).insertTileFormReference(index);
	}
	
	public static void removeReference(int index) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).removeTileFormReference(index);
	}
	
	public static void switchReferences(int index1, int index2) {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).switchTileFormReferences(index1, index2);
	}
	
	public static void readAllLabelsFromFile(GBitReader input) throws IOException {
		int							labelCounter;
		
		sLabel = new String[3][];
		
		sLabel[0] = new String[getStaticNumberCount()];
		for (labelCounter = 0; labelCounter < getStaticNumberCount(); labelCounter++)
			sLabel[0][labelCounter] = input.pullString();
		
		sLabel[1] = new String[getStaticFlagCount()];
		for (labelCounter = 0; labelCounter < getStaticFlagCount(); labelCounter++)
			sLabel[1][labelCounter] = input.pullString();
		
		sLabel[2] = new String[getStaticTextCount()];
		for (labelCounter = 0; labelCounter < getStaticTextCount(); labelCounter++)
			sLabel[2][labelCounter] = input.pullString();
	}
	
	public static void clearLabels() {
		sLabel = new String[3][0];
		setLabelCount(0, getStaticNumberCount());
		setLabelCount(1, getStaticFlagCount());
		setLabelCount(2, getStaticTextCount());
	}
	
	public static String getLabel(int type, int index) {
		if (type < 0 || type > 2 || index < 0 || index >= sLabel[type].length)
			return "-";
		
		return sLabel[type][index];
	}
	
	public static String[] getLabels(int type) {
		if (type < 0 || type > 2)
			return new String[] {"incorrect type"};
		
		if (sLabel[type] == null)
			return new String[0];
		return sLabel[type];
	}
	
	public static void setLabel(int type, int index, String label) {
		if (type == GProperties.kNumber && index >= 0 && index < getStaticNumberCount())
			sLabel[type][index] = label;
		
		if (type == GProperties.kFlag && index >= 0 && index < getStaticFlagCount())
			sLabel[type][index] = label;
		
		if (type == GProperties.kText && index >= 0 && index < getStaticTextCount())
			sLabel[type][index] = label;
	}
	
	protected static void setLabelCount(int number, int flag, int text) {
		setLabelCount(0, number);
		setLabelCount(1, flag);
		setLabelCount(2, text);
	}
	
	protected static void setLabelCount(int type, int count) {
		if (type < 0 || type > 2 || count < 0 || sLabel == null)
			return;
		
		if (count <= 0)
			sLabel[type] = new String[0];
		else {
			String[]				oldLabel;
			int						labelCounter;
			
			oldLabel = sLabel[type];
			sLabel[type] = new String[count];
			for (labelCounter = 0; labelCounter < count; labelCounter++) {
				if (oldLabel != null && labelCounter < oldLabel.length)
					sLabel[type][labelCounter] = oldLabel[labelCounter];
				else
					sLabel[type][labelCounter] = "";
			}
		}
	}
}

