import glib.*;

/***************************************************************

Copyright 2001, Bjorn Carlin.

GAMEFORM LICENSE AGREEMENT
--------------------------

This license agreement covers the source code in the folders
"GLib Source" and "GEditor Source", and the libraries "GLib.jar"
and "GEditor.jar" ("GameForm"). It does not cover the example
code distributed with GameForm.

Permission to use GameForm, whole or in part, is granted for any
personal, non-commercial purposes, to any person understanding
and fully agreeing with this license agreement.

GameForm may NOT be used, whole or in part, by any commercial
company or for any commercial software or development of
commercial software, without prior written permission from the
copyright holder. 'Commercial software' includes, but is not
limited to, charging money for using or distributing the software,
or requiring a paid subscription or license to use the software.

GameForm is provided 'as-is', without any express or implied
warranty. The author shall not be held liable for any damages
caused by the use of GameForm.

GameForm may be freely distibuted in an unaltered state, and
with no part of the package missing. Especially this copyright
notice and license agreement must be present in all source files
of all distributed copies.


CONTACT INFORMATION
-------------------

GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;


/*
Handles the main window, including all menu items. Also
the main GEditor class, and takes care of initializing.
*/


public class GEditor extends JFrame
		implements ActionListener, GEImageCanvasOwner {
	public static final boolean		kLog = true; // set to false to mute GOutput
	public static final boolean		kConfirmRemove = true; // set to false to disable confirmation dialogs for "dangerous" actions
	
	static GEditor					sInstance;
	
	protected int					cTileIndex;
	protected GEImageCanvas			cTileCanvas;
	protected GEImageCanvas			cTileImageCanvas;
	
	protected int					cSpriteIndex;
	protected GEImageCanvas			cSpriteCanvas;
	protected GEImageCanvas			cSpriteImageCanvas;
	
	protected int					cLevelIndex;
	protected GEImageCanvas			cLevelCanvas;
	
	protected static File			sFile;
	protected static File			sLastOpenedFile = new File("x");
	
	
	public GEditor() {
		JMenuBar					menuBar;
		
		JPanel						contents;
		JPanel						tileSprite;
		
		setTitle("GEditor");
		setLocation(5, 5);
		setSize(Toolkit.getDefaultToolkit().getScreenSize().width - 10, 260);
		
		// menu bar
		menuBar = new JMenuBar();
		setJMenuBar(menuBar);
		menuBar.add(createMenu("File", "", new Object[] {"New", "Open", null, "Save", "Save As", null, "Revert", null,
			"Remove Labels", null, "Exit"},
			new int[] {KeyEvent.VK_N, KeyEvent.VK_O, 0, KeyEvent.VK_S}, this));
		menuBar.add(createMenu("Tile", "T", new Object[] {
			"Edit Property Count", "Edit Properties", "Edit Labels", null,
			createMenu("Image", "TI", new Object[] {
				"Add Image", "Duplicate Image", "Edit Image", "Copy Image", "Paste Image", "Import Image", null, "Remove Image"},
				new int[0], this), null,
			"Add Tile", "Duplicate Tile", null, "Remove Tile", null,
			"Change Tile Size"},
			new int[0], this));
		menuBar.add(createMenu("Sprite", "S", new Object[] {
			"Edit Property Count", "Edit Properties", "Edit Labels", null,
			createMenu("Image", "SI", new Object[] {
				"Add Image", "Duplicate Image", "Edit Image", "Copy Image", "Paste Image", "Import Image", null, "Remove Image"},
				new int[0], this), null,
			"Add Sprite", "Duplicate Sprite", null, "Remove Sprite", null,
			"Edit Instance Property Count", "Edit Instance Labels", null, "Reset Instances"},
			new int[0], this));
		menuBar.add(createMenu("Level", "L", new Object[] {
			"Edit Property Count", "Edit Properties", "Edit Labels", null,
			"Add Level", "Duplicate Level", "Edit Level", null, "Remove Level"},
			new int[0], this));
		
		// content layout
		contents = new JPanel(new BorderLayout(3, 3));
		contents.setBorder(new EmptyBorder(3, 3, 3, 3));
		getContentPane().add(contents, BorderLayout.NORTH);
		
		tileSprite = new JPanel(new GridLayout(1, 2));
		contents.add(tileSprite, BorderLayout.NORTH);
		
		{ // tile panel
			JPanel					tileRoot;
			JPanel					tile;
			JScrollBar				tileScroll;
			JPanel					image;
			JScrollBar				imageScroll;
			
			tileRoot = new JPanel(new BorderLayout(3, 3));
			tileRoot.setBorder(new TitledBorder("Tile"));
			
			// tiles
			tile = new JPanel(new BorderLayout());
			tileScroll = new JScrollBar(JScrollBar.HORIZONTAL);
			tile.add(tileScroll, BorderLayout.SOUTH);
			cTileCanvas = new GEImageCanvas(tileScroll, GTileForm.getTileHeight(), this);
			tile.add(cTileCanvas, BorderLayout.NORTH);
			tileRoot.add(tile, BorderLayout.NORTH);
			
			// images
			image = new JPanel(new BorderLayout());
			imageScroll = new JScrollBar(JScrollBar.HORIZONTAL);
			image.add(imageScroll, BorderLayout.SOUTH);
			cTileImageCanvas = new GEImageCanvas(imageScroll, GTileForm.getTileHeight(), this);
			image.add(cTileImageCanvas, BorderLayout.NORTH);
			tileRoot.add(image, BorderLayout.SOUTH);
			
			tileSprite.add(tileRoot);
		}
		
		{ // sprite panel
			JPanel					spriteRoot;
			JPanel					sprite;
			JScrollBar				spriteScroll;
			JPanel					image;
			JScrollBar				imageScroll;
			
			spriteRoot = new JPanel(new BorderLayout(3, 3));
			spriteRoot.setBorder(new TitledBorder("Sprite"));
			
			// sprites
			sprite = new JPanel(new BorderLayout());
			spriteScroll = new JScrollBar(JScrollBar.HORIZONTAL);
			sprite.add(spriteScroll, BorderLayout.SOUTH);
			cSpriteCanvas = new GEImageCanvas(spriteScroll, GTileForm.getTileHeight(), this);
			sprite.add(cSpriteCanvas, BorderLayout.NORTH);
			spriteRoot.add(sprite, BorderLayout.NORTH);
			
			// images
			image = new JPanel(new BorderLayout());
			imageScroll = new JScrollBar(JScrollBar.HORIZONTAL);
			image.add(imageScroll, BorderLayout.SOUTH);
			cSpriteImageCanvas = new GEImageCanvas(imageScroll, GTileForm.getTileHeight(), this);
			image.add(cSpriteImageCanvas, BorderLayout.NORTH);
			spriteRoot.add(image, BorderLayout.SOUTH);
			
			tileSprite.add(spriteRoot);
		}
		
		{ // level panel
			JPanel					levelRoot;
			JPanel					level;
			JScrollBar				levelScroll;
			
			levelRoot = new JPanel(new BorderLayout(3, 3));
			levelRoot.setBorder(new TitledBorder("Level"));
			
			// levels
			level = new JPanel(new BorderLayout());
			levelScroll = new JScrollBar(JScrollBar.HORIZONTAL);
			level.add(levelScroll, BorderLayout.SOUTH);
			cLevelCanvas = new GEImageCanvas(levelScroll, GTileForm.getTileHeight() * 3, this);
			level.add(cLevelCanvas, BorderLayout.NORTH);
			levelRoot.add(level, BorderLayout.NORTH);
			
			contents.add(levelRoot, BorderLayout.SOUTH);
		}
		
		cTileIndex = cTileCanvas.getSelected();
		cSpriteIndex = cSpriteCanvas.getSelected();
		cLevelIndex = cLevelCanvas.getSelected();
		
		reset();
	}
	
	public void setVisible(boolean b) {
		// can't hide the main window
		if (b)
			super.setVisible(true);
	}
	
	// new data file or new document
	public void reset() {
		createLevelThumbnails();
		
		cTileCanvas.setSelected(0);
		cTileIndex = 0;
		cTileImageCanvas.setSelected(0);
		
		cSpriteCanvas.setSelected(0);
		cSpriteIndex = 0;
		cSpriteImageCanvas.setSelected(0);
		
		cLevelCanvas.setSelected(0);
		cLevelIndex = 0;
		
		updateImageCanvas(true);
		repaint();
	}
	
	// Some graphics have been changed
	public void updateImageCanvas(boolean resize) {
		if (resize) {
			cTileCanvas.setBufferThickness(GTileForm.getTileHeight());
			cTileImageCanvas.setBufferThickness(GTileForm.getTileHeight());
			
			cSpriteCanvas.setBufferThickness(GTileForm.getTileHeight());
			cSpriteImageCanvas.setBufferThickness(GTileForm.getTileHeight());
			
			cLevelCanvas.setBufferThickness(getMaxLevelThumbnailHeight());
		}
		
		cTileCanvas.updateScrollBar();
		cTileImageCanvas.updateScrollBar();
		cSpriteCanvas.updateScrollBar();
		cSpriteImageCanvas.updateScrollBar();
		cLevelCanvas.updateScrollBar();
		
		repaint();
	}
	
	// determines height of level list
	public int getMaxLevelThumbnailHeight() {
		int							maxHeight;
		int							levelCounter;
		
		maxHeight = 8;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			if (GELevel.getE(levelCounter).getThumbnail().getHeight() > maxHeight)
				maxHeight = GELevel.getE(levelCounter).getThumbnail().getHeight();
		
		return maxHeight;
	}
	
	public void createLevelThumbnails() {
		int							levelCounter;
		
		for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).createThumbnail();
	}
	
	// GImageCanvasOwner mehtods:
	public int getImageCount(GEImageCanvas canvas) {
		if (canvas == cTileCanvas)
			return GTileForm.getCount();
		else if (canvas == cTileImageCanvas)
			return GTileForm.get(cTileIndex).getImageCount();
		else if (canvas == cSpriteCanvas)
			return GSpriteForm.getCount();
		else if (canvas == cSpriteImageCanvas)
			return GSpriteForm.get(cSpriteIndex).getImageCount();
		else
			return GLevel.getCount();
	}
	
	public GPixels getImage(GEImageCanvas canvas, int imageIndex) {
		if (canvas == cTileCanvas)
			return GTileForm.get(imageIndex).getImage(0);
		else if (canvas == cTileImageCanvas)
			return GTileForm.get(cTileIndex).getImage(imageIndex);
		else if (canvas == cSpriteCanvas)
			return GSpriteForm.get(imageIndex).getImage(0);
		else if (canvas == cSpriteImageCanvas)
			return GSpriteForm.get(cSpriteIndex).getImage(imageIndex);
		else
			return GELevel.getE(imageIndex).getThumbnail();
	}
	
	public void newSelectedImage(GEImageCanvas canvas) {
		if (canvas == cTileCanvas) {
			cTileIndex = cTileCanvas.getSelected();
			cTileImageCanvas.updateScrollBar();
		} else if (canvas == cSpriteCanvas) {
			cSpriteIndex = cSpriteCanvas.getSelected();
			cSpriteImageCanvas.updateScrollBar();
		} else if (canvas == cLevelCanvas) {
			cLevelIndex = cLevelCanvas.getSelected();
			cLevelCanvas.updateScrollBar();
		}
	}
	
	public void clickOnSelectedImage(GEImageCanvas canvas) {
		if (canvas == cTileImageCanvas) {
			editTileImage();
		} else if (canvas == cSpriteImageCanvas) {
			editSpriteImage();
		} else if (canvas == cLevelCanvas) {
			editLevel();
		}
	}
	
	public void draggedImage(GEImageCanvas canvas, int fromIndex, int toIndex) {
		if (canvas == cTileCanvas) {
			if (toIndex < 0)
				toIndex = 0;
			if (toIndex > (GTileForm.getCount() - 1))
				toIndex = GTileForm.getCount() - 1;
			
			GETileForm.move(fromIndex, toIndex);
			
			canvas.setSelected(toIndex);
			cTileIndex = canvas.getSelected();
			cTileImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GETileForm.updateFrames(-1);
		} else if (canvas == cTileImageCanvas) {
			GTileForm				tile;
			
			tile = GTileForm.get(cTileIndex);
			if (toIndex < 0)
				toIndex = 0;
			if (toIndex > (tile.getImageCount() - 1))
				toIndex = tile.getImageCount() - 1;
			
			((GETileForm) tile).moveImage(fromIndex, toIndex);
			
			canvas.setSelected(toIndex);
			canvas.repaint();
			cTileCanvas.repaint();
			
			GELevel.updateFrames();
			GETileForm.updateFrames(cTileIndex);
		} else if (canvas == cSpriteCanvas) {
			if (toIndex < 0)
				toIndex = 0;
			if (toIndex > (GSpriteForm.getCount() - 1))
				toIndex = GSpriteForm.getCount() - 1;
			
			GESpriteForm.move(fromIndex, toIndex);
			
			canvas.setSelected(toIndex);
			cSpriteIndex = canvas.getSelected();
			cSpriteImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GESpriteForm.updateFrames(-1);
		} else if (canvas == cSpriteImageCanvas) {
			GSpriteForm				sprite;
			
			sprite = GSpriteForm.get(cSpriteIndex);
			if (toIndex < 0)
				toIndex = 0;
			if (toIndex > (sprite.getImageCount() - 1))
				toIndex = sprite.getImageCount() - 1;
			
			((GESpriteForm) sprite).moveImage(fromIndex, toIndex);
			
			canvas.setSelected(toIndex);
			canvas.repaint();
			cSpriteCanvas.repaint();
			
			GELevel.updateFrames();
			GESpriteForm.updateFrames(cSpriteIndex);
		} else {
			if (toIndex < 0)
				toIndex = 0;
			if (toIndex > (GLevel.getCount() - 1))
				toIndex = GLevel.getCount() - 1;
			
			GELevel.move(fromIndex, toIndex);
			
			canvas.setSelected(toIndex);
			cLevelIndex = canvas.getSelected();
			
			GELevel.updateFrames();
		}
	}
	
	// Menu command support methods
	
	// close all frames except the main window when a file is closed
	public void closeFrames() {
		for (int levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
			GELevel.getE(levelCounter).closeFrame();
		
		for (int tileCounter = 0; tileCounter < GTileForm.getCount(); tileCounter++)
			for (int imageCounter = 0; imageCounter < GTileForm.get(tileCounter).getImageCount(); imageCounter++)
				((GEPixels) GTileForm.get(tileCounter).getImage(imageCounter)).closeFrame();
		
		for (int spriteCounter = 0; spriteCounter < GSpriteForm.getCount(); spriteCounter++)
			for (int imageCounter = 0; imageCounter < GSpriteForm.get(spriteCounter).getImageCount(); imageCounter++)
				((GEPixels) GSpriteForm.get(spriteCounter).getImage(imageCounter)).closeFrame();
	}
	
	public int[] editPropertyCount(int number, int flag, int text) {
		Object[]					answer;
		
		answer = GEAskDialog.ask(this, 350, "Edit Property Count",
			new String[] {"Numbers", "Flags", "Texts"},
			new Integer[] {new Integer(number),
				new Integer(flag),
				new Integer(text)});
		
		if (answer == null)
			return null;
		else
			return new int[] {
				((Integer) answer[0]).intValue(),
				((Integer) answer[1]).intValue(),
				((Integer) answer[2]).intValue()};
	}
	
	public void editLabels(String[] numberLabels, String[] flagLabels, String[] textLabels) {
		int							numberCount;
		int							numberCounter;
		int							flagCount;
		int							flagCounter;
		int							textCount;
		int							textCounter;
		int							statCount;
		int							statCounter;
		
		String[]					metaLabel;
		String[]					label;
		Object[]					answer;
		
		numberCount = numberLabels.length;
		flagCount = flagLabels.length;
		textCount = textLabels.length;
		statCount = numberCount + flagCount + textCount;
		
		if (statCount <= 0) {
			System.err.println("No labels to edit");
			return;
		}
		
		metaLabel = new String[statCount];
		label = new String[statCount];
		
		statCounter = 0;
		
		for (numberCounter = 0; numberCounter < numberCount; numberCounter++) {
			metaLabel[statCounter] = "Number " + numberCounter;
			label[statCounter] = "" + numberLabels[numberCounter];
			
			statCounter++;
		}
		
		for (flagCounter = 0; flagCounter < flagCount; flagCounter++) {
			metaLabel[statCounter] = "Flag " + flagCounter;
			label[statCounter] = "" + flagLabels[flagCounter];
			
			statCounter++;
		}
		
		for (textCounter = 0; textCounter < textCount; textCounter++) {
			metaLabel[statCounter] = "Text " + textCounter;
			label[statCounter] = "" + textLabels[textCounter];
			
			statCounter++;
		}
		
		answer = GEAskDialog.ask(this, 350, "Edit Labels", metaLabel, label);
		
		if (answer != null) {
			statCounter = 0;
			
			for (numberCounter = 0; numberCounter < numberCount; numberCounter++)
				numberLabels[numberCounter] = (String) answer[statCounter++];
			
			for (flagCounter = 0; flagCounter < flagCount; flagCounter++)
				flagLabels[flagCounter] = (String) answer[statCounter++];
			
			for (textCounter = 0; textCounter < textCount; textCounter++)
				textLabels[textCounter] = (String) answer[statCounter++];
		}
	}
	
	/*
	importImage() can return before the image is completely
	imported.
	*/
	public GPixels importImage(JFrame frame) {
		JFileChooser				chooser;
		int							option;
		
		try {
			chooser = new JFileChooser(sLastOpenedFile);
			option = chooser.showOpenDialog(frame);
			
			if (option == JFileChooser.APPROVE_OPTION) {
				Image					image;
				byte[]					fileData;
				int						fileCounter;
				File					file;
				FileInputStream			input;
				
				file = chooser.getSelectedFile();
				sLastOpenedFile = file;
				
				fileData = new byte[(int) file.length()];
				
				input = new FileInputStream(file);
				input.read(fileData, 0, fileData.length);
				image = Toolkit.getDefaultToolkit().createImage(fileData);
				
				input.close();
				
				return new GEPixels(image);
			}
		} catch (Exception e) {
			System.err.println("importImages(): caught exception: " + e);
		}
		
		return null;
	}
	
	// Menu commands, in the order of appearance (basically)
	
	// FILE MENU
	
	public void emptyFile() {
		int							command;
		String						error;
		String						fileName;
		
		if (sFile == null)
			fileName = "the current file";
		else
			fileName = sFile.getName();
		
		command = JOptionPane.showConfirmDialog(new JFrame(),
				"Do you want to save changes to " + fileName + "?", "Closing " + fileName,
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
		
		if (command == JOptionPane.YES_OPTION)
			save();
		else if (command == JOptionPane.CANCEL_OPTION)
			return;
		
		closeFrames();
		
		sFile = null;
		clearAll();
		
		setTitle("GEditor: Untitled");
		reset();
	}
	
	public void open() {
		JFileChooser				chooser;
		int							option;
		
		chooser = new JFileChooser(sLastOpenedFile);
		option = chooser.showOpenDialog(new JFrame());
		if (option == JFileChooser.APPROVE_OPTION) {
			int						command;
			String					error;
			String					fileName;
			
			sLastOpenedFile = chooser.getSelectedFile();
			
			if (sFile == null)
				fileName = "the current file";
			else
				fileName = sFile.getName();
			
			command = JOptionPane.showConfirmDialog(new JFrame(),
					"Do you want to save changes to " + fileName + "?", "Closing " + fileName,
					JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
			
			if (command == JOptionPane.YES_OPTION)
				save();
			else if (command == JOptionPane.CANCEL_OPTION)
				return;
			
			closeFrames();
			
			sFile = chooser.getSelectedFile();
			setTitle("GEditor: " + sFile.getName());
			
			try {
				GBitReader			reader;
				
				reader = new GBitReader(sFile);
				readFromFile(reader);
				reader.close();
			} catch (IOException e) {
				System.err.println("GEditor.open(): error loading file: " + e);
				sFile = null;
				setTitle("GEditor: Untitled");
				clearAll();
			}
			
			reset();
		}
	}
	
	public void save() {
		if (sFile == null) {
			saveAs();
			return;
		}
		
		try {
			GEBitWriter				output;
			
			output = new GEBitWriter(sFile);
			
			writeToFile(output);
			
			output.close();
		} catch (IOException theException) {
			System.err.println("IOException while saving data: " + theException);
		}
	}
	
	// returns true if any data was saved
	public boolean saveAs() {
		File						defaultFile;
		JFileChooser				chooser;
		int							option;
		
		if (sFile == null)
			defaultFile = new File("data.data"); // path of application resides in new File()
		else
			defaultFile = sFile;
		
		chooser = new JFileChooser(defaultFile);
		option = chooser.showSaveDialog(new JFrame());
		if (option == JFileChooser.APPROVE_OPTION) {
			sFile = chooser.getSelectedFile();
			if (sFile != null) {
				save();
				sLastOpenedFile = sFile;
				setTitle("GEditor: " + sFile.getName());
				return true;
			}
		}
		return false;
	}
	
	public void revert() {
		int							command;
		
		if (sFile == null) {
			System.err.println("No file to revert to.");
			return;
		}
		
		command = JOptionPane.showConfirmDialog(new JFrame(),
				"Do you want to revert to the last saved version of " + sFile.getName() + "?", "Revert",
				JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
		
		if (command == JOptionPane.YES_OPTION) {
			String					error;
			
			closeFrames();
			
			try {
				GBitReader			reader;
				
				reader = new GBitReader(sFile);
				readFromFile(reader);
				reader.close();
			} catch (IOException e) {
				System.err.println("error loading file: " + e);
				sFile = null;
				clearAll();
			}
			
			reset();
		}
	}
	
	public void removeLabels() {
		if (GEAskDialog.ask(this, 300, "Remove Labels", new String[] {
				"Are you sure you want to remove the labels from this file? The operation can't be undone."}, null) != null) {
			GETileForm.clearLabels();
			GESpriteForm.clearLabels();
			GESpriteForm.clearAllInstanceLabels();
			GELevel.clearLabels();
		}
	}
	
	public void quit() {
		String						fileName;
		int							command;
		
		if (sFile == null)
			fileName = "the current file";
		else
			fileName = sFile.getName();
		
		command = JOptionPane.showConfirmDialog(new JFrame(),
				"Do you want to save changes to " + fileName + "?", "Exit",
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
		
		if (command == JOptionPane.YES_OPTION) {
			if (sFile != null) {
				save();
				System.exit(0);
			} else {
				if (saveAs())
					System.exit(0);
				else
					return; // canceled save as, cancel quit
			}
		} else if (command == JOptionPane.NO_OPTION)
			System.exit(0);
		else if (command == JOptionPane.CANCEL_OPTION)
			return;
	}
	
	// TILE MENU
	
	public void editTilePropertyCount() {
		int[]						answer;
		
		answer = editPropertyCount(GTileForm.getStaticNumberCount(),
				GTileForm.getStaticFlagCount(),
				GTileForm.getStaticTextCount());
		
		if (answer != null) {
			GTileForm.setStaticNumberCount(answer[0]);
			GTileForm.setStaticFlagCount(answer[1]);
			GTileForm.setStaticTextCount(answer[2]);
			
			GETileForm.setLabelCount(answer[0], answer[1], answer[2]);
		}
	}
	
	public void editTileProperties() {
		int							numberCount;
		int							numberCounter;
		int							flagCount;
		int							flagCounter;
		int							textCount;
		int							textCounter;
		int							statCount;
		int							statCounter;
		
		GTileForm					tileForm;
		String[]					label;
		Object[]					answer;
		
		numberCount = GTileForm.getStaticNumberCount();
		flagCount = GTileForm.getStaticFlagCount();
		textCount = GTileForm.getStaticTextCount();
		statCount = 2 + numberCount + flagCount + textCount;
		
		label = new String[statCount];
		answer = new Object[statCount];
		tileForm = GTileForm.get(cTileIndex);
		
		statCounter = 0;
		
		// tile-specific properties
		label[statCounter] = "Animation Period";
		answer[statCounter++] = new Integer(tileForm.getAnimationPeriod());
		
		label[statCounter] = "Solid";
		answer[statCounter++] = new Boolean(tileForm.getSolid());
		
		for (numberCounter = 0; numberCounter < numberCount; numberCounter++) {
			label[statCounter] = GETileForm.getLabel(GProperties.kNumber, numberCounter);
			answer[statCounter] = new Integer(tileForm.getNumber(numberCounter));
			
			statCounter++;
		}
		
		for (flagCounter = 0; flagCounter < flagCount; flagCounter++) {
			label[statCounter] = GETileForm.getLabel(GProperties.kFlag, flagCounter);
			answer[statCounter] = new Boolean(tileForm.getFlag(flagCounter));
			
			statCounter++;
		}
		
		for (textCounter = 0; textCounter < textCount; textCounter++) {
			label[statCounter] = GETileForm.getLabel(GProperties.kText, textCounter);
			answer[statCounter] = GETileForm.getE(cTileIndex).getText(textCounter);
			
			statCounter++;
		}
		
		answer = GEAskDialog.ask(this, 350, "Edit Properties", label, answer);
		
		if (answer != null) {
			statCounter = 0;
			
			// tile-specific properties
			tileForm.setAnimationPeriod(((Integer) answer[statCounter++]).intValue());
			
			tileForm.setSolid(((Boolean) answer[statCounter++]).booleanValue());
			
			for (numberCounter = 0; numberCounter < numberCount; numberCounter++)
				tileForm.getProperties().setNumber(numberCounter, ((Integer) answer[statCounter++]).intValue());
			
			for (flagCounter = 0; flagCounter < flagCount; flagCounter++)
				tileForm.getProperties().setFlag(flagCounter, ((Boolean) answer[statCounter++]).booleanValue());
			
			for (textCounter = 0; textCounter < textCount; textCounter++)
				tileForm.getProperties().setText(textCounter, (String) answer[statCounter++]);
		}
	}
	
	public void editTileLabels() {
		editLabels(GETileForm.getLabels(0), GETileForm.getLabels(1), GETileForm.getLabels(2));
	}
	
	// Ask the GEPixels object to open its frame
	public void editTileImage() {
		((GEPixels) GTileForm.get(cTileIndex).getImage(cTileImageCanvas.getSelected())).openFrame("Tile " + cTileIndex + ": Image " + cTileImageCanvas.getSelected(), false);
	}
	
	public void changeTileSize() {
		Object[]					answer;
		
		answer = GEAskDialog.ask(this, 350, "Change Tile Size",
			new String[] {"Width", "Height"},
			new Integer[] {new Integer(GTileForm.getTileWidth()),
					new Integer(GTileForm.getTileHeight())});
		
		if (answer != null) {
			int						width;
			int						height;
			
			width = ((Integer) answer[0]).intValue();
			height = ((Integer) answer[1]).intValue();
			
			if (width > 3 && height > 3 && width <= GPixels.kMaxSize && height <= GPixels.kMaxSize) {
				for (int index = 0; index < GTileForm.getCount(); index++)
					for (int image = 0; image < GTileForm.get(index).getImageCount(); image++)
						((GEPixels) GTileForm.get(index).getImage(image)).scaleTo(width, height);
				
				GELevel.updateFrames();
				GETileForm.updateFrames(-1);
				
				updateImageCanvas(true);
			}
		}
	}
	
	// SPRITE MENU
	
	public void editSpritePropertyCount() {
		int[]						answer;
		
		answer = editPropertyCount(GSpriteForm.getStaticNumberCount(),
				GSpriteForm.getStaticFlagCount(),
				GSpriteForm.getStaticTextCount());
		
		if (answer != null) {
			GSpriteForm.setStaticNumberCount(answer[0]);
			GSpriteForm.setStaticFlagCount(answer[1]);
			GSpriteForm.setStaticTextCount(answer[2]);
			
			GESpriteForm.setLabelCount(answer[0], answer[1], answer[2]);
		}
	}
	
	public void editSpriteProperties() {
		int							numberCount;
		int							numberCounter;
		int							flagCount;
		int							flagCounter;
		int							textCount;
		int							textCounter;
		int							statCount;
		int							statCounter;
		
		GSpriteForm					spriteForm;
		String[]					label;
		Object[]					answer;
		
		numberCount = GSpriteForm.getStaticNumberCount();
		flagCount = GSpriteForm.getStaticFlagCount();
		textCount = GSpriteForm.getStaticTextCount();
		statCount = 5 + numberCount + flagCount + textCount;
		
		label = new String[statCount];
		answer = new Object[statCount];
		spriteForm = GSpriteForm.get(cSpriteIndex);
		
		statCounter = 0;
		
		// sprite-specific properties
		label[statCounter] = "Default can pass solid tiles";
		answer[statCounter++] = new Boolean(spriteForm.getCanPassSolidTiles());
		
		label[statCounter] = "Default collision effect";
		answer[statCounter++] = new Integer(spriteForm.getCollisionEffect());
		
		label[statCounter] = "Default accept collision events";
		answer[statCounter++] = new Boolean(spriteForm.getDetectCollision());
		
		label[statCounter] = "Default collision width";
		answer[statCounter++] = new Integer(spriteForm.getCollisionWidth());
		
		label[statCounter] = "Default collision height";
		answer[statCounter++] = new Integer(spriteForm.getCollisionHeight());
		
		// numbers, flags and texts
		for (numberCounter = 0; numberCounter < numberCount; numberCounter++) {
			label[statCounter] = GESpriteForm.getLabel(GProperties.kNumber, numberCounter);
			answer[statCounter] = new Integer(spriteForm.getNumber(numberCounter));
			
			statCounter++;
		}
		
		for (flagCounter = 0; flagCounter < flagCount; flagCounter++) {
			label[statCounter] = GESpriteForm.getLabel(GProperties.kFlag, flagCounter);
			answer[statCounter] = new Boolean(spriteForm.getFlag(flagCounter));
			
			statCounter++;
		}
		
		for (textCounter = 0; textCounter < textCount; textCounter++) {
			label[statCounter] = GESpriteForm.getLabel(GProperties.kText, textCounter);
			answer[statCounter] = spriteForm.getText(textCounter);
			
			statCounter++;
		}
		
		answer = GEAskDialog.ask(this, 450, "Edit Properties", label, answer);
		
		if (answer != null) {
			statCounter = 0;
			
			// sprite form-specific properties
			spriteForm.setCanPassSolidTiles(((Boolean) answer[statCounter++]).booleanValue());
			spriteForm.setCollisionEffect(((Integer) answer[statCounter++]).intValue());
			spriteForm.setDetectCollision(((Boolean) answer[statCounter++]).booleanValue());
			spriteForm.setCollisionWidth(((Integer) answer[statCounter++]).intValue());
			spriteForm.setCollisionHeight(((Integer) answer[statCounter++]).intValue());
			
			// numbers, flags and texts
			for (numberCounter = 0; numberCounter < numberCount; numberCounter++)
				spriteForm.getProperties().setNumber(numberCounter, ((Integer) answer[statCounter++]).intValue());
			
			for (flagCounter = 0; flagCounter < flagCount; flagCounter++)
				spriteForm.getProperties().setFlag(flagCounter, ((Boolean) answer[statCounter++]).booleanValue());
			
			for (textCounter = 0; textCounter < textCount; textCounter++)
				spriteForm.getProperties().setText(textCounter, (String) answer[statCounter++]);
		}
	}
	
	public void editSpriteLabels() {
		editLabels(GESpriteForm.getLabels(0), GESpriteForm.getLabels(1), GESpriteForm.getLabels(2));
	}
	
	// Ask the GEPixels object to open its frame
	public void editSpriteImage() {
		((GEPixels) GSpriteForm.get(cSpriteIndex).getImage(cSpriteImageCanvas.getSelected())).openFrame("Sprite " + cSpriteIndex + ": Image " + cSpriteImageCanvas.getSelected(), true);
	}
	
	public void editSpriteInstancePropertyCount() {
		int[]						answer;
		
		answer = editPropertyCount(GSpriteForm.getStaticNumberCount(),
				GSpriteForm.getStaticFlagCount(),
				GSpriteForm.getStaticTextCount());
		
		if (answer != null) {
			GSpriteForm.get(cSpriteIndex).setInstanceNumberCount(answer[0]);
			GSpriteForm.get(cSpriteIndex).setInstanceFlagCount(answer[1]);
			GSpriteForm.get(cSpriteIndex).setInstanceTextCount(answer[2]);
		}
	}
	
	public void editSpriteInstanceLabels() {
		editLabels(GESpriteForm.getE(cSpriteIndex).getLabels(0), 
				GESpriteForm.getE(cSpriteIndex).getLabels(1), 
				GESpriteForm.getE(cSpriteIndex).getLabels(2));
	}
	
	// LEVEL MENU
	
	public void editLevelPropertyCount() {
		int[]						answer;
		
		answer = editPropertyCount(GLevel.getStaticNumberCount(),
				GLevel.getStaticFlagCount(),
				GLevel.getStaticTextCount());
		
		if (answer != null) {
			GLevel.setStaticNumberCount(answer[0]);
			GLevel.setStaticFlagCount(answer[1]);
			GLevel.setStaticTextCount(answer[2]);
			
			GELevel.setLabelCount(answer[0], answer[1], answer[2]);
		}
	}
	
	public void editLevelProperties() {
		int							numberCount;
		int							numberCounter;
		int							flagCount;
		int							flagCounter;
		int							textCount;
		int							textCounter;
		int							statCount;
		int							statCounter;
		
		GLevel						level;
		String[]					label;
		Object[]					answer;
		
		numberCount = GLevel.getStaticNumberCount();
		flagCount = GLevel.getStaticFlagCount();
		textCount = GLevel.getStaticTextCount();
		statCount = 2 + numberCount + flagCount + textCount;
		
		label = new String[statCount];
		answer = new Object[statCount];
		level = GLevel.get(cLevelIndex);
		
		statCounter = 0;
		
		// level-specific properties
		label[statCounter] = "Start Scroll X";
		answer[statCounter++] = new Integer(level.getStartScrollX());
		
		label[statCounter] = "Start Scroll Y";
		answer[statCounter++] = new Integer(level.getStartScrollY());
		
		for (numberCounter = 0; numberCounter < numberCount; numberCounter++) {
			label[statCounter] = GELevel.getLabel(GProperties.kNumber, numberCounter);
			answer[statCounter] = new Integer(level.getNumber(numberCounter));
			
			statCounter++;
		}
		
		for (flagCounter = 0; flagCounter < flagCount; flagCounter++) {
			label[statCounter] = GELevel.getLabel(GProperties.kFlag, flagCounter);
			answer[statCounter] = new Boolean(level.getFlag(flagCounter));
			
			statCounter++;
		}
		
		for (textCounter = 0; textCounter < textCount; textCounter++) {
			label[statCounter] = GELevel.getLabel(GProperties.kText, textCounter);
			answer[statCounter] = level.getText(textCounter);
			
			statCounter++;
		}
		
		answer = GEAskDialog.ask(this, 350, "Edit Properties", label, answer);
		
		if (answer != null) {
			statCounter = 0;
			
			// level-specific properties
			level.setStartScrollX(((Integer) answer[statCounter++]).intValue());
			level.setStartScrollY(((Integer) answer[statCounter++]).intValue());
			
			for (numberCounter = 0; numberCounter < numberCount; numberCounter++)
				level.getProperties().setNumber(numberCounter, ((Integer) answer[statCounter++]).intValue());
			
			for (flagCounter = 0; flagCounter < flagCount; flagCounter++)
				level.getProperties().setFlag(flagCounter, ((Boolean) answer[statCounter++]).booleanValue());
			
			for (textCounter = 0; textCounter < textCount; textCounter++)
				level.getProperties().setText(textCounter, (String) answer[statCounter++]);
		}
	}
	
	public void editLevelLabels() {
		editLabels(GELevel.getLabels(0), GELevel.getLabels(1), GELevel.getLabels(2));
	}
	
	// Ask the GELevel object to open its frame
	public void editLevel() {
		GELevel.getE(cLevelIndex).openFrame("Level " + cLevelIndex);
	}
	
	// MENU SWITCH
	public void actionPerformed(ActionEvent theEvent) {
		String						command;
		
		command = theEvent.getActionCommand();
		
		// File menu items
		if (command.equals("New"))
			emptyFile();
		else if (command.equals("Open"))
			open();
		else if (command.equals("Save"))
			save();
		else if (command.equals("Save As"))
			saveAs();
		else if (command.equals("Revert"))
			revert();
		else if (command.equals("Remove Labels"))
			removeLabels();
		else if (command.equals("Memory Status")) {
			// removed from file menu
			System.out.println("Used / Total Before GC:  " + (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1024 + " kB / " + Runtime.getRuntime().totalMemory() / 1024 + " kB");
			Runtime.getRuntime().gc();
			System.out.println("Used / Total After GC:   " + (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1024 + " kB / " + Runtime.getRuntime().totalMemory() / 1024 + " kB");
		} else if (command.equals("Exit"))
			quit();
		
		// Tile menu items
		else if (command.equals("TEdit Property Count"))
			editTilePropertyCount();
		else if (command.equals("TEdit Properties"))
			editTileProperties();
		else if (command.equals("TEdit Labels"))
			editTileLabels();
		else if (command.equals("TAdd Tile")) {
			GETileForm.add();
			cTileCanvas.setSelected(GTileForm.getCount() - 1);
			cTileIndex = cTileCanvas.getSelected();
			cTileCanvas.updateScrollBar();
			cTileImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
		} else if (command.equals("TDuplicate Tile")) {
			GETileForm.duplicate(cTileCanvas.getSelected());
			cTileCanvas.setSelected(cTileCanvas.getSelected() + 1);
			cTileIndex = cTileCanvas.getSelected();
			cTileCanvas.updateScrollBar();
			cTileImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GETileForm.updateFrames(-1);
		} else if (command.equals("TRemove Tile")) {
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Remove Tile", "Are you sure you want to remove this tile?"))
					return;
			
			// close image frames
			for (int imageCounter = 0; imageCounter < GETileForm.get(cTileCanvas.getSelected()).getImageCount(); imageCounter++)
				((GEPixels) GETileForm.get(cTileCanvas.getSelected()).getImage(imageCounter)).closeFrame();
			
			GETileForm.remove(cTileCanvas.getSelected());
			cTileCanvas.setSelected(cTileCanvas.getSelected());
			cTileIndex = cTileCanvas.getSelected();
			cTileCanvas.updateScrollBar();
			cTileImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GETileForm.updateFrames(-1);
		} else if (command.equals("TChange Tile Size"))
			changeTileSize();
		
		// Tile->Image menu items
		else if (command.equals("TIAdd Image")) {
			GETileForm.getE(cTileIndex).addImage();
			cTileImageCanvas.updateScrollBar();
			cTileImageCanvas.setSelected(GTileForm.get(cTileIndex).getImageCount() - 1);
		} else if (command.equals("TIDuplicate Image")) {
			GETileForm.getE(cTileIndex).duplicateImage(cTileImageCanvas.getSelected());
			cTileImageCanvas.updateScrollBar();
			cTileImageCanvas.setSelected(cTileImageCanvas.getSelected() + 1);
			
			GETileForm.updateFrames(cTileIndex);
		} else if (command.equals("TIEdit Image")) {
			editTileImage();
		} else if (command.equals("TICopy Image")) {
			GEPixels.setClip(GTileForm.get(cTileIndex).getImage(cTileImageCanvas.getSelected()));
		} else if (command.equals("TIPaste Image")) {
			if (!GEPixels.hasClip()) {
				System.err.println("Clip is empty");
				return;
			} else if (GEPixels.getClip().getWidth() != GTileForm.getTileWidth() ||
					GEPixels.getClip().getHeight() != GTileForm.getTileHeight()) {
				System.err.println("Clip has Invalid tile size (" + GEPixels.getClip().getWidth() + " * " + GEPixels.getClip().getHeight() + ")");
				return;
			}
			
			GETileForm.getE(cTileIndex).addImage((GPixels) GEPixels.getClip().clone());
			cTileImageCanvas.updateScrollBar();
			cTileImageCanvas.setSelected(GTileForm.get(cTileIndex).getImageCount() - 1);
		} else if (command.equals("TIImport Image")) {
			GPixels					image;
			
			image = importImage(this);
			if (image == null)
				return;
			if (image.getWidth() != GTileForm.getTileWidth() ||
					image.getHeight() != GTileForm.getTileHeight()) {
				System.err.println("Image has an invalid tile size (" + image.getWidth() + " * " + image.getHeight() + ")");
				return;
			}
			
			GETileForm.getE(cTileIndex).addImage(image);
			
			cTileImageCanvas.updateScrollBar();
			cTileImageCanvas.setSelected(GTileForm.get(cTileIndex).getImageCount() - 1);
		} else if (command.equals("TIRemove Image")) {
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Remove Tile Image", "Are you sure you want to remove this tile image?"))
					return;
			
			// close image frame
			((GEPixels) GTileForm.get(cTileIndex).getImage(cTileImageCanvas.getSelected())).closeFrame();
			
			GETileForm.getE(cTileIndex).removeImage(cTileImageCanvas.getSelected());
			cTileImageCanvas.updateScrollBar();
			cTileImageCanvas.setSelected(cTileImageCanvas.getSelected());
			
			GETileForm.updateFrames(cTileIndex);
		}
		
		// Sprite menu items
		else if (command.equals("SEdit Property Count"))
			editSpritePropertyCount();
		else if (command.equals("SEdit Properties"))
			editSpriteProperties();
		else if (command.equals("SEdit Labels"))
			editSpriteLabels();
		else if (command.equals("SAdd Sprite")) {
			GESpriteForm.add();
			cSpriteCanvas.setSelected(GSpriteForm.getCount() - 1);
			cSpriteIndex = cSpriteCanvas.getSelected();
			cSpriteCanvas.updateScrollBar();
			cSpriteImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
		} else if (command.equals("SDuplicate Sprite")) {
			GESpriteForm.duplicate(cSpriteCanvas.getSelected());
			cSpriteCanvas.setSelected(cSpriteCanvas.getSelected() + 1);
			cSpriteIndex = cSpriteCanvas.getSelected();
			cSpriteCanvas.updateScrollBar();
			cSpriteImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GESpriteForm.updateFrames(-1);
		} else if (command.equals("SRemove Sprite")) {
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Remove Sprite", "Are you sure you want to remove this sprite?"))
					return;
			
			// close image frames
			for (int imageCounter = 0; imageCounter < GESpriteForm.get(cSpriteCanvas.getSelected()).getImageCount(); imageCounter++)
				((GEPixels) GESpriteForm.get(cSpriteCanvas.getSelected()).getImage(imageCounter)).closeFrame();
			
			GESpriteForm.remove(cSpriteCanvas.getSelected());
			cSpriteCanvas.setSelected(cSpriteCanvas.getSelected());
			cSpriteIndex = cSpriteCanvas.getSelected();
			cSpriteCanvas.updateScrollBar();
			cSpriteImageCanvas.updateScrollBar();
			
			GELevel.updateFrames();
			GESpriteForm.updateFrames(-1);
		} else if (command.equals("SEdit Instance Property Count")) {
			editSpriteInstancePropertyCount();
		} else if (command.equals("SEdit Instance Labels")) {
			editSpriteInstanceLabels();
		} else if (command.equals("SReset Instances")) {
			int									levelCounter;
			int									spriteCounter;
			GSpriteForm							form;
			GSprite								sprite;
			
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Reset Instances", "Are you sure you want to reset all instances of this sprite form?"))
					return;
			
			// reset the relative variables of all instances of this form
			
			form = GSpriteForm.get(cSpriteIndex);
			
			for (levelCounter = 0; levelCounter < GLevel.getCount(); levelCounter++)
				for (spriteCounter = 0; spriteCounter < GLevel.get(levelCounter).getSpriteCount(); spriteCounter++)
					if (GLevel.get(levelCounter).getSprite(spriteCounter).getFormIndex() == cSpriteIndex) {
						sprite = GLevel.get(levelCounter).getSprite(spriteCounter);
						
						sprite.setCanPassSolidTiles(form.getCanPassSolidTiles());
						sprite.setCollisionEffect(form.getCollisionEffect());
						sprite.setDetectCollision(form.getDetectCollision());
						sprite.setCollisionWidth(form.getCollisionWidth());
						sprite.setCollisionHeight(form.getCollisionHeight());
					}
		}
		
		// Sprite->Image menu items
		else if (command.equals("SIAdd Image")) {
			GESpriteForm.getE(cSpriteIndex).addImage();
			cSpriteImageCanvas.updateScrollBar();
			cSpriteImageCanvas.setSelected(GSpriteForm.get(cSpriteIndex).getImageCount() - 1);
		} else if (command.equals("SIDuplicate Image")) {
			GESpriteForm.getE(cSpriteIndex).duplicateImage(cSpriteImageCanvas.getSelected());
			cSpriteImageCanvas.updateScrollBar();
			cSpriteImageCanvas.setSelected(cSpriteImageCanvas.getSelected() + 1);
			
			GESpriteForm.updateFrames(cSpriteIndex);
		} else if (command.equals("SIEdit Image")) {
			editSpriteImage();
		} else if (command.equals("SICopy Image")) {
			GEPixels.setClip(GSpriteForm.get(cSpriteIndex).getImage(cSpriteImageCanvas.getSelected()));
		} else if (command.equals("SIPaste Image")) {
			if (!GEPixels.hasClip()) {
				System.err.println("Clip is empty");
				return;
			}
			
			GESpriteForm.getE(cSpriteIndex).addImage((GPixels) GEPixels.getClip().clone());
			cSpriteImageCanvas.updateScrollBar();
			cSpriteImageCanvas.setSelected(GSpriteForm.get(cSpriteIndex).getImageCount() - 1);
		} else if (command.equals("SIImport Image")) {
			GPixels					image;
			
			image = importImage(this);
			if (image == null)
				return;
			
			GESpriteForm.getE(cSpriteIndex).addImage(image);
			
			cSpriteImageCanvas.updateScrollBar();
			cSpriteImageCanvas.setSelected(GSpriteForm.get(cSpriteIndex).getImageCount() - 1);
		} else if (command.equals("SIRemove Image")) {
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Remove Sprite Image", "Are you sure you want to remove this sprite image?"))
					return;
			
			// close image frame
			((GEPixels) GSpriteForm.get(cSpriteIndex).getImage(cSpriteImageCanvas.getSelected())).closeFrame();
			
			GESpriteForm.getE(cSpriteIndex).removeImage(cSpriteImageCanvas.getSelected());
			cSpriteImageCanvas.updateScrollBar();
			cSpriteImageCanvas.setSelected(cSpriteImageCanvas.getSelected());
			
			GESpriteForm.updateFrames(cSpriteIndex);
		}
		
		// Level menu items
		else if (command.equals("LEdit Property Count"))
			editLevelPropertyCount();
		else if (command.equals("LEdit Properties"))
			editLevelProperties();
		else if (command.equals("LEdit Labels"))
			editLevelLabels();
		else if (command.equals("LAdd Level")) {
			GELevel.add();
			updateImageCanvas(true);
			cLevelCanvas.setSelected(GLevel.getCount() - 1);
			cLevelIndex = cLevelCanvas.getSelected();
		} else if (command.equals("LDuplicate Level")) {
			GELevel.duplicate(cLevelCanvas.getSelected());
			updateImageCanvas(true);
			cLevelCanvas.setSelected(cLevelCanvas.getSelected() + 1);
			cLevelIndex = cLevelCanvas.getSelected();
			
			GELevel.updateFrames();
		} else if (command.equals("LEdit Level")) {
			editLevel();
		} else if (command.equals("LRemove Level")) {
			if (kConfirmRemove)
				if (!GEAskDialog.confirm(this, "Remove Level", "Are you sure you want to remove this level?"))
					return;
			
			GELevel.getE(cLevelCanvas.getSelected()).closeFrame();
			
			GELevel.remove(cLevelCanvas.getSelected());
			updateImageCanvas(true);
			cLevelCanvas.setSelected(cLevelCanvas.getSelected());
			cLevelIndex = cLevelCanvas.getSelected();
			
			GELevel.updateFrames();
		}
	}
	
	// convenience method for menu creation
	public static JMenu createMenu(String name, String prefix, Object[] item, int[] accelerator, ActionListener listener) {
		JMenu						theMenu;
		int							itemCounter;
		JMenuItem					theItem;
		
		theMenu = new JMenu(name);
		if (item != null)
			for (itemCounter = 0; itemCounter < item.length; itemCounter++) {
				if (item[itemCounter] == null)
					theMenu.addSeparator();
				else {
					if (item[itemCounter] instanceof JMenuItem)
						theItem = (JMenuItem) item[itemCounter];
					else if (item[itemCounter] instanceof String) {
						theItem = new JMenuItem((String) item[itemCounter]);
						theItem.setActionCommand(prefix + item[itemCounter]);
					} else
						continue;
					
					if (listener != null)
						theItem.addActionListener(listener);
					
					if (accelerator.length > itemCounter && accelerator[itemCounter] != 0)
						theItem.setAccelerator(KeyStroke.getKeyStroke(accelerator[itemCounter],
								InputEvent.CTRL_MASK));
					
					theMenu.add(theItem);
				}
			}
		
		return theMenu;
	}
	
	public static JMenu createMenu(String name, Object[] item, int[] accelerator, ActionListener listener) {
		return createMenu(name, "", item, accelerator, listener);
	}
	
	protected static void readFromFile(GBitReader input) throws IOException {
		int							version;
		
		GOutput.logln("reading " + sFile.getName() + "...");
		
		if (input == null)
			throw new IOException("GEditor.readFromFile: invalid file");
		
		version = input.pullInt(8);
		
		GOutput.logln("version " + version);
		
		if (version == 1) {
			GTileForm.readAllFromFile(input);
			GSpriteForm.readAllFromFile(input);
			GLevel.readAllFromFile(input);
			
			GETileForm.readAllLabelsFromFile(input);
			GESpriteForm.readAllLabelsFromFile(input);
			GELevel.readAllLabelsFromFile(input);
			
			GOutput.logln("reading complete");
		} else {
			throw new IOException("GEditor.readFromFile: Unknown version: " + version);
		}
	}
	
	protected static void writeToFile(GEBitWriter output) throws IOException {
		int							last;
		
		GOutput.logln("writing " + sFile.getName() + "...");
		
		GOutput.log(" version...                     ");
		last = output.getBitsWritten();
		output.pushInt(8, 1);
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GETileForm.writeAllToFile(output);
		GESpriteForm.writeAllToFile(output);
		GELevel.writeAllToFile(output);
		
		GOutput.log(" labels...                      ");
		last = output.getBitsWritten();
		GETileForm.writeAllLabelsToFile(output);
		GESpriteForm.writeAllLabelsToFile(output);
		GELevel.writeAllLabelsToFile(output);
		GOutput.logln("done, " + (output.getBitsWritten() - last) + " bits (" + output.getBitsWritten() + ")");
		
		GOutput.logln("writing complete");
	}
	
	public static void clearAll() {
		GLevel.clearAll();
		GTileForm.clearAll();
		GSpriteForm.clearAll();
		
		GELevel.clearLabels();
		GETileForm.clearLabels();
		GESpriteForm.clearLabels();
	}
	
	public static void main(String args[]) {
		// initialize GameForm
		GOutput.sMute = !kLog;
		GOutput.sThrowExceptions = false;
		GOutput.sTraceStack = false;
		GFactory.setFactory(new GEFactory());
		
		// try to open the file "data.data"
		sFile = new File("data.data");
		if (sFile.exists()) {
			try {
				GBitReader				reader;
				
				reader = new GBitReader(sFile);
				readFromFile(reader);
				reader.close();
			} catch (IOException e) {
				System.err.println("Error loading file data.data: " + e);
				sFile = null;
				clearAll();
			}
		} else {
			sFile = null;
			clearAll();
		}
		
		// open the main window
		sInstance = new GEditor();
		if (sFile == null)
			sInstance.setTitle("GEditor: Untitled");
		else
			sInstance.setTitle("GEditor: " + sFile.getName());
		sInstance.show();
	}
}
