
public class BouncingBall extends GSprite
{
	// number indices
	static final int				kSpeedProperty = 0;
	static final int				kColorShiftProperty = 1;
	static final int				kColorProperty = 2;
	
	static final int				kMaxColor = 360;
	
	static final int[][]			kMod = new int[][] {
											{255, 255, 0},
											{127, 225, 127},
											{0, 255, 255},
											{127, 127, 255},
											{255, 0, 255},
											{255, 127, 127},
											{255, 255, 0}};
	
	// 0 = 360
	int								cColor;
	
	// a touched ball is about to be removed
	boolean							cTouched;
	
	
	public BouncingBall(int formIndex) {
		super(formIndex);
		
		cTouched = false;
	}
	
	/*
	Uses a single GPixels image to draw the different colors, by setting
	the cColorMod of BouncingPixels, which then uses the custom copyPixelsColor()
	method to draw the ball.
	*/
	public void draw(int[] canvas, int canvasWidth, int scrollX, int scrollY) {
		GPixels						image;
		
		int							redMod;
		int							greenMod;
		int							blueMod;
		
		// check if sprite is visible
		if ((getRight() >> 10) < scrollX ||
				(getBottom() >> 10) < scrollY ||
				(getLeft() >> 10) >= scrollX + canvasWidth ||
				(getTop() >> 10) >= scrollY + canvas.length / canvasWidth)
			return;
		
		redMod = (kMod[cColor * 6 / kMaxColor + 1][0] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][0] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		greenMod = (kMod[cColor * 6 / kMaxColor + 1][1] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][1] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		blueMod = (kMod[cColor * 6 / kMaxColor + 1][2] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][2] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		
		image = cForm.getImage(0);
		((BouncingPixels) image).setColorMod((redMod << 16) + (greenMod << 8) + blueMod);
		image.drawWrapped(canvas, canvasWidth, getLeft() >> 10, getTop() >> 10,
			0, scrollX, scrollY);
		((BouncingPixels) image).setColorMod(0xFFFFFFFF);
	}
	
	/*
	The default action for a sprite when hitting a solid tile is
	"sticking" to the tile. BouncingBall.moveCanNotPassSolid() makes
	the ball bounce, reusing most of the code from GSprite.
	*/
	public void moveCanNotPassSolid() {
		// assumes the sprite is not in a solid tile before moving
		
		// note: a sprite incapable of passing solid tiles
		//       can not move faster than the tile size
		// note: a sprite incapable of passing solid tiles
		//       should not be rotated (change collision
		//       bounds) close to solid tiles
		
		// assume not touching solid
		setTouchingSolid(false);
		
		if (getSpeedX() != 0) {
			// horizontal movement
			int						column;
			int						topRow;
			int						bottomRow;
			int						y;
			int						suggestedX;
			boolean					bounce;
			
			if (getSpeedX() > 0) {
				// moving right
				if (getSpeedX() > ((GTileForm.getTileWidth() - 1) << 10))
					setSpeedX((GTileForm.getTileWidth() - 1) << 10);
				
				column = ((getCollisionRight() + getSpeedX() - 1) >> 10) / GTileForm.getTileWidth();
				
				suggestedX = column * (GTileForm.getTileWidth() << 10) - (getCurrentCollisionWidth() << 9);
			} else {
				// moving left
				if (getSpeedX() < -((GTileForm.getTileWidth() - 1) << 10))
					setSpeedX(-((GTileForm.getTileWidth() - 1) << 10));
				
				column = ((getCollisionLeft() + getSpeedX()) >> 10) / GTileForm.getTileWidth();
				
				suggestedX = (column + 1) * (GTileForm.getTileWidth() << 10) + (getCurrentCollisionWidth() << 9);
			}
			
			topRow = (getCollisionTop() >> 10) / GTileForm.getTileHeight();
			bottomRow = ((getCollisionBottom() >> 10) - 1) / GTileForm.getTileHeight();
			
			bounce = false;
			if (column < cLevel.getWidth() && column >= 0) {
				if (topRow < 0) {
					// above level, trim search area
					topRow = 0;
				} else if (bottomRow >= cLevel.getHeight()) {
					// below level, trim search area
					bottomRow = cLevel.getHeight() - 1;
				}
				
				if (topRow <= bottomRow)
					for (y = topRow; y <= bottomRow; y++)
						if (column >= 0 && column < cLevel.getWidth() && y >= 0 && y < cLevel.getHeight())
							if (GTileForm.get(cLevel.getTile(column, y)).getSolid())
								bounce = true;
			}
			
			if (bounce) {
				setTouchingSolid(true);
				setLocationX(suggestedX);
				setSpeedX(-getSpeedX());
			} else
				setLocationX(getLocationX() + getSpeedX());
		}
		
		if (getSpeedY() != 0) {
			// vertical movement
			int						row;
			int						leftColumn;
			int						rightColumn;
			int						x;
			int						suggestedY;
			boolean					bounce;
			
			if (getSpeedY() > 0) {
				// moving down
				if (getSpeedY() > ((GTileForm.getTileHeight() - 1) << 10))
					setSpeedY((GTileForm.getTileHeight() - 1) << 10);
				
				row = ((getCollisionBottom() + getSpeedY() - 1) >> 10) / GTileForm.getTileHeight();
				
				suggestedY = row * (GTileForm.getTileHeight() << 10) - (getCurrentCollisionHeight() << 9);
			} else {
				// moving up
				if (getSpeedY() < -((GTileForm.getTileHeight() - 1) << 10))
					setSpeedY(-((GTileForm.getTileHeight() - 1) << 10));
				
				row = ((getCollisionTop() + getSpeedY()) >> 10) / GTileForm.getTileHeight();
				
				suggestedY = (row + 1) * (GTileForm.getTileHeight() << 10) + (getCurrentCollisionHeight() << 9);
			}
			
			leftColumn = (getCollisionLeft() >> 10) / GTileForm.getTileWidth();
			rightColumn = ((getCollisionRight() >> 10) - 1) / GTileForm.getTileWidth();
			
			bounce = false;
			if (row < cLevel.getHeight() && row >= 0) {
				if (leftColumn < 0) {
					// left of level, trim search area
					leftColumn = 0;
				} else if (rightColumn >= cLevel.getWidth()) {
					// right of level, trim search area
					rightColumn = cLevel.getWidth() - 1;
				}
				
				if (leftColumn <= rightColumn)
					for (x = leftColumn; x <= rightColumn; x++)
						if (x >= 0 && x < cLevel.getWidth() && row >= 0 && row < cLevel.getHeight())
							if (GTileForm.get(cLevel.getTile(x, row)).getSolid())
								bounce = true;
			}
			
			if (bounce) {
				setTouchingSolid(true);
				setLocationY(suggestedY);
				setSpeedY(-getSpeedY());
			} else
				setLocationY(getLocationY() + getSpeedY());
		}
	}
	
	public void preProcess(int frame) {
		if (frame == 0) {
			double					direction;
			
			// because properties have not been loaded at "constructor time",
			// properties-dependent initialization happen at the beginning of frame 0.
			
			cColor = getNumber(kColorProperty);
			direction = Math.random() * 2 * Math.PI;
			setSpeedX((int) (getNumber(kSpeedProperty) * Math.cos(direction) * 1024));
			setSpeedY((int) (getNumber(kSpeedProperty) * Math.sin(direction) * 1024));
		}
		
		cColor = (cColor + getNumber(kColorShiftProperty)) % kMaxColor;
	}
	
	public void postProcess(int frame) {
		if (cTouched) {
			// the player has picked this ball, remove it from the level
			cLevel.removeSprite(this);
		}
	}
	
	public void collideWith(GSprite other, int frame) {
		if (other instanceof BouncingPlayer) {
			// "picked" by the player sprite
			// we are not allowed to remove sprites during collideWith(),
			// so this ball is marked as touched, and removed at postProcess().
			cTouched = true;
		}
	}
}


