
import java.awt.*;


public class BouncingLevelCanvas extends GLevelCanvas
{
	public BouncingLevelCanvas(Component imageProducer, GLevel level, int width, int height) {
		super(imageProducer, level, width, height);
	}
	
	/*
	 * Updates the buffer, including new areas exposed by scrolling,
	 * tiles previously covered by sprites, animated tiles and
	 * all sprites.
	 
	 Adds support for drawing tiles twice, a bottom layer under the
	 sprites and a top layer above the sprites.
	 
	 */
	public void paint(Graphics finalGraphics, Graphics midGraphics, int frame) {
		int							breakX;
		int							breakY;
		boolean[][]					topNeedUpdate; // the part of tiles above the sprites
		
		
		if (!preparePaint())
			return;
		
		cFrame = frame;
		
		breakX = cScrollX % cBufferWidth;
		breakY = cScrollY % cBufferHeight;
		
		// fill the entire buffer with grey, helps seeing how much is updated
		// each frame
		if (cDebug)
			for (int index = 0; index < cBufferPixels.length; index++)
				cBufferPixels[index] = 0xFF7F7F7F;
		
		BouncingTile.cTopLayer = false;
		
		// if update area is significant, force update all
		if (Math.abs(cScrollX - cLastUpdateScrollX) * 3 > cBufferWidth * 2)
			cFreshBuffer = true;
		if (Math.abs(cScrollY - cLastUpdateScrollY) * 3 > cBufferHeight * 2)
			cFreshBuffer = true;
		
		topNeedUpdate = new boolean[cBufferTileWidth][cBufferTileHeight];
		
		if (cFreshBuffer) {
			paintChanged();
			cFreshBuffer = false;
			for (int x = 0; x < cBufferTileWidth; x++)
				for (int y = 0; y < cBufferTileHeight; y++)
					topNeedUpdate[x][y] = true;
		} else {
			// make sure animated tiles are updated
			if (cFrame != cLastUpdateFrame) {
				int					x;
				int					y;
				
				// this will unfortunately update an entire tile when part of it is animated
				for (x = cScrollX / cTileWidth; x <= (cScrollX / cTileWidth + cBufferTileWidth); x++)
					if (x >= 0 && x < cLevel.getWidth())
						for (y = cScrollY / cTileHeight; y <= (cScrollY / cTileHeight + cBufferTileHeight); y++)
							if (y >= 0 && y < cLevel.getHeight())
								if (GTileForm.get(cLevel.getTile(x, y)).getFlag(0) && BouncingTile.cAnimate)
									cNeedUpdate[x % cBufferTileWidth][y % cBufferTileHeight] = true;
			}
			
			// make sure tiles under sprites are updated this frame.
			// otherwise, the frame a sprite moved to a tile, the top
			// part of the tile would be drawn twice while the bottom part
			// only be drawn once, and transparent top tiles would end
			// up looking more opaque that frame.
			
			for (int spriteCounter = 0; spriteCounter < cLevel.getSpriteCount(); spriteCounter++) {
				GSprite				theSprite;
				int					x;
				int					y;
				
				theSprite = cLevel.getSprite(spriteCounter);
				if (theSprite == null) {
					// should never be here
				} else if ((theSprite.getRight() >> 10) < cScrollX ||
					(theSprite.getBottom() >> 10) < cScrollY ||
					(theSprite.getLeft() >> 10) > cScrollX + cBufferWidth ||
					(theSprite.getTop() >> 10) > cScrollY + cBufferHeight) {
					// sprite is not visible
				} else {
					// make sure the tiles behind the sprite get updated next frame
					for (x = (theSprite.getLeft() >> 10) / cTileWidth;
							x <= (((theSprite.getRight() >> 10) - 1) / cTileWidth); x++)
						for (y = (theSprite.getTop() >> 10) / cTileHeight;
								y <= (((theSprite.getBottom() >> 10) - 1) / cTileHeight); y++)
							if (x >= 0 && y >= 0)
								setNeedUpdate(x, y); // will trim to the buffer size
				}
			}
			
			if (cScrollX != cLastUpdateScrollX) {
				// update between any left and right panel
				paintVerticalUpdate();
			}
			
			if (cScrollY != cLastUpdateScrollY) {
				// update between any top and bottom panel
				paintHorizontalUpdate();
			}
			
			// make sure the top parts of tiles are updated if the bottom parts are
			for (int x = 0; x < cBufferTileWidth; x++)
				System.arraycopy(cNeedUpdate[x], 0, topNeedUpdate[x], 0, cBufferTileHeight);
			
			paintChanged();
		}
		
		if (cLevel.getSpriteCount() > 0)
			paintSprites();
		
		// draw top layer of tiles
		{
			boolean[][]				needUpdate;
			int						x;
			int						y;
			
			// store cNeedUpdate for next frame, topNeedUpdate keeps
			// track of which tiles need an update of the top part.
			needUpdate = cNeedUpdate;
			
			cNeedUpdate = topNeedUpdate;
			
			BouncingTile.cTopLayer = true;
			
			if (cScrollX != cLastUpdateScrollX) {
				// update between any left and right panel
				paintVerticalUpdate();
			}
			
			if (cScrollY != cLastUpdateScrollY) {
				// update between any top and bottom panel
				paintHorizontalUpdate();
			}
			
			paintChanged();
			
			cNeedUpdate = needUpdate;
		}
		
		cBufferSource.newPixels();
		cLastUpdateScrollX = cScrollX;
		cLastUpdateScrollY = cScrollY;
		cLastUpdateFrame = cFrame;
		
		copyBuffers(finalGraphics, midGraphics, breakX, breakY);
	}
}

