
public class BouncingPlayer extends GSprite
{
	// four pixels per frame = 100 pixels per second
	static final int				kSpeed = 4096;
	
	static final int				kMaxColor = 360;
	
	static final int[][]			kMod = new int[][] {
											{255, 255, 205},
											{225, 255, 225},
											{205, 255, 255},
											{225, 255, 255},
											{255, 205, 255},
											{255, 225, 225},
											{255, 255, 205}};
	
	// 0 = 360
	int								cColor;
	
	// allows communicating with the BouncingApplet, telling
	// it where to scroll
	static BouncingApplet			cApplet;
	
	// keeping track of how much the ball has been rolling
	int								cAnimationFrame;
	
	
	public BouncingPlayer(int formIndex) {
		super(formIndex);
		cAnimationFrame = 0;
	}
	
	/*
	Uses a single GPixels image to draw the different colors, by setting
	the cColorMod of BouncingPixels, which then uses the custom copyPixelsColor()
	method to draw the ball.
	*/
	public void draw(int[] canvas, int canvasWidth, int scrollX, int scrollY) {
		GPixels						image;
		
		int							redMod;
		int							greenMod;
		int							blueMod;
		
		// check if sprite is visible
		if ((getRight() >> 10) < scrollX ||
				(getBottom() >> 10) < scrollY ||
				(getLeft() >> 10) >= scrollX + canvasWidth ||
				(getTop() >> 10) >= scrollY + canvas.length / canvasWidth)
			return;
		
		redMod = (kMod[cColor * 6 / kMaxColor + 1][0] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][0] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		greenMod = (kMod[cColor * 6 / kMaxColor + 1][1] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][1] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		blueMod = (kMod[cColor * 6 / kMaxColor + 1][2] * (cColor % (kMaxColor / 6)) +
				kMod[cColor * 6 / kMaxColor][2] * ((kMaxColor - cColor - 1) % (kMaxColor / 6))) / (kMaxColor / 6);
		
		image = cForm.getImage(cCurrentImage);
		((BouncingPixels) image).setColorMod((redMod << 16) + (greenMod << 8) + blueMod);
		image.drawWrapped(canvas, canvasWidth, getLeft() >> 10, getTop() >> 10,
			cCurrentImageMode, scrollX, scrollY);
		((BouncingPixels) image).setColorMod(0xFFFFFFFF);
	}
	
	public void preProcess(int frame) {
		// handle movement
		if (cApplet.isKeyActive(cApplet.kLeft))
			setSpeedX(-kSpeed);
		else if (cApplet.isKeyActive(cApplet.kRight))
			setSpeedX(kSpeed);
		else
			setSpeedX(0);
		
		if (cApplet.isKeyActive(cApplet.kUp))
			setSpeedY(-kSpeed);
		else if (cApplet.isKeyActive(cApplet.kDown))
			setSpeedY(kSpeed);
		else
			setSpeedY(0);
		
		if (frame % 2 == 0) {
			// update animation
			cAnimationFrame++;
			
			if (cAnimationFrame % 4 == 0)
				cCurrentImage = 0;
			else if (cAnimationFrame % 4 == 2)
				cCurrentImage = 2;
			else
				cCurrentImage = 1;
			
			if (cAnimationFrame % 4 == 3)
				cCurrentImageMode = 4;
			else
				cCurrentImageMode = 0;
			
			if (cAnimationFrame % 8 >= 4)
				cCurrentImageMode = GPixels.getRotateMode(cCurrentImageMode, 1);
		}
		
		cColor = (cColor + 4) % kMaxColor;
	}
	
	public void postProcess(int frame) {
		// tell the applet where to scroll
		cApplet.cTargetCenterX = cLocationX >> 10;
		cApplet.cTargetCenterY = cLocationY >> 10;
		
		if (cLevel.getSpriteCount() == 1 &&
				BouncingTile.cAnimate) {
			// all the balls have been picked.
			// stop the flashing bricks
			BouncingTile.cAnimate = false;
			
			// make sure we update all tiles
			cApplet.cCanvas.setNeedUpdate(true);
		}
	}
	
	public void collideWith(GSprite other, int frame) {
		// get points for catching a bouncing ball
		if (other instanceof BouncingBall)
			cApplet.cPoints += 10;
	}
}


