
public class SnakeBody extends GSprite
{
	static SnakeIIApplet			cApplet;
	
	static int[]					sMode = new int[] {4, 0, 0, 7, 7, 1, 1, 6, 6, 2, 2, 5, 5, 3, 3, 4, 4};
	static int[]					sImage = new int[] {0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0};
	
	static final int				kImageBase = 0;
	static final int				kTailImageBase = 2;
	static final int				kBoneImageBase = 4;
	
	GSprite							cLeader;
	int								cAngle;
	boolean							cWiggle;
	boolean							cWaitForNextLevel;
	boolean							cTail;
	int								cFlesh;
	
	
	public SnakeBody(int formIndex) {
		super(formIndex);
		
		cFlesh = 255;
		cWiggle = false;
		cWaitForNextLevel = false;
		cAngle = 4;
	}
	
	public void draw(int[] canvas, int canvasWidth, int scrollX, int scrollY) {
		if (cFlesh == 255)
			// flesh
			super.draw(canvas, canvasWidth, scrollX, scrollY);
		else if (cFlesh == 0) {
			// bone
			cCurrentImage += kBoneImageBase;
			super.draw(canvas, canvasWidth, scrollX, scrollY);
			cCurrentImage -= kBoneImageBase;
		} else {
			// part flesh, part bone
			GPixels					image;
			
			// check if sprite is visible
			if ((cLocationX >> 10) + getCurrentWidth() < scrollX ||
					(cLocationY >> 10) + getCurrentHeight() < scrollY ||
					(cLocationX >> 10) >= scrollX + canvasWidth ||
					(cLocationY >> 10) >= scrollY + canvas.length / canvasWidth)
				return;
			
			// draw flesh partially transparent
			image = cForm.getImage(cCurrentImage);
			
			if (image != null) {
				image.setAlphaMod(cFlesh);
				
				image.drawWrapped(canvas, canvasWidth, getLeft() >> 10, getTop() >> 10,
					cCurrentImageMode, scrollX, scrollY);
				image.setAlphaMod(255);
			}
			
			// draw bone partially transparent
			image = cForm.getImage(cCurrentImage + kBoneImageBase);
			
			if (image != null) {
				image.setAlphaMod(255 - cFlesh);
				
				image.drawWrapped(canvas, canvasWidth, getLeft() >> 10, getTop() >> 10,
					cCurrentImageMode, scrollX, scrollY);
				image.setAlphaMod(255);
			}
		}
	}
	
	public void waitForNextLevel() {
		if (cLeader instanceof SnakeBody)
			((SnakeBody) cLeader).waitForNextLevel();
		cWaitForNextLevel = true;
	}
	
	// from head to tail
	public boolean decaySegment() {
		if (cFlesh == 0) {
			if (cLeader instanceof SnakeBody)
				((SnakeBody) cLeader).decaySegment();
			
			return true;
		} else {
			if (cLeader instanceof SnakeBody) {
				// body
				if (((SnakeBody) cLeader).decaySegment()) {
					cFlesh -= 35;
					if (cFlesh < 0)
						cFlesh = 0;
				}
			} else {
				// leader == head, decay this
				cFlesh -= 35;
				if (cFlesh < 0)
					cFlesh = 0;
			}
			
			if (cFlesh < 180)
				return true;
			else
				return false;
		}
	}
	
	// flesh mode only, from tail to head
	public void removeFlesh() {
		if (cFlesh == 255) {
			cFlesh -= 35;
			return;
		} else {
			int							leaderFlesh;
			
			if (cLeader instanceof SnakeBody) {
				((SnakeBody) cLeader).removeFlesh();
				leaderFlesh = ((SnakeBody) cLeader).cFlesh;
			} else {
				((SnakeHead) cLeader).removeFlesh();
				leaderFlesh = ((SnakeHead) cLeader).cFlesh;
			}
			
			if (cFlesh > leaderFlesh - 35)
				cFlesh = leaderFlesh - 35;
			if (cFlesh < 0)
				cFlesh = 0;
		}
	}
	
	// flesh mode only, from tail to head
	// returns suggested action for follower
	public boolean addFlesh() {
		if (cFlesh == 255)
			return true;
		
		if (cLeader instanceof SnakeBody) {
			// body
			if (((SnakeBody) cLeader).addFlesh()) {
				cFlesh += 35;
				if (cFlesh > 255)
					cFlesh = 255;
			}
		} else {
			// head; if snake is alive, head is flesh
			cFlesh += 35;
			if (cFlesh > 255)
				cFlesh = 255;
		}
		
		return cFlesh > 75;
	}
	
	public void updateImage() {
		int							angle;
		
		if (cWiggle)
			angle = cAngle + 1;
		else
			angle = cAngle;
		
		if (cTail)
			setCurrentImage(sImage[angle] + kTailImageBase);
		else
			setCurrentImage(sImage[angle] + kImageBase);
		
		setCurrentImageMode(sMode[angle]);
	}
	
	public void preProcess(int frame) {
		if (frame % 3 != 0)
			return;
		
		updateImage();
	}
	
	public void postProcess(int frame) {
		if (frame % 3 == 0 && !cApplet.isSnakeDead() && !cWaitForNextLevel)
			cWiggle = !cWiggle;
	}
}

