/***************************************************************

Snake II shows how you can use the GLevelCanvas' buffer for a
title screen with a progress bar when loading the data file. It
also shows how to use multiple levels, how to map a single
GSprite subclass to several GSpriteForm objects (the bugs), and
how to create sprites both in the start of a game (the snake),
and during a game (bugs).


The source for Snake II is in the public domain. This means you
can use it in your own projects, whole or in part. Please note
that GameForm is not in the public domain, see the GameForm
source, documentation, or the GameForm homepage for information
about the allowed and disallowed usage of GameForm.


GameForm home page:
	http://home.swipnet.se/carlin/gameform/

E-mail address to Bjorn Carlin:
	bjorn.carlin@swipnet.se

Snail mail address to Bjorn Carlin:
	Bjorn Carlin
	Gnejsvagen 18 B
	752 42 Uppsala
	Sweden

***************************************************************/


import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;


public class SnakeIIApplet extends GameApplet implements GLoadListener, MouseListener, MouseMotionListener
{
	// keys
	static final int				kPlayerLeft = 0;
	static final int				kPlayerRight = 1;
	
	static final int				kSegments = 14; // even numbers work best
	
	static final int				kPreLoading = 0;
	static final int				kLoading = 1;
	static final int				kPriming = 2;
	static final int				kFailedLoading = 3;
	static final int				kDoneLoading = 4;
	
	static final int				kGameModeCount = 2;
	static final int				kNormalMode = 0;
	static final int				kFleshMode = 1;
	
	static final int				kFirstBug = 2;
	static final int				kBugTypeCount = 4;
	
	static final int				kFileSize = 63795; // as reported by GEditor
	static final int				kFadeFrameCount = 25;
	static final int				kNoShowFrameCount = 23;
	
	static final String				kDefaultDescription =
			"To start a new game, press one of the buttons to the left.\nUse the left and right arrow keys to control the snake.\nReturn of the Snake 2 is Copyright  2000-2001, Bjrn Carlin. All rights reserved.";
	static final String[]			kDescription = new String[] {
			"Eat all the bugs in each level to continue to the next. Running into obstacles will kill the snake.\nA High Score is awarded to the highest level reached, and the most bugs eaten on that level.",
			"Try to stay alive by eating as many bugs as possible as fast as you can.\nThe snake will lose flesh continously, and only regain it when eating a bug. New bugs will appear when the old ones are consumed."};
	
	GLevelCanvas					cCanvas;
	SnakeHead						cHeadSprite;
	
	String							cLoadingMessage;
	int								cLoadingStatus;
	int								cLoadingBits;
	
	Image							cBuffer; // shared with cCanvas
	Graphics						cBufferGraphics;
	
	GLevel							cLevel;
	int								cLevelIndex;
	int								cCompleteLevelFrame;
	int								cRestartLevelFrame;
	int								cEndGameFrame;
	int								cGameMode;
	int								cStartFadeFrame;
	
	int								cTargetCenterX;
	int								cTargetCenterY;
	int								cScrollDeltaX;
	int								cScrollDeltaY;
	
	int								cCreateCanvasFrame;
	int								cDeciseconds;
	
	// high score
	boolean							cCheater;
	int								cNormalBestLevel;
	int								cNormalMostBugs;
	int								cFleshBestDeciseconds;
	int								cFleshMostBugs;
	
	// last score
	int								cNormalLastLevel;
	int								cNormalLastBugs;
	int								cFleshLastDeciseconds;
	int								cFleshLastBugs;
	
	// layout
	boolean[]						cModeButtonPressed = new boolean[kGameModeCount];
	boolean[]						cModeButtonHilited = new boolean[kGameModeCount];
	Rectangle[]						cModeButtonBounds = new Rectangle[kGameModeCount];
	
	Font							cSmallFont;
	Font							cLargeFont;
	
	boolean							cSaveHighScore;
	
	
	public void init() {
		super.init(kNoBuffer);
		
		cLevel = null;
		cCanvas = null;
		
		SnakeHead.cApplet = this;
		SnakeBody.cApplet = this;
		SnakeBug.cApplet = this;
		
		GFactory.setFactory(new SnakeFactory());
		GOutput.sMute = true;
		
		setKeys(5, 3,
				new int[] {KeyEvent.VK_LEFT,
							KeyEvent.VK_RIGHT},
				new int[] {kKeyContinous,
							kKeyContinous});
		
		setCheats(new int[][] {
			{KeyEvent.VK_W, KeyEvent.VK_R, KeyEvent.VK_Y},
			{KeyEvent.VK_B, KeyEvent.VK_U, KeyEvent.VK_G}});
		
		// init double buffer
		cCanvas = new GLevelCanvas(this, null, getSize().width, getSize().height);
		cBuffer = cCanvas.getBuffer();
		cBufferGraphics = cBuffer.getGraphics();
		addMouseListener(this);
		addMouseMotionListener(this);
		
		cLoadingStatus = kPreLoading;
		cLoadingMessage = "Looking for Data File...";
		cLoadingBits = 0;
		
		resetHighScore();
		
		cSmallFont = new Font("SansSerif", Font.PLAIN, 10);
		cLargeFont = new Font("Serif", Font.PLAIN, 24);
		
		for (int buttonCounter = 0; buttonCounter < kGameModeCount; buttonCounter++) {
			cModeButtonPressed[buttonCounter] = false;
			cModeButtonHilited[buttonCounter] = false;
			cModeButtonBounds[buttonCounter] = new Rectangle(15, 75 + (getHeight() - 75) * buttonCounter / kGameModeCount, 100, (getHeight() - 75) / kGameModeCount - 15);
		}
		
		cFrameLength = 67; // 15 fps
	}
	
	public void resetHighScore() {
		cNormalBestLevel = 0;
		cNormalMostBugs = 0;
		cFleshBestDeciseconds = 0;
		cFleshMostBugs = 0;
		
		cNormalLastLevel = -1;
		cNormalLastBugs = -1;
		cFleshLastDeciseconds = -1;
		cFleshLastBugs = -1;
	}
	
	public int getWidth() {
		return getSize().width;
	}
	
	public int getHeight() {
		return getSize().height;
	}
	
	public boolean isSnakeDead() {
		if (cHeadSprite == null)
			return false;
		else
			return cHeadSprite.cDead;
	}
	
	public int getGameMode() {
		return cGameMode;
	}
	
	public void createCanvas() {
		int							x;
		int							y;
		GSprite[]					snake;
		int							segmentCounter;
		int							bugs;
		
		cCreateCanvasFrame = cFrame;
		cCompleteLevelFrame = cFrame - 1;
		cRestartLevelFrame = cFrame - 1;
		cEndGameFrame = cFrame - 1;
		cStartFadeFrame = cFrame - kFadeFrameCount * 2;
		
		if (cGameMode == kNormalMode) {
			// levels past last will roll over
			cLevel = GFactory.createLevel(GLevel.get(cLevelIndex % (GLevel.getCount() - 1))); // get a deep clone
		} else
			cLevel = GFactory.createLevel(GLevel.get(cLevelIndex)); // get a deep clone
		bugs = cLevel.getSpriteCount();
		
		// make snake
		x = 40 * 1024;
		y = 40 * 1024;
		snake = new GSprite[kSegments];
		
		for (segmentCounter = 0; segmentCounter < kSegments - 1; segmentCounter++) {
			snake[segmentCounter] = cLevel.addSprite(x, y, 1);
			((SnakeBody) snake[segmentCounter]).cWiggle = (segmentCounter % 2 == 0);
			((SnakeBody) snake[segmentCounter]).updateImage();
		}
		
		((SnakeBody) snake[0]).cTail = true;
		snake[kSegments - 1] = cLevel.addSprite(x + 13 * 1024, y, 0);
		
		for (segmentCounter = 0; segmentCounter < kSegments - 1; segmentCounter++)
			((SnakeBody) snake[segmentCounter]).cLeader = snake[segmentCounter + 1];
		
		cHeadSprite = (SnakeHead) snake[kSegments - 1];
		
		cHeadSprite.cSecondHead = (SnakeBody) snake[kSegments - 2];
		cHeadSprite.cSecondTail = (SnakeBody) snake[1];
		cHeadSprite.cTail = (SnakeBody) snake[0];
		cHeadSprite.cBugs = bugs;
		cHeadSprite.updateImage();
		
		for (int i = 0; i < kNoShowFrameCount; i++)
			cLevel.nextFrame(++cFrame);
		
		// fool internal time keeper
		cFirstFrameTime = System.currentTimeMillis() - cFrame * cFrameLength;
		
		cCanvas.setLevel(cLevel);
		
		cBuffer = cCanvas.getBuffer();
		cBufferGraphics = cBuffer.getGraphics();
		
		// reset scrolling
		cTargetCenterX = 0;
		cTargetCenterY = 0;
		cScrollDeltaX = 0;
		cScrollDeltaY = 0;
	}
	
	public synchronized void advanceFrame() {
		int							targetScrollX;
		int							targetScrollY;
		
		if (cFrame == cCompleteLevelFrame) {
			cLevelIndex++;
			if (cGameMode == kNormalMode)
				SnakeHead.sTotalBugs = 0;
			createCanvas();
		} else if (cFrame == cRestartLevelFrame) {
			createCanvas();
		} else if (cFrame == cEndGameFrame) {
			cEndGameFrame = cFrame - 1;
			endGame();
			return;
		}
		
		super.advanceFrame();
		
		if (cLevel == null)
			return;
		
		cLevel.nextFrame(cFrame);
		
		if (cFrame % 3 == 0) {
			targetScrollX = cTargetCenterX - getWidth() / 2;
			targetScrollY = cTargetCenterY - getHeight() / 2;
			cScrollDeltaX = (targetScrollX - cCanvas.getScrollX()) / 3;
			cScrollDeltaY = (targetScrollY - cCanvas.getScrollY()) / 3;
		}
		
		cCanvas.setScrollX(cCanvas.getScrollX() + cScrollDeltaX);
		cCanvas.setScrollY(cCanvas.getScrollY() + cScrollDeltaY);
	}
	
	public void advanceLevel(int delay) {
		cStartFadeFrame = cFrame + delay;
		cCompleteLevelFrame = cStartFadeFrame + kFadeFrameCount + 2;
	}
	
	public void restartLevel(int delay) {
		cStartFadeFrame = cFrame + delay;
		cRestartLevelFrame = cStartFadeFrame + kFadeFrameCount + 2;
	}
	
	public void endLevel(int delay) {
		cStartFadeFrame = cFrame + delay;
		cEndGameFrame = cStartFadeFrame + kFadeFrameCount + 2;
	}
	
	public int getFrame() {
		return cFrame;
	}
	
	public void spawnBug() {
		// assumes bug is smaller than tile
		
		int							suggestedX = 3; // tile
		int							suggestedY = 3; // tile
		boolean						goodPlace;
		int							attemptCount;
		
		goodPlace = false;
		attemptCount = 0;
		while (!goodPlace && attemptCount < 30) {
			attemptCount++;
			
			suggestedX = (int) (Math.random() * cLevel.getWidth());
			suggestedY = (int) (Math.random() * cLevel.getHeight());
			
			// make sure tile is not solid
			if (GTileForm.get(cLevel.getTile(suggestedX, suggestedY)).getSolid())
				goodPlace = false;
			else {
				// make sure location is not visible
				
				// convert location to pixels
				suggestedX = suggestedX * GTileForm.getTileWidth() + GTileForm.getTileWidth() / 2;
				suggestedY = suggestedY * GTileForm.getTileHeight() + GTileForm.getTileHeight() / 2;
				
				if ((suggestedX - cCanvas.getScrollX() < -8) ||
						(suggestedY - cCanvas.getScrollY() < -8) ||
						(suggestedX - cCanvas.getScrollX() > getWidth() + 8) ||
						(suggestedY - cCanvas.getScrollY() > getHeight() + 8))
					goodPlace = true;
				else
					goodPlace = false;
			}
		}
		
		cLevel.addSprite(suggestedX << 10, suggestedY << 10,
			(int) (kFirstBug + kBugTypeCount * Math.random()));
	}
	
	public void pauseGame(boolean pause) {
		super.pauseGame(pause);
	}
	
	public void newGame(int mode) {
		if (cLoadingStatus != kDoneLoading)
			return;
		
		if (cGameInProgress)
			endGame();
		
		cGameMode = mode;
		
		cCheater = false;
		if (cGameMode == kNormalMode) {
			SnakeHead.sLives = 5;
			cLevelIndex = 0;
		} else if (cGameMode == kFleshMode) {
			// last level is flesh level
			cLevelIndex = GLevel.getCount() - 1;
		}
		
		SnakeHead.sTotalBugs = 0;
		cFrame = -1;
		createCanvas();
		
		cPauseTime = 0;
		
		cGameInProgress = true;
		cGamePaused = false;
		cGameOver = false;
	}
	
	public void endGame() {
		if (!cGameInProgress)
			return;
		
		super.endGame();
		
		if (cGameMode == kNormalMode) {
			cNormalLastLevel = cLevelIndex;
			cNormalLastBugs = SnakeHead.sTotalBugs;
			
			if (!cCheater)
				if (cLevelIndex > cNormalBestLevel ||
						cLevelIndex == cNormalBestLevel && SnakeHead.sTotalBugs > cNormalMostBugs) {
					cNormalBestLevel = cLevelIndex;
					cNormalMostBugs = SnakeHead.sTotalBugs;
				}
		} else if (cGameMode == kFleshMode) {
			cFleshLastDeciseconds = cDeciseconds;
			cFleshLastBugs = SnakeHead.sTotalBugs;
			
			if (!cCheater)
				if (cDeciseconds > cFleshBestDeciseconds) {
					cFleshBestDeciseconds = cDeciseconds;
					cFleshMostBugs = SnakeHead.sTotalBugs;
				}
		}
	}
	
	protected void cheatActivated(int cheatID) {
		if (cGameMode == kNormalMode && cheatID == 0)
			advanceLevel(1 - kFadeFrameCount);
		if (cGameMode == kFleshMode && cheatID == 1)
			for (int b = 0; b < 10; b++)
				spawnBug();
		cCheater = true;
	}
	
	public void startLoading() {
		try {
			GFileManager.load(new URL(getCodeBase(), getParameter("file")), this);
		} catch (Exception e) {
			// malformed url etc.
			cLoadingStatus = kFailedLoading;
			repaint();
		}
	}
	
	public void newProgress(int bitsRead) {
		if (bitsRead == GFileManager.kFinished) {
			// priming the JIT
			cLoadingStatus = kPriming;
			repaint();
			Thread.yield();
			cLevelIndex = 0;
			createCanvas();
			cLevel.nextFrame(1);
			cLevel.nextFrame(2);
			repaint();
			Thread.yield();
			cCanvas.paint(null, 2);
			repaint();
			Thread.yield();
			cLevel.nextFrame(3);
			cLevel.nextFrame(4);
			cCanvas.paint(null, 4);
			cLoadingStatus = kDoneLoading;
		} else if (bitsRead == GFileManager.kFailed)
			cLoadingStatus = kFailedLoading;
		
		cLoadingMessage = "Loading...";
		cLoadingBits = bitsRead;
		repaint();
	}
	
	public void drawShadowedString(String text, int x, int y) {
		cBufferGraphics.setColor(Color.black);
		cBufferGraphics.drawString(text, x + 1, y + 1);
		cBufferGraphics.setColor(Color.white);
		cBufferGraphics.drawString(text, x, y);
	}
	
	// returns suggested Y for next string
	public int drawWrappedString(String text, int x, int y, int width) {
		int							currentRowY;
		int							currentRowWidth;
		String						currentRow;
		String						nextToken;
		StringTokenizer				tokenizer;
		FontMetrics					metrics;
		
		if (text.length() == 0)
			return y;
		
		currentRowY = y;
		tokenizer = new StringTokenizer(text, " ");
		metrics = cBufferGraphics.getFontMetrics();
		nextToken = tokenizer.nextToken(); // prepare first row
		
		while (nextToken != null) {
			currentRowWidth = 0;
			currentRow = "";
			while (currentRowWidth + metrics.stringWidth(nextToken) < width) {
				currentRow += " " + nextToken;
				currentRowWidth = metrics.stringWidth(currentRow);
				
				if (tokenizer.hasMoreTokens())
					nextToken = tokenizer.nextToken();
				else {
					nextToken = null;
					break;
				}
			}
			
			cBufferGraphics.drawString(currentRow, x, currentRowY);
			currentRowY += metrics.getHeight();
		}
		
		return currentRowY;
	}
	
	public int drawWrappedStringLines(String text, int x, int y, int width) {
		int							currentRowY;
		StringTokenizer				tokenizer;
		
		if (text.length() == 0)
			return y;
		
		currentRowY = y;
		tokenizer = new StringTokenizer(text, "\n");
		while (tokenizer.hasMoreTokens())
			currentRowY = 5 + drawWrappedString(tokenizer.nextToken(), x, currentRowY, width);
		
		return currentRowY;
	}
	
	public void updateTitleScreen() {
		// title screen
		int							shownDescription;
		String						description;
		int							currentColumn;
		int							currentRow;
		
		// clear
		cBufferGraphics.setColor(Color.black);
		cBufferGraphics.fillRect(0, 0, getWidth(), getHeight());
		
		// draw title
		// figure out horizontal center of window
		cBufferGraphics.setFont(new Font("Serif", Font.PLAIN, 24));
		currentColumn = cBufferGraphics.getFontMetrics().stringWidth("Return ");
		cBufferGraphics.setFont(new Font("Serif", Font.PLAIN, 48));
		currentColumn += cBufferGraphics.getFontMetrics().stringWidth("Snake 2");
		currentColumn = (getWidth() - currentColumn) / 2;
		
		cBufferGraphics.setColor(Color.white);
		
		cBufferGraphics.setFont(new Font("Serif", Font.PLAIN, 24));
		cBufferGraphics.drawString("Return", currentColumn, 5 + cBufferGraphics.getFontMetrics().getHeight());
		cBufferGraphics.drawString("of the", currentColumn, 5 + cBufferGraphics.getFontMetrics().getHeight() * 2);
		currentColumn += cBufferGraphics.getFontMetrics().stringWidth("Return ");
		
		cBufferGraphics.setFont(new Font("Serif", Font.PLAIN, 48));
		currentRow = 5 + cBufferGraphics.getFontMetrics().getHeight();
		cBufferGraphics.drawString("Snake 2", currentColumn, currentRow);
		currentColumn += cBufferGraphics.getFontMetrics().stringWidth("Snake 2");
		
		cBufferGraphics.setFont(new Font("Serif", Font.PLAIN, 14));
		cBufferGraphics.drawString("by Bjrn Carlin", getWidth() - 15 - cBufferGraphics.getFontMetrics().stringWidth("by Bjrn Carlin"), currentRow + cBufferGraphics.getFontMetrics().getHeight());
		currentRow += cBufferGraphics.getFontMetrics().getHeight();
		
		// new game buttons
		shownDescription = -1;
		cBufferGraphics.setFont(cSmallFont);
		for (int buttonCounter = 0; buttonCounter < kGameModeCount; buttonCounter++) {
			Color					backColor;
			Color					highColor;
			Color					lowColor;
			Color					textColor;
			int						left;
			int						baseLine;
			
			if (cLoadingStatus != kDoneLoading) {
				highColor = new Color(75, 75, 75);
				backColor = new Color(37, 37, 37);
				lowColor = new Color(25, 25, 25);
				textColor = new Color(93, 93, 93);
				
				if (cModeButtonPressed[buttonCounter] && cModeButtonHilited[buttonCounter])
					shownDescription = buttonCounter;
			} else if (cModeButtonPressed[buttonCounter] && cModeButtonHilited[buttonCounter]) {
				highColor = new Color(15, 18, 45);
				backColor = new Color(20, 25, 60);
				lowColor = new Color(40, 50, 120);
				textColor = new Color(50, 62, 150);
				shownDescription = buttonCounter;
			} else {
				highColor = new Color(60, 75, 180);
				backColor = new Color(30, 37, 90);
				lowColor = new Color(20, 25, 60);
				textColor = new Color(75, 93, 225);
			}
			
			cBufferGraphics.setColor(backColor);
			
			cBufferGraphics.fillRect(cModeButtonBounds[buttonCounter].x + 3, 
					cModeButtonBounds[buttonCounter].y + 3,
					cModeButtonBounds[buttonCounter].width - 6,
					cModeButtonBounds[buttonCounter].height - 6);
			
			cBufferGraphics.setColor(highColor);
			cBufferGraphics.drawLine(cModeButtonBounds[buttonCounter].x + 3, 
					cModeButtonBounds[buttonCounter].y + 3,
					cModeButtonBounds[buttonCounter].x + 3,
					cModeButtonBounds[buttonCounter].y + cModeButtonBounds[buttonCounter].height - 3);
			cBufferGraphics.drawLine(cModeButtonBounds[buttonCounter].x + 3, 
					cModeButtonBounds[buttonCounter].y + 3,
					cModeButtonBounds[buttonCounter].x + cModeButtonBounds[buttonCounter].width - 3,
					cModeButtonBounds[buttonCounter].y + 3);
			
			cBufferGraphics.setColor(lowColor);
			cBufferGraphics.drawLine(cModeButtonBounds[buttonCounter].x + 3, 
					cModeButtonBounds[buttonCounter].y + cModeButtonBounds[buttonCounter].height - 3,
					cModeButtonBounds[buttonCounter].x + cModeButtonBounds[buttonCounter].width - 3,
					cModeButtonBounds[buttonCounter].y + cModeButtonBounds[buttonCounter].height - 3);
			cBufferGraphics.drawLine(cModeButtonBounds[buttonCounter].x + cModeButtonBounds[buttonCounter].width - 3, 
					cModeButtonBounds[buttonCounter].y + 3,
					cModeButtonBounds[buttonCounter].x + cModeButtonBounds[buttonCounter].width - 3,
					cModeButtonBounds[buttonCounter].y + cModeButtonBounds[buttonCounter].height - 3);
			
			if (cModeButtonHilited[buttonCounter])
				if (shownDescription < 0)
					shownDescription = buttonCounter;
			
			cBufferGraphics.setFont(new Font("SansSerif", Font.BOLD, 14));
			
			// new game
			if (cModeButtonPressed[buttonCounter] && cModeButtonHilited[buttonCounter])
				cBufferGraphics.setColor(highColor);
			else
				cBufferGraphics.setColor(lowColor);
			
			left = cModeButtonBounds[buttonCounter].x + (cModeButtonBounds[buttonCounter].width - cBufferGraphics.getFontMetrics().stringWidth("New Game")) / 2;
			baseLine = cModeButtonBounds[buttonCounter].y + (cModeButtonBounds[buttonCounter].height - 5) / 2 - cBufferGraphics.getFontMetrics().getDescent();
			
			cBufferGraphics.drawString("New game", left, baseLine + 1);
			cBufferGraphics.drawString("New game", left - 1, baseLine);
			cBufferGraphics.drawString("New game", left, baseLine - 1);
			cBufferGraphics.drawString("New game", left + 1, baseLine);
			
			cBufferGraphics.setColor(textColor);
			cBufferGraphics.drawString("New game", left, baseLine);
			
			// mode
			if (cModeButtonPressed[buttonCounter] && cModeButtonHilited[buttonCounter])
				cBufferGraphics.setColor(highColor);
			else
				cBufferGraphics.setColor(lowColor);
			
			left = cModeButtonBounds[buttonCounter].x + (cModeButtonBounds[buttonCounter].width - cBufferGraphics.getFontMetrics().stringWidth("Mode " + (buttonCounter + 1))) / 2;
			baseLine = cModeButtonBounds[buttonCounter].y + (cModeButtonBounds[buttonCounter].height + 5) / 2 + cBufferGraphics.getFontMetrics().getHeight() - cBufferGraphics.getFontMetrics().getDescent();
			
			cBufferGraphics.drawString("Mode " + (buttonCounter + 1), left, baseLine + 1);
			cBufferGraphics.drawString("Mode " + (buttonCounter + 1), left - 1, baseLine);
			cBufferGraphics.drawString("Mode " + (buttonCounter + 1), left, baseLine - 1);
			cBufferGraphics.drawString("Mode " + (buttonCounter + 1), left + 1, baseLine);
			
			if (cModeButtonPressed[buttonCounter] && cModeButtonHilited[buttonCounter])
				cBufferGraphics.setColor(lowColor);
			else
				cBufferGraphics.setColor(highColor);
			cBufferGraphics.drawString("Mode " + (buttonCounter + 1), left, baseLine);
		}
		
		// description frame
		cBufferGraphics.setColor(new Color(60, 90, 30));
		
		currentColumn = cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width;
		currentRow += cBufferGraphics.getFontMetrics().getHeight();
		cBufferGraphics.drawRect(currentColumn, currentRow,
			getWidth() - 15 - currentColumn, getHeight() - 15 - currentRow);
		
		if (shownDescription >= 0) {
			int						connectionY;
			
			connectionY = cModeButtonBounds[shownDescription].y + 
				cModeButtonBounds[shownDescription].height / 2;
			
			// around button
			cBufferGraphics.drawRect(cModeButtonBounds[shownDescription].x,
				cModeButtonBounds[shownDescription].y,
				cModeButtonBounds[shownDescription].width - 1,
				cModeButtonBounds[shownDescription].height - 1);
			
			// tube endings
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.drawLine(currentColumn - cModeButtonBounds[0].x - 1, connectionY - 2,
					currentColumn - cModeButtonBounds[0].x - 1, connectionY + 2);
			cBufferGraphics.drawLine(currentColumn, connectionY + 2,
					currentColumn, connectionY - 2);
			
			// tube
			cBufferGraphics.setColor(new Color(60, 90, 30));
			cBufferGraphics.drawLine(currentColumn - cModeButtonBounds[0].x - 1, connectionY - 2,
					currentColumn, connectionY - 2);
			cBufferGraphics.drawLine(currentColumn - cModeButtonBounds[0].x - 1, connectionY + 2,
					currentColumn, connectionY + 2);
		}
		
		if (shownDescription == -1 && cLoadingStatus == kLoading) {
			int							left;
			int							right;
			
			cBufferGraphics.setFont(cSmallFont);
			cBufferGraphics.setColor(new Color(175, 250, 100));
			
			left = cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 5;
			right = getWidth() - 20;
			
			drawWrappedString("Please Wait, " + cLoadingMessage,
				left + 5,
				currentRow + cBufferGraphics.getFontMetrics().getHeight() * 2,
				right - left - 10);
			
			currentRow += 4 * cBufferGraphics.getFontMetrics().getHeight();
			
			// progress bar
			cBufferGraphics.setColor(new Color(100, 150, 50));
			cBufferGraphics.fillRect(left,
				currentRow - 10,
				cLoadingBits * (right - left) / kFileSize,
				10);
			cBufferGraphics.setColor(new Color(60, 90, 30));
			cBufferGraphics.drawRect(left,
				currentRow - 10,
				right - left,
				10);
		} else if (shownDescription == -1 && cLoadingStatus == kPriming) {
			cBufferGraphics.setFont(cSmallFont);
			cBufferGraphics.setColor(new Color(175, 250, 100));
			
			cBufferGraphics.drawString("Almost done...",
				cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 10,
				currentRow + cBufferGraphics.getFontMetrics().getHeight() * 2);
		} else {
			// description
			cBufferGraphics.setFont(cSmallFont);
			cBufferGraphics.setColor(new Color(175, 250, 100));
			if (shownDescription == -1)
				description = kDefaultDescription;
			else
				description = kDescription[shownDescription];
			
			drawWrappedStringLines(description,
				cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width,
				currentRow + cBufferGraphics.getFontMetrics().getHeight(),
				getWidth() - cModeButtonBounds[0].x * 2 - cModeButtonBounds[0].width - 18);
			
			// high score
			if (shownDescription == kNormalMode) {
				String					text;
				
				if (cNormalMostBugs == 1)
					text = "High Score: 1 bug on level " + (cNormalBestLevel + 1);
				else
					text = "High Score: " + cNormalMostBugs + " bugs on level " + (cNormalBestLevel + 1);
				
				cBufferGraphics.drawString(text, cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 3, getHeight() - 20);
			} else if (shownDescription == kFleshMode) {
				String					text;
				String					time;
				
				time = (cFleshBestDeciseconds / 600) + ":" + ((cFleshBestDeciseconds / 100) % 6) + "" + ((cFleshBestDeciseconds / 10) % 10) + "." + (cFleshBestDeciseconds % 10);
				if (cFleshMostBugs == 1)
					text = "High Score: 1 bug in " + time;
				else
					text = "High Score: " + cFleshMostBugs + " bugs in " + time;
				
				cBufferGraphics.drawString(text,
					cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 3,
					getHeight() - 20);
			}
			
			// last score
			if (shownDescription == kNormalMode && cNormalLastLevel >= 0) {
				String					text;
				
				if (cNormalLastBugs == 1)
					text = "Last Score: 1 bug on level " + (cNormalLastLevel + 1);
				else
					text = "Last Score: " + cNormalLastBugs + " bugs on level " + (cNormalLastLevel + 1);
				
				cBufferGraphics.drawString(text, cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 3, getHeight() - 20 - cBufferGraphics.getFontMetrics().getHeight());
			} else if (shownDescription == kFleshMode && cFleshLastDeciseconds >= 0) {
				String					text;
				String					time;
				
				time = (cFleshLastDeciseconds / 600) + ":" + ((cFleshLastDeciseconds / 100) % 6) + "" + ((cFleshLastDeciseconds / 10) % 10) + "." + (cFleshLastDeciseconds % 10);
				if (cFleshLastBugs == 1)
					text = "Last Score: 1 bug in " + time;
				else
					text = "Last Score: " + cFleshLastBugs + " bugs in " + time;
				
				cBufferGraphics.drawString(text, cModeButtonBounds[0].x * 2 + cModeButtonBounds[0].width + 3, getHeight() - 20 - cBufferGraphics.getFontMetrics().getHeight());
			}
		}
	}
	
	public void updateGame() {
		cCanvas.paint(null, cFrame); // paints to cBuffer
		
		// draw statistics
		if (cGameMode == kNormalMode) {
			String					bugs;
			String					lives;
			
			if (cHeadSprite.bugsLeft() != 1)
				bugs = cHeadSprite.bugsLeft() + " bugs";
			else
				bugs = "1 bug";
			
			if (SnakeHead.sLives > 1)
				lives = SnakeHead.sLives + " lives";
			else
				lives = "last life";
			
			cBufferGraphics.setFont(cSmallFont);
			drawShadowedString(bugs, getWidth() - cBufferGraphics.getFontMetrics().stringWidth(bugs) - 4, 15);
			drawShadowedString(lives, getWidth() - cBufferGraphics.getFontMetrics().stringWidth(lives) - 4, 30);
		} else if (cGameMode == kFleshMode) {
			String					bugs;
			String					time;
			
			if (SnakeHead.sTotalBugs != 1)
				bugs = SnakeHead.sTotalBugs + " bugs";
			else
				bugs = "1 bug";
			
			if (!cHeadSprite.cDead)
				cDeciseconds = (cFrame - cCreateCanvasFrame - kNoShowFrameCount) * cFrameLength / 100;
			
			time = (cDeciseconds / 600) + ":" + ((cDeciseconds / 100) % 6) + "" + ((cDeciseconds / 10) % 10) + "." + (cDeciseconds % 10);
			
			cBufferGraphics.setFont(cSmallFont);
			drawShadowedString(bugs, getWidth() - cBufferGraphics.getFontMetrics().stringWidth(bugs) - 4, 15);
			drawShadowedString(time, getWidth() - cBufferGraphics.getFontMetrics().stringWidth(time) - 4, 30);
		}
		
		int							fadeFrame;
		
		fadeFrame = cFrame - cStartFadeFrame;
		
		if (fadeFrame > kFadeFrameCount + 8)
			fadeFrame = (kNoShowFrameCount + kFadeFrameCount - (cFrame - cCreateCanvasFrame)) * 2 - kFadeFrameCount;
		
		if (fadeFrame > 0 && fadeFrame <= kFadeFrameCount) {
			// draw black frame
			int						focusX;
			int						focusY;
			int						left;
			int						top;
			int						right;
			int						bottom;
			
			focusX = (cHeadSprite.getLocationX() >> 10) - cCanvas.getScrollX();
			focusY = (cHeadSprite.getLocationY() >> 10) - cCanvas.getScrollY();
			
			left = focusX * fadeFrame / kFadeFrameCount;
			top = focusY * fadeFrame / kFadeFrameCount;
			right = getWidth() - ((getWidth() - focusX) * fadeFrame / kFadeFrameCount);
			bottom = getHeight() - ((getHeight() - focusY) * fadeFrame / kFadeFrameCount);
			
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.fillRect(0, 0, left, getHeight());
			cBufferGraphics.fillRect(right, 0, getWidth() - right, getHeight());
			
			cBufferGraphics.fillRect(left, 0, right - left, top);
			cBufferGraphics.fillRect(left, bottom, right - left, getWidth() - bottom);
		} else if (fadeFrame > 0 && fadeFrame <= kFadeFrameCount + 8) {
			// frame fills entire window
			cBufferGraphics.setColor(Color.black);
			cBufferGraphics.fillRect(0, 0, getWidth(), getHeight());
		}
		
		if (cGamePaused) {
			// draw "PAUSE"
			cBufferGraphics.setFont(cLargeFont);
			drawShadowedString("PAUSE", (getWidth() - cBufferGraphics.getFontMetrics().stringWidth("PAUSE")) / 2, getHeight() / 2);
			
			cBufferGraphics.setFont(cSmallFont);
			drawShadowedString("Click to Continue", (getWidth() - cBufferGraphics.getFontMetrics().stringWidth("Click to Continue")) / 2, getHeight() / 2 + 20);
		} else if (cFrame - cCreateCanvasFrame < 50) {
			if (cGameMode == kNormalMode) {
				// draw "Level x" / "y bugs to eat"
				String				levelx;
				String				ybugs;
				
				levelx = "Level " + (cLevelIndex + 1);
				ybugs = cHeadSprite.bugsLeft() + " bugs to eat";
				
				cBufferGraphics.setFont(cLargeFont);
				drawShadowedString(levelx, (getWidth() - cBufferGraphics.getFontMetrics().stringWidth(levelx)) / 2, getHeight() / 2);
				
				cBufferGraphics.setFont(cSmallFont);
				drawShadowedString(ybugs, (getWidth() - cBufferGraphics.getFontMetrics().stringWidth(ybugs)) / 2, getHeight() / 2 + 20);
			} else if (cGameMode == kFleshMode) {
				// draw "Go Gain Weight"
				String				text;
				
				text = "Go Gain Weight";
				
				cBufferGraphics.setFont(cSmallFont);
				drawShadowedString(text, (getWidth() - cBufferGraphics.getFontMetrics().stringWidth(text)) / 2, getHeight() / 2 + 20);
			}
		}
	}
	
	/*
	repaint() has been called.
	*/
	public synchronized void update(Graphics theGraphics) {
		if (cGameInProgress) {
			updateGame();
		} else {
			if (cLoadingStatus == kPreLoading) {
				startLoading();
				cLoadingStatus = kLoading;
			}
			
			if (cLoadingStatus == kFailedLoading) {
				cBufferGraphics.setColor(Color.black);
				cBufferGraphics.fillRect(0, 0, getWidth(), getHeight());
				cBufferGraphics.setColor(Color.white);
				if (getParameter("file") == null)
					drawWrappedStringLines("Failed loading data.\nThe applet failed to locate the data file, because the HTML page you are viewing was improperly set up.", 10, 20, getWidth() - 20);
				else
					drawWrappedStringLines("Failed loading data.\nPossible causes are:\n- The applet failed to locate the data file on the server.\n- The host server is having problems.\n- There was a problem transferring the data file information from the server to your computer.\n- The applet encountered unforseen internal problems.\nThe applet can not proceed after failing to load the data.", 10, 20, getWidth() - 20);
			} else
				updateTitleScreen();
		}
		
		theGraphics.drawImage(cBuffer, 0, 0, null);
	}
	
	/*
	The view is "damaged".
	*/
	public void paint(Graphics theGraphics) {
		update(theGraphics);
	}
	
	public void mouseMoved(MouseEvent theEvent) {
		int							buttonCounter;
		boolean						newHilited;
		
		for (buttonCounter = 0; buttonCounter < kGameModeCount; buttonCounter++) {
			if (theEvent.getX() >= cModeButtonBounds[buttonCounter].x &&
					theEvent.getY() >= cModeButtonBounds[buttonCounter].y &&
					theEvent.getX() <= cModeButtonBounds[buttonCounter].x + cModeButtonBounds[buttonCounter].width &&
					theEvent.getY() <= cModeButtonBounds[buttonCounter].y + cModeButtonBounds[buttonCounter].height)
				newHilited = true;
			else
				newHilited = false;
			
			if (cModeButtonHilited[buttonCounter] != newHilited) {
				cModeButtonHilited[buttonCounter] = newHilited;
				if (!cGameInProgress)
					repaint();
			}
		}
	}
	
	public void mouseDragged(MouseEvent theEvent) {
		mouseMoved(theEvent);
	}
	
	public void mousePressed(MouseEvent theEvent) {
		int							buttonCounter;
		
		if (cLoadingStatus == kDoneLoading)
			for (buttonCounter = 0; buttonCounter < kGameModeCount; buttonCounter++)
				if (cModeButtonHilited[buttonCounter]) {
					cModeButtonPressed[buttonCounter] = true;
					repaint();
				}
	}
	
	public void mouseReleased(MouseEvent theEvent) {
		int						buttonCounter;
		
		for (buttonCounter = 0; buttonCounter < kGameModeCount; buttonCounter++)
			if (cModeButtonPressed[buttonCounter]) {
				cModeButtonPressed[buttonCounter] = false;
				
				if (!cGameInProgress) {
					if (cModeButtonHilited[buttonCounter]) {
						newGame(buttonCounter);
						return;
					} else
						repaint();
				}
			}
		
		if (cGameInProgress)
			pauseGame(!cGamePaused);
	}
	
	public void keyPressed(KeyEvent theEvent) {
		int							keyEventCode;
		
		super.keyPressed(theEvent);
		
		keyEventCode = theEvent.getKeyCode();
		
		if (cLoadingStatus == kDoneLoading) {
			if (!cGameInProgress && (keyEventCode == KeyEvent.VK_1 || keyEventCode == KeyEvent.VK_ENTER))
				newGame(kNormalMode);
			else if (!cGameInProgress && (keyEventCode == KeyEvent.VK_2 || keyEventCode == KeyEvent.VK_SPACE))
				newGame(kFleshMode);
			else if (cGameInProgress && keyEventCode == KeyEvent.VK_E)
				endGame();
		}
	}
	
	public void mouseClicked(MouseEvent theEvent) {}
	public void mouseEntered(MouseEvent theEvent) {}
	public void mouseExited(MouseEvent theEvent) {}
}

